/* Checks files for consitency against information in checksums and attrs */

/* {{{ Apache License version 2.0
 */
/*
 * Copyright 2003-2018 David Shea <david@reallylongword.org>
 *                     David Cantrell <david.l.cantrell@gmail.com>
 *                     Chris Lumens <chris@bangmoney.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/* }}} */

#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <openssl/md5.h>
#include <openssl/sha.h>
#include <endian.h>
#include <limits.h>
#include "dist-tools.h"
#include "db.h"
#include "signature.h"

#define MODE_ERR(str) { \
   fprintf(stderr, "Unrecognized mode string %s\n", (str)); \
   return -1; \
}

/* Local prototypes */
static char *print_type(mode_t);

/* do the weird ones first, since that's probably why this was called */
static char *print_type(mode_t mode) {
    if (S_ISSOCK(mode)) {
        return "socket";
    } else if (S_ISFIFO(mode)) {
        return "named pipe";
    } else if (S_ISBLK(mode)) {
        return "block device";
    } else if (S_ISCHR(mode)) {
        return "character device";
    } else if (S_ISREG(mode)) {
        return "regular file";
    } else if (S_ISDIR(mode)) {
        return "directory";
    } else if (S_ISLNK(mode)) {
        return "symbolic link";
    } else {
        return "unknown";
    }
}

/*
 * checks the file for consistency
 * First ensures the type is correct
 * Next verifies the checksum or symlinkage, if applicable
 * if strict checking requested and is a hardlink, check linkage on that
 * lastly, verifies uid/gid/mode if strict checking requested
 *
 * 'repair' is a set of flags, it means distchk has called us with the
 * repair-in-place option specified.  These are the available flags:
 *
 *     OPTS_NO_REPAIR    Do not attempt a database repair (default).
 *     OPTS_REPAIR       Perform the database repair.
 *     OPTS_DRY_RUN      Used with OPTS_REPAIR to show repair steps but
 *                       not actually perform the repair.
 *
 * Here is the basic idea behind repair:
 *
 * 1) If the specified file does not exist on the filesystem, remove it
 *    from the package database.
 * 2) If the stored checksum does not match the file on disk, update the
 *    database.
 * 3) If the UID and/or GID do not match the database, update the database.
 * 4) If the device major and minor numbers do not match, update the
 *    database to reflect the filesystem.
 * 5) If the file is a soft or hard link and does not match what is in the
 *    database, update the linkdest.
 *
 * Note that repair only updates the package database.  It does not change
 * files or directories on the system.  The option should be used with
 * caution as you may not always want to blindly update the database to
 * reflect the system.  The --dry-run option is provided as a test.
 */
int checkfile(const struct finfo *info, char *prepend,
              const bool strict, const uint64_t repair) {
    char *filename = NULL;
    char *buf = NULL;
    struct stat realattr;
    struct stat test;
    int i = 0;
    int success = 0;
    enum checksum sumtype = NULLSUM;

    assert(info != NULL);

    /*
     * Remember we store paths without a leading slash.  Use either the
     * prepend value (which indicates the user specified an alternate root
     * path) or '/'.
     */
    filename = joinpath(prepend, info->filename, NULL);

    if (lstat(filename, &realattr) == -1) {
        if (errno == ENOENT) {
            fprintf(stderr, "*** %s: %s not found\n", __func__, filename);
        } else {
            fprintf(stderr, "*** %s: could not stat %s: %s\n", __func__,
                    filename, strerror(errno));
        }

        fflush(stderr);
        return -1;
    }

    switch (info->type) {
        case TYPE_REG:
            if (!S_ISREG(realattr.st_mode)) {
                printf("%s is not a regular file\n", filename);
                fflush(stdout);
                return -1;
            }

            if ((info->checksum) == NULL) {
                printf("%s has no checksum\n", filename);
                fflush(stdout);
                success = -1;
                break;
            }

            if (!strcmp(info->checksum, "*")) {
                success = 0;
                break;
            }

            /* determine actual checksum type to use here */
            if (strlen(info->checksum) == (MD5_DIGEST_LENGTH * 2)) {
                sumtype = MD5SUM;
            } else if (strlen(info->checksum) == (SHA_DIGEST_LENGTH * 2)) {
                sumtype = SHA1SUM;
            } else if (strlen(info->checksum) == (SHA256_DIGEST_LENGTH * 2)) {
                sumtype = SHA256SUM;
            }

            if (sumtype < 0) {
                printf("unknown checksum type used on %s\n", filename);
                fflush(stdout);
                success = -1;
                break;
            }

            buf = checksum(filename, sumtype);

            if (buf == NULL) {
                printf("unable to read checksum for %s\n", filename);
                fflush(stdout);
                success = -1;
            } else if (strcmp(buf, info->checksum)) {
                printf("bad checksum for %s\n", filename);
                fflush(stdout);
                success = -1;
            }

            free(buf);
            break;
        case TYPE_DIR:
            if (!S_ISDIR(realattr.st_mode)) {
                printf("%s is not a directory\n", filename);
                fflush(stdout);
                return -1;
            }

            break;
        case TYPE_SOFT:
            if (!S_ISLNK(realattr.st_mode)) {
                printf("%s is not a symbolic link\n", filename);
                fflush(stdout);
                return -1;
            } else if (info->linkname == NULL) {
                printf("%s has no symlink information\n", filename);
                fflush(stdout);
                success = -1;
            } else {
                if ((buf = calloc(1, PATH_MAX)) == NULL) {
                    fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__,
                            strerror(errno));
                    fflush(stderr);
                    abort();
                }

                if ((i = readlink(filename, buf, PATH_MAX)) == -1) {
                    printf("error reading symlink information for %s: %s\n",
                           filename, strerror(errno));
                    fflush(stdout);
                    success = -1;
                }

                /* make sure our readlink value is NUL terminated */
                buf[i >= PATH_MAX ? PATH_MAX - 1 : i] = '\0';

                if (strcmp(buf, info->linkname)) {
                    printf("symlink %s should point to %s; "
                           "points to %s instead\n", filename,
                           info->linkname, buf);
                    fflush(stdout);
                    success = -1;
                }

                free(buf);
            }

            break;
        case TYPE_HARD:
            /* nothing to do in this particular block for hard links */
            break;
        case TYPE_CHAR:
            if (!S_ISCHR(realattr.st_mode)) {
                printf("%s is not a character device\n", filename);
                fflush(stdout);
                return -1;
            }

            if (realattr.st_rdev != info->attr->st_rdev) {
                printf("device number inconsistency on %s "
                       "(expected %ld, found %ld)\n",
                       filename, realattr.st_rdev, info->attr->st_rdev);
                fflush(stdout);
                return -1;
            }

            break;
        case TYPE_BLOCK:
            if (!S_ISBLK(realattr.st_mode)) {
                printf("%s is not a block device\n", filename);
                fflush(stdout);
                return -1;
            }

            if (realattr.st_rdev != info->attr->st_rdev) {
                printf("device number inconsistency on %s "
                       "(expected %ld, found %ld)\n",
                       filename, realattr.st_rdev, info->attr->st_rdev);
                fflush(stdout);
                return -1;
            }

            break;
        case TYPE_FIFO:
            if (!S_ISFIFO(realattr.st_mode)) {
                printf("%s is not a fifo\n", filename);
                fflush(stdout);
                return -1;
            }

            break;
        case TYPE_SOCK:
            if (!S_ISSOCK(realattr.st_mode)) {
                printf("%s is not a socket\n", filename);
                fflush(stdout);
                return -1;
            }

            break;
        default:
            printf("file type of %s (%s) not supported\n",
                   filename, print_type(realattr.st_mode));
            fflush(stdout);
            return -1;
    }

    if (strict) {
        if ((info->type) == TYPE_HARD) {
            if (info->linkname == NULL) {
                printf("file %s has no hardlink information\n", filename);
                fflush(stdout);
                success = -1;
            } else {
                buf = joinpath(prepend, info->linkname, NULL);

                if (lstat(buf, &test) == -1) {
                    fprintf(stderr, "*** unable to stat %s: %s\n",
                            buf, strerror(errno));
                    fflush(stderr);
                    success = -1;
                }

                if (test.st_ino != realattr.st_ino) {
                    printf("%s is not hardlinked to %s\n", filename, buf);
                    fflush(stdout);
                    success = -1;
                }

                free(buf);
            }
        }

        if (info->attr != NULL) {
            if (info->attr->st_uid != realattr.st_uid) {
                printf("UID of %s (%ld) does not match expected (%ld)\n",
                       filename, (long) realattr.st_uid,
                       (long) (info->attr->st_uid));
                fflush(stdout);
                success = -1;
            }

            if (info->attr->st_gid != realattr.st_gid) {
                printf("GID of %s (%ld) does not match expected (%ld)\n",
                       filename, (long) realattr.st_gid,
                       (long) (info->attr->st_gid));
                fflush(stdout);
                success = -1;
            }

            if (info->type != TYPE_SOFT &&
                ((info->attr->st_mode) & S_INTER) !=
                 ((realattr.st_mode) & S_INTER)) {
                printf("mode of %s (%#4o) does not match expected (%#4o)\n",
                       filename, (realattr.st_mode) & S_INTER,
                       (info->attr->st_mode) & S_INTER);
                fflush(stdout);
                success = -1;
            }
        }
    }

    free(filename);
    return success;
}

/*
 * Reads the package contents from the database, returns table of filenames.
 * Directories and hardlinks (optionally) returned in lists, since order
 * matters on these.
 *
 * If the packageid is < 0, assume we are reading a bom.db and the contents
 * table should be read entirely.  Otherwise, we are reading the installed
 * package database and need to select rows that match the packageid.
 */
int read_contents(sqlite3 *db, int packageid, enum signature sigver,
                  finfolist_t **manifest, finfolist_t *dirs, finfolist_t *hards) {
    struct finfo *info = NULL;
    sqlite3_stmt *stmt = NULL;
    mode_t mode = 0;

    assert(db != NULL);
    assert(manifest != NULL);

    /* build correct prepared statement */
    if (sqlite3_prepare(db, CONTENTS_SELECT, -1, &stmt, 0) != SQLITE_OK) {
        FAIL_SQLITE(db, NULL, -1);
    }

    *manifest = init_finfolist();

    /* iterate over each row of the results */
    while (sqlite3_step(stmt) == SQLITE_ROW) {
        /* allocate memory for this entry */
        if ((info = calloc(1, sizeof(*info))) == NULL) {
            fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
            fflush(stderr);
            abort();
        }

        if ((info->attr = calloc(1, sizeof(*(info->attr)))) == NULL) {
            fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
            fflush(stderr);
            abort();
        }

        /* initialize all info pointers */
        info->filename = NULL;
        info->fullpath = NULL;
        info->checksum = NULL;
        info->linkname = NULL;

        /* populate struct info with bom.db data */
        info->type = sqlite3_column_int(stmt, 8);
        info->filename = strdup((char *) sqlite3_column_text(stmt, 0));

        if (sigver == SIGVER_21) {
            /*
             * This is a known bug in v2.1 packages.  If you build a noarch
             * package, the host you install it on must match the endianness
             * of the build host.  That's why for v2.2 packages, we store the
             * mode in big endian format in the packages.
             */
            memcpy(&(info->attr->st_mode), sqlite3_column_blob(stmt, 5), sqlite3_column_bytes(stmt, 5));
        } else if (sigver >= SIGVER_22) {
            memcpy(&mode, sqlite3_column_blob(stmt, 5), sqlite3_column_bytes(stmt, 5));
            mode = be32toh(mode);
            memcpy(&(info->attr->st_mode), &mode, sizeof(mode));
        } else {
            fprintf(stderr, "%s (%d): unknown package signature %d\n", __func__, __LINE__, sigver);
            fflush(stderr);
            abort();
        }

        info->attr->st_uid = (uid_t) sqlite3_column_int(stmt, 2);
        info->attr->st_gid = (gid_t) sqlite3_column_int(stmt, 4);
        info->attr->st_size = (off_t) sqlite3_column_int(stmt, 6);

        if (sqlite3_column_text(stmt, 9) != NULL) {
            info->linkname = strdup((char *) sqlite3_column_text(stmt, 9));
        }

        if (sqlite3_column_text(stmt, 10) != NULL) {
            info->checksum = strdup((char *) sqlite3_column_text(stmt, 10));
        }

        /* if we have no checksum, but expect one, display a warning */
        if (info->checksum == NULL && info->type == TYPE_REG) {
            fprintf(stderr, "%s: WARNING checksum for %s not found\n", __func__, info->filename);
            fflush(stderr);
        }

        /* save dirs and hard links in appropriate lists */
        if (info->type == TYPE_DIR && dirs != NULL) {
            dirs = sorted_finfolist_insert(dirs, dup_finfo(info));
        }

        if (info->type == TYPE_HARD && hards != NULL) {
            hards = sorted_finfolist_insert(hards, dup_finfo(info));
        }

        /* record the entry and prep info for the next one */
        *manifest = sorted_finfolist_insert(*manifest, info);
    }

    return 0;
}

/* vim:set shiftwidth=3 softtabstop=3: */
