/* Check the integrity of an installed dist package. */

/* {{{ Apache License version 2.0
 */
/*
 * Copyright 2015-2018 David Cantrell <david.l.cantrell@gmail.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/* }}} */

#include <stdlib.h>
#include <stdint.h>
#include <assert.h>
#include <string.h>
#include "dist-tools.h"
#include "db.h"

/*
 * Perform integrity checks on the specified package,
 * optionally repair database.
 */
int checkpkg(char *dbpath, struct dist *package, uint64_t repair, char *prepend) {
    int ret = 0;
    char *query = NULL;
    char *err = NULL;
    sqlite3 *db = NULL;
    sqlite3_stmt *stmt = NULL;
    struct finfo fi;

    assert(dbpath != NULL);
    assert(package != NULL);

    /* open a connection to our database */
    if (sqlite3_open(dbpath, &db) != SQLITE_OK) {
        FAIL_SQLITE(db, NULL, -1);
    }

    if (sqlite3_exec(db, "PRAGMA synchronous = OFF", NULL, NULL, &err) != SQLITE_OK) {
        FAIL_SQLITE(db, err, -1);
    }

    if (sqlite3_exec(db, "PRAGMA journal_mode = MEMORY", NULL, NULL, &err) != SQLITE_OK) {
        FAIL_SQLITE(db, err, -1);
    }

    /* iterate and check contents */
    xasprintf(&query, PACKAGE_CONTENTS, package->id);

    if (sqlite3_prepare(db, query, -1, &stmt, 0) != SQLITE_OK) {
        FAIL_SQLITE(db, NULL, -2);
    }

    while (sqlite3_step(stmt) == SQLITE_ROW) {
        /* build a struct finfo for each package entry */
        if (read_finfo(package->id, package->sigver, stmt, &fi, prepend)) {
            fprintf(stderr, "error reading finfo for some row\n");
            fflush(stderr);
            ret = -3;
        }

        /* check this package entry */
        if (checkfile(&fi, prepend, true, repair)) {
            ret = -4;
        }

        /* reset the struct finfo */
        free_finfo(&fi);
    }

    if (sqlite3_finalize(stmt) != SQLITE_OK) {
        FAIL_SQLITE(db, NULL, -5);
    }

    sqlite3_close(db);
    free(query);
    return ret;
}
