/* Check a package file for a valid signature. */

/* {{{ Apache License version 2.0
 */
/*
 * Copyright 2003-2018 David Cantrell <david.l.cantrell@gmail.com>
 *                     David Shea <david@reallylongword.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/* }}} */

#include <stdlib.h>
#include <stdbool.h>
#include <assert.h>
#include <string.h>
#include "dist-tools.h"

/*
 * A dist package will have a file in the archive named 'signature'.  It
 * will contain the text defined in SIGNATURE.  To verify a file is a
 * package, we check for the signature in the archive.  Not foolproof, but
 * ok for our purposes.
 */
bool checksig(char *filename) {
    int prefixlen = strlen(SIGNATURE);
    int verlen = strlen(SIGNATURE_IN_USE);
    char *buf = NULL;
    bool have_signature = true;
    bool valid_signature = false;

    assert(filename != NULL);

    if ((buf = calloc(1, BUFSIZ)) == NULL) {
        fprintf(stderr, "*** error allocating buf\n");
        abort();
    }

    if (unpack_metadata2str(filename, DIST_SIGNATURE, buf, true)) {
        return false;
    }

    /* trim trailing newline and/or carriage return characters */
    buf[strcspn(buf, "\r\n")] = '\0';

    if (strlen(buf) > (prefixlen + verlen)) {
        free(buf);
        return false;
    }

    /* Check for valid signature prefix followed by a valid version. */
    if (!strncmp(buf, SIGNATURE, prefixlen) &&
        (!strncmp(buf+prefixlen, SIGNATURE_21, verlen) ||
         !strncmp(buf+prefixlen, SIGNATURE_22, verlen))) {
        valid_signature = true;
    }

    free(buf);

    if (have_signature && valid_signature) {
        return true;
    } else {
        return false;
    }
}
