/* Check the status of an installed dist package. */

/* {{{ Apache License version 2.0
 */
/*
 * Copyright 2003-2018 David Shea <david@reallylongword.org>
 *                     David Cantrell <david.l.cantrell@gmail.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/* }}} */

#include <stdlib.h>
#include <stdbool.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <libgen.h>
#include <getopt.h>
#include "dist-tools.h"
#include "db.h"
#include "opts.h"

/* Local prototypes */
static void usage(void);
static void _checkpkg(struct dist *, uint64_t);

/* Local global variables */
static char *prog = NULL;
static char *db_installed = NULL;
static char *prepend = NULL;
static int success = EXIT_SUCCESS;
static bool first = true;

/* Command line options */
static struct option long_options[] = {
    {"repair", no_argument, NULL, 'R'},
    {"dry-run", no_argument, NULL, 'd'},
    {"root", required_argument, NULL, 'r'},
    {"version", no_argument, NULL, 'V'},
    {"help", no_argument, NULL, 'h'},
    {NULL, 0, NULL, 0}
};

/* Usage screen displayed when the user passes --help as an argument. */
static void usage(void) {
    printf("Usage: %s [OPTIONS] packagespec...\n", prog);
    puts("Check the integrity of an installed Peachtree Linux package\n" "\n"
         "  -R, --repair               Repair database entries to match filesystem\n"
         "  -d, --dry-run              Only show what database repairs would be made\n"
         "  -r PATH, --root=PATH       Use PATH for the root (/) directory\n"
         "  -V, --version              Display version information\n"
         "  -h, --help                 This help message\n");
    return;
}

/*
 * Perform integrity checks on the specified package,
 * optionally repair database.
 */
static void _checkpkg(struct dist *package, uint64_t repair) {
    assert(package != NULL);

    if (first) {
        first = false;
    } else {
        printf("\n");
    }

    printf("=> %s-%s-%d%s.%s <=\n", package->name, package->ver,
           package->rel, (package->debug) ? ".debug" : "",
           package->arch);

    if (checkpkg(db_installed, package, repair, prepend)) {
        success = EXIT_FAILURE;
    } else {
        printf("OK\n");
    }

    fflush(stdout);

    return;
}

/* MAIN */
int main(int argc, char **argv) {
    int i = 0;
    pkglist_t *pkglist = NULL;
    pkglist_entry_t *entry = NULL;
    struct stat sb;
    uint64_t repair = 0;

    prog = basename(argv[0]);
    repair |= OPTS_NO_REPAIR;

    while ((i = getopt_long(argc, argv, "Rdr:Vh?", long_options, NULL)) != -1) {
        switch (i) {
            case 'R':
                repair |= OPTS_REPAIR;
                repair &= ~OPTS_NO_REPAIR;
                break;
            case 'd':
                repair |= OPTS_DRY_RUN;
                break;
            case 'r':
                if (*optarg != '/') {
                    printf("Root directory must be an absolute pathname\n");
                    fflush(stdout);
                    return EXIT_FAILURE;
                }

                prepend = strdup(optarg);
                break;
            case 'V':
                ver(prog);
                return EXIT_SUCCESS;
            case 'h':
            case '?':
                usage();
                return EXIT_SUCCESS;
            default:
                usage();
                return EXIT_FAILURE;
        }
    }

    /* check for mismatched options */
    if ((repair & OPTS_DRY_RUN) && !(repair & OPTS_REPAIR)) {
        fprintf(stderr, "The dry-run option (-d) requires the repair (-R) option.\n");
        fflush(stderr);
        return EXIT_FAILURE;
    }

    /* use / if the user did not specify a root path */
    if (prepend == NULL) {
        prepend = strdup("/");
    }

    /* handle an alternative root path for installation */
    db_installed = get_path(prepend, DB_FILE);

    if (stat(db_installed, &sb) == -1 || !S_ISREG(sb.st_mode)) {
        fprintf(stderr, "This system lacks a Peachtree Linux package database.\n");
        fflush(stderr);
        return EXIT_FAILURE;
    }

    /* no packages specified on the command line, iterate over everything */
    if (argc == optind) {
        pkglist = scanpkg(db_installed, NULL);

        if (pkglist == NULL) {
            return EXIT_FAILURE;
        }

        /* sort the list alphabetically */
        pkglist = sort_pkglist(pkglist);

        /* check the package */
        TAILQ_FOREACH(entry, pkglist, items) {
            _checkpkg(entry->pkg, repair);
        }

        /* free the list of packages */
        free_pkglist(pkglist);
        pkglist = NULL;
    } else {
        /* iterate over the arguments specified */
        for (i = optind; i < argc; i++) {
            /* make a list of possible installed packages */
            pkglist = scanpkg(db_installed, argv[i]);

            if (pkglist == NULL) {
                printf("No installed package matching '%s' found.\n", argv[i]);
                fflush(stdout);
                success = EXIT_FAILURE;
                continue;
            }

            /* require more specification */
            if (length_pkglist(pkglist) > 1) {
                fprintf(stderr, "*** ERROR: More than one version of %s found, "
                        "please specify one to check.\n", argv[i]);
                fflush(stderr);
                free_pkglist(pkglist);
                success = EXIT_FAILURE;
                continue;
            }

            /* check the package */
            TAILQ_FOREACH(entry, pkglist, items) {
                _checkpkg(entry->pkg, repair);
            }

            /* free the list of packages */
            free_pkglist(pkglist);
            pkglist = NULL;
        }
    }

    free(db_installed);
    return success;
}
