/* Make a new dist package. */

/* {{{ Apache License version 2.0
 */
/*
 * Copyright 2003-2018 David Cantrell <david.l.cantrell@gmail.com>
 *                     David Shea <david@reallylongword.org>
 *                     Chris Lumens <chris@bangmoney.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/* }}} */

#include <stdlib.h>
#include <assert.h>
#include <stdbool.h>
#include <string.h>
#include <strings.h>
#include <ftw.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <paths.h>
#include <limits.h>
#include <archive.h>
#include <archive_entry.h>
#include "dist-tools.h"
#include "plist.h"
#include "db.h"
#include "opts.h"
#include "signature.h"

extern char *optarg;
extern int optind, opterr, optopt;

/* Global variables */
/*
 * prepend needs to be set to "/" for use by gendist.c
 * It's really only necessary for -g on distadd, but since distmk calls
 * the same functions, we need to define it in distmk too.
 */
char *prepend = "/";

/* Local prototypes */
static void usage(void);
static void set_archive_compression(unsigned int);
static int _add_payload_entry(const char *, const struct stat *, int, struct FTW *);
static int create_reloc(struct dist *);
static int create_etc(struct dist *);
static int create_dist(struct dist *);
static int mkstagedir(struct dist *);
static int genstrings(struct dist *, char *, char *);
static int cleanup(struct dist *);
static void cpmeta(struct dist *, char *, bool);
static int _tree_entry_size(const char *, const struct stat *, int, struct FTW *);
static int get_tree_size(const char *);
static int dometa(struct dist *);

/* Global variables */
static char *targetarch = NULL;
static uint64_t opts = 0;
static char *prog = NULL;
static char *tmppath = NULL;
static char *staging = NULL;
static int verbose = 0;
static struct archive *output = NULL;
static unsigned int treesize = 0;
static int payload = PAYLOAD_NULL;
static int trimlen = -1;
static unsigned int plist_checked = 0;
static plist_t *plist = NULL;

static struct option long_options[] = {
    {"tree", required_argument, NULL, 't'},
    {"meta", required_argument, NULL, 'm'},
    {"pkg", required_argument, NULL, 'p'},
    {"ver", required_argument, NULL, 'v'},
    {"rel", required_argument, NULL, 'r'},
    {"plist", required_argument, NULL, 'l'},
    {"debug", no_argument, NULL, 'd'},
    {"repo", required_argument, NULL, 'R'},
    {"arch", required_argument, NULL, 'a'},
    {"compression", required_argument, NULL, 'z'},
    {"checksum", required_argument, NULL, 's'},
    {"clean", no_argument, NULL, 'c'},
    {"version", no_argument, NULL, 'V'},
    {"help", no_argument, NULL, 'h'},
    {"verbose", no_argument, &verbose, -1},
    {"tmpdir", required_argument, NULL, 0},
    {NULL, 0, NULL, 0}
};

#define REASON_TREE "Missing package root tree (-t)"
#define REASON_PKGNAME "Missing package name (-p)"
#define REASON_PKGVERSION "Missing package version (-v)"
#define REASON_PKGRELEASE "Missing package release (-r)"
#define REASON_PLIST "Missing path to plist file (-l)"

/*
 * Usage screen.
 */
static void usage(void) {
    printf("Usage: %s [OPTION]...\n", prog);
    puts("Make a Peachtree Linux dist package\n"
         "\nRequired:\n"
         "  -t PATH, --tree=PATH         Source path to package root tree\n"
         "  -p NAME, --pkg=NAME          Package name (e.g., dist-tools)\n"
         "  -v VER, --ver=VER            Package version (e.g., 2.0)\n"
         "  -r REL, --rel=REL            Package release (e.g., 1)\n"
         "  -l FILE, --plist=FILE        Path to the packing list file\n"
         "\nOptional:\n"
         "  -d, --debug                  Create a debug package\n"
         "  -m PATH, --meta=PATH         Source path to package meta data\n"
         "  -R REPO, --repo=REPO         Package repository name\n"
         "  -a ARCH, --arch=ARCH         Target architecture\n"
         "  -o PATH, --outdir=PATH       Output directory for final package\n"
         "  -z TYPE, --compression=TYPE  Set compression type (see below)\n"
         "  -s TYPE, --checksum=TYPE     Set the checksum type (see below)\n"
         "  -c, --clean                  Clean up temporary packaging files\n"
         "  --verbose                    Display more output\n"
         "  --tmpdir=PATH                Use PATH for temporary files");
    printf("                               (default: %s)\n", _PATH_VARTMP);
    puts("  -V, --version                Display version information\n"
         "  -h, --help                   This help message\n");
    printf("Target architecture is '%s' by default, ", targetarch);
    printf("but can be explicitly\nset to 'noarch'.\n\n");
    puts("Compression type is LZMA by default, but you may specify any of\n"
         "the following types:\n"
         "    bzip2, bz2               bzip2(1) compression\n"
         "    gzip, gz                 gzip(1) compression\n"
         "    compress, Z              compress(1) compression\n"
         "    xz                       xz(1) compression\n"
         "    lzma                     LZMA compression (default)\n"
         "    none                     Do not compress payload\n\n"
         "Checksum type is SHA1 by default, but you may specify any of the\n"
         "following types:\n"
         "    md5                      MD5 128-bit, as described in RFC 1321\n"
         "    sha1                     SHA-1 160-bit, as described in FIPS-180-1\n"
         "    sha256                   SHA-256 256-bit, as described in FIPS-180-2 \n\n"
         "Both compression type and checksum type are compared as case\n"
         "insensitive strings.");
    return;
}

/*
 * Set the compression type on the payload based on what the user
 * specified.  The default is LZMA set in initpkg().
 */
static void set_archive_compression(unsigned int compression) {
    switch (compression) {
        case ARCHIVE_COMPRESSION_LZMA:
#if ARCHIVE_VERSION_NUMBER < 3000000
            archive_write_set_compression_lzma(output);
#else
            archive_write_add_filter_lzma(output);
#endif
            break;
        case ARCHIVE_COMPRESSION_BZIP2:
#if ARCHIVE_VERSION_NUMBER < 3000000
            archive_write_set_compression_bzip2(output);
#else
            archive_write_add_filter_bzip2(output);
#endif
            break;
        case ARCHIVE_COMPRESSION_COMPRESS:
#if ARCHIVE_VERSION_NUMBER < 3000000
            archive_write_set_compression_compress(output);
#else
            archive_write_add_filter_compress(output);
#endif
            break;
        case ARCHIVE_COMPRESSION_GZIP:
#if ARCHIVE_VERSION_NUMBER < 3000000
            archive_write_set_compression_gzip(output);
#else
            archive_write_add_filter_gzip(output);
#endif
            break;
        case ARCHIVE_COMPRESSION_XZ:
#if ARCHIVE_VERSION_NUMBER < 3000000
            archive_write_set_compression_xz(output);
#else
            archive_write_add_filter_xz(output);
#endif
            break;
        case ARCHIVE_COMPRESSION_NONE:
#if ARCHIVE_VERSION_NUMBER < 3000000
            archive_write_set_compression_none(output);
#else
            archive_write_add_filter_none(output);
#endif
            break;
    }

    return;
}

/*
 * Adds an individual entry to a payload
 * (called by create_reloc() and create_etc())
 *
 * Callers need to set up 'output' and reset 'trimlen' to -1 before calling
 * this function.
 */
static int _add_payload_entry(const char *fpath, const struct stat *sb,
                              int tflag, struct FTW *ftwbuf) {
    int fd;
    ssize_t len;
    char buf[BUFSIZ];
    char *hardlink = NULL;
    bool is_hardlink = false;
    const char *vpath = NULL;
    struct archive_entry *entry = NULL;
    plist_entry_t *plistentry = NULL;
    plist_entry_t *iterator = NULL;

    if (trimlen == -1) {
        trimlen = strlen(fpath);
        return 0;
    } else {
        vpath = fpath + trimlen;

        while (*vpath == '/') {
            vpath++;
        }
    }

    /* Skip parts of the tree that do not match this payload type. */
    if ((strncmp(vpath, CONFDIR_PREFIX, strlen(CONFDIR_PREFIX)) && payload == PAYLOAD_ETC) ||
        (!strncmp(vpath, CONFDIR_PREFIX, strlen(CONFDIR_PREFIX)) && payload == PAYLOAD_RELOC)) {
        return 0;
    }

    entry = archive_entry_new();

    /* Is this file a hard link? */
    hardlink = lookup_inode(sb->st_ino);
    if (hardlink != NULL && strcmp(vpath, hardlink)) {
        is_hardlink = true;
    }

    if (S_ISREG(sb->st_mode)) {
        if (is_hardlink) {
            archive_entry_set_hardlink(entry, hardlink);
        } else {
            archive_entry_set_filetype(entry, AE_IFREG);
        }
    } else if (S_ISLNK(sb->st_mode) && (tflag == FTW_SL || tflag == FTW_SLN)) {
        archive_entry_set_filetype(entry, AE_IFLNK);
        memset(buf, '\0', sizeof(buf));

        if (readlink(fpath, buf, sizeof(buf)) == -1) {
            fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__,
                    strerror(errno));
            return -1;
        }

        archive_entry_set_symlink(entry, buf);
    } else if (S_ISSOCK(sb->st_mode)) {
        archive_entry_set_filetype(entry, AE_IFSOCK);
    } else if (S_ISCHR(sb->st_mode)) {
        archive_entry_set_filetype(entry, AE_IFCHR);
    } else if (S_ISBLK(sb->st_mode)) {
        archive_entry_set_filetype(entry, AE_IFBLK);
    } else if (S_ISDIR(sb->st_mode)) {
        archive_entry_set_filetype(entry, AE_IFDIR);
    } else if (S_ISFIFO(sb->st_mode)) {
        archive_entry_set_filetype(entry, AE_IFIFO);
    } else {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__,
                strerror(errno));
        return -1;
    }

    archive_entry_set_pathname(entry, vpath);
    archive_entry_set_atime(entry, sb->st_atime, 0);
    archive_entry_set_mtime(entry, sb->st_mtime, 0);
    archive_entry_set_ctime(entry, sb->st_ctime, 0);
    archive_entry_set_perm(entry, sb->st_mode);
    archive_entry_set_gid(entry, sb->st_uid);
    archive_entry_set_uid(entry, sb->st_gid);

    if (!is_hardlink) {
        archive_entry_set_size(entry, sb->st_size);
    }

    if (payload != PAYLOAD_DIST) {
        if (verbose) {
            printf(">>> ");

            if (payload == PAYLOAD_RELOC) {
                printf("reloc");
            } else if (payload == PAYLOAD_ETC) {
                printf("  "CONFDIR_PREFIX);
            }

            printf(": /%s\n", vpath);
        }

        /*
         * We get the mode, owner, and group from the plist file.  We also
         * check and ensure there are no entries in the package root tree
         * that are not specified in the plist file.
         */
        TAILQ_FOREACH(iterator, plist, items) {
            if (!strcmp(iterator->filename, vpath)) {
                plistentry = iterator;
                break;
            }
        }

        if (plistentry == NULL) {
            fprintf(stderr, "*** WARNING: %s not specified in plist file.\n", vpath);
        } else {
            if (!plistentry->is_symlink && S_ISLNK(sb->st_mode)) {
                fprintf(stderr, "*** WARNING: %s not a 'symlink' in plist file.\n", vpath);
            }

            if (!plistentry->is_symlink) {
                archive_entry_set_perm(entry, plistentry->mode);
            }

            archive_entry_set_gid(entry, plistentry->uid);
            archive_entry_set_uid(entry, plistentry->gid);
            plist_checked++;
        }
    }

    /* write the entry header */
    archive_write_header(output, entry);

    /* write the data to the payload for regular files only */
    if (S_ISREG(sb->st_mode) && !is_hardlink) {
        if ((fd = open(fpath, O_RDONLY)) == -1) {
            fprintf(stderr, "%s (%d): %s: %s\n", __func__, __LINE__,
                    strerror(errno), fpath);
            return -1;
        } else {
            memset(buf, 0, sizeof(buf));
        }

        if ((len = read(fd, buf, sizeof(buf))) == -1) {
            fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__,
                    strerror(errno));
            return -1;
        }

        while (len > 0) {
            archive_write_data(output, buf, len);

            if ((len = read(fd, buf, sizeof(buf))) == -1) {
                fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__,
                        strerror(errno));
                return -1;
            }
        }

        if (close(fd) == -1) {
            fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__,
                    strerror(errno));
            return -1;
        }
    }

    archive_entry_free(entry);
    return 0;
}

/*
 * Create the DIST_RELOC file of package->tree, excluding /etc.
 */
static int create_reloc(struct dist *package) {
    int flags = FTW_MOUNT | FTW_PHYS;

    output = archive_write_new();
    set_archive_compression(package->compression);
    archive_write_set_format_pax_restricted(output);
    archive_write_open_filename(output, package->reloc);

    /* add all tree directories except /etc to the payload */
    payload = PAYLOAD_RELOC;
    trimlen = -1;

    if (nftw(package->tree, _add_payload_entry, 25, flags) == -1) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__,
                strerror(errno));
#if ARCHIVE_VERSION_NUMBER < 3000000
        archive_write_finish(output);
#else
        archive_write_free(output);
#endif
        return -1;
    }

    archive_write_close(output);
#if ARCHIVE_VERSION_NUMBER < 3000000
    archive_write_finish(output);
#else
    archive_write_free(output);
#endif

    return 0;
}

/*
 * Pack up /etc tree if it exists.
 */
static int create_etc(struct dist *package) {
    int flags = FTW_MOUNT | FTW_PHYS;
    struct stat sb;

    /* debug packages do not have configuration files */
    if (package->debug) {
        return 0;
    }

    memset(&sb, 0, sizeof(sb));

    /* do we even have an /etc directory? */
    if (chdir(package->tree)) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__,
                strerror(errno));
        return -1;
    } else if (lstat(CONFDIR_PREFIX, &sb) || !S_ISDIR(sb.st_mode)) {
        return 0;
    }

    /* create an /etc payload */
    output = archive_write_new();
    set_archive_compression(package->compression);
    archive_write_set_format_pax_restricted(output);
    archive_write_open_filename(output, package->etc);

    payload = PAYLOAD_ETC;
    trimlen = -1;

    if (nftw(package->tree, _add_payload_entry, 25, flags) == -1) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__,
                strerror(errno));
#if ARCHIVE_VERSION_NUMBER < 3000000
        archive_write_finish(output);
#else
        archive_write_free(output);
#endif
        return -1;
    }

    archive_write_close(output);
#if ARCHIVE_VERSION_NUMBER < 3000000
    archive_write_finish(output);
#else
    archive_write_free(output);
#endif

    return 0;
}

/*
 * Create the package file.
 */
static int create_dist(struct dist *package) {
    int flags = FTW_MOUNT | FTW_PHYS;
    char cwd[PATH_MAX + 1];

    memset(cwd, '\0', sizeof(cwd));

    if (getcwd(cwd, PATH_MAX) == NULL) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
        abort();
    }

    if (chdir(staging) != 0) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
        return -1;
    }

    /* create dist package */
    output = archive_write_new();
#if ARCHIVE_VERSION_NUMBER < 3000000
    archive_write_set_compression_none(output);
#else
    archive_write_add_filter_none(output);
#endif
    archive_write_set_format_ar_bsd(output);
    archive_write_open_filename(output, package->distname);

    payload = PAYLOAD_DIST;
    trimlen = -1;

    if (nftw(staging, _add_payload_entry, 25, flags) == -1) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
#if ARCHIVE_VERSION_NUMBER < 3000000
        archive_write_finish(output);
#else
        archive_write_free(output);
#endif
        return -1;
    }

    archive_write_close(output);
#if ARCHIVE_VERSION_NUMBER < 3000000
    archive_write_finish(output);
#else
    archive_write_free(output);
#endif

    if (chdir(cwd) != 0) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
        return -1;
    }

    return 0;
}

/*
 * Creating the package staging directory.
 */
static int mkstagedir(struct dist *package) {
    int mode = S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH;
    struct stat statbuf;

    assert(staging != NULL);

    if (stat(staging, &statbuf)) {
        /* directory doesn't exist, go ahead and make it */
        if (mkdirp(staging, mode)) {
            fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__,
                    strerror(errno));
            return -1;
        }
    } else {
        /* we fail if the staging directory already exists */
        if (S_ISDIR(statbuf.st_mode)) {
            fprintf(stderr, "Staging directory for %s already exists\n",
                    package->filename);
        } else {
            fprintf(stderr, "%s already exists\n", package->filename);
        }

        return -1;
    }

    return 0;
}

/*
 * Generate string values in the dist structure.
 */
static int genstrings(struct dist *package, char *targetarch, char *outdir) {
    package->arch = targetarch;

    if (outdir == NULL) {
        if ((outdir = getcwd(outdir, 0)) == NULL) {
            fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
            abort();
        }
    }

    if (package->debug) {
        xasprintf(&(package->filename), "%s-%s-%d.debug.%s", package->name,
                  package->ver, package->rel, package->arch);
    } else {
        xasprintf(&(package->filename), "%s-%s-%d.%s", package->name,
                  package->ver, package->rel, package->arch);
    }

    /* package bill of materials database */
    xasprintf(&(package->bom), "%s/%s", staging, DIST_BOM);

    /* package reloc file full path */
    xasprintf(&(package->reloc), "%s/%s", staging, DIST_RELOC);

    /* package etc file full path */
    xasprintf(&(package->etc), "%s/%s", staging, DIST_ETC);

    /* package distname file full path */
    xasprintf(&(package->distname), "%s/%s.dist", outdir, package->filename);

    return 0;
}

/*
 * Clean up temporary files and directories.
 */
static int cleanup(struct dist *package) {
    int ret = 0;

    if (package->do_cleanup == true) {
        ret = rmtree(staging, false, false);
    }

    /* return memory */
    freepkg(package);

    return ret;
}

/*
 * Copy a meta file to the package staging directory.
 */
static void cpmeta(struct dist *package, char *filename, bool create) {
    char *src = NULL;
    char *dest = NULL;
    struct stat sb;
    FILE *fp = NULL;

    if (package->meta == NULL) {
        return;
    }

    /* set the source and dest filenames */
    xasprintf(&src, "%s/%s", package->meta, filename);
    xasprintf(&dest, "%s/%s", staging, filename);

    /* exit if the source doesn't exist */
    /* or create empty file if create is True */
    if (stat(src, &sb)) {
        if (create == true) {
            if ((fp = fopen(dest, "w")) == NULL) {
                fprintf(stderr, "%s (%d): unable to create %s: %s\n",
                        __func__, __LINE__, dest, strerror(errno));
                free(src);
                free(dest);
                return;
            }

            fprintf(fp, "Missing %s file.\n", src);

            if (fclose(fp) == EOF) {
                fprintf(stderr, "%s (%d): unable to close %s: %s\n",
                        __func__, __LINE__, dest, strerror(errno));
                free(src);
                free(dest);
                return;
            }
        }

        free(src);
        free(dest);
        return;
    }

    /* copy the file */
    if (copyfile(src, dest, true, true)) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__,
                strerror(errno));
    }

    free(src);
    free(dest);
    return;
}

/* Called by get_tree_size() for each directory entry. */
static int _tree_entry_size(const char *fpath, const struct stat *sb,
                            int tflag, struct FTW *ftwbuf) {
    if (tflag == FTW_DNR || tflag == FTW_NS) {
        return -1;
    }

    treesize += sb->st_size;
    return 0;
}

/* Computes the size of the given tree */
static int get_tree_size(const char *tree) {
    int flags = FTW_MOUNT | FTW_PHYS;

    if (tree == NULL) {
        return -1;
    }

    treesize = 0;

    if (nftw(tree, _tree_entry_size, 25, flags) == -1) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__,
                strerror(errno));
        return -1;
    }

    return treesize;
}

/*
 * Add package meta data to the staging directory.
 * NOTE:  When this function is entered, we are already in the pkgroot
 * directory.  Any chdir() out of the current directory needs to take
 * in to account the genattrs() call, which expects us to be in the
 * pkgroot directory.  Just make sure you are chdir()'ed in to pkgroot
 * before genattrs() is called.
 */
static int dometa(struct dist *package) {
    char pkgroot[PATH_MAX + 1];
    char *filename = NULL, *t = NULL, *v = NULL;
    char *line = NULL;
    size_t len = 0;
    FILE *fd = NULL;
    struct stat sb;
    char *err = NULL;
    sqlite3 *bom = NULL;
    sqlite3_stmt *stmt = NULL;

    memset(pkgroot, '\0', sizeof(pkgroot));

    if (getcwd(pkgroot, PATH_MAX) == NULL) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
        abort();
    }

    /* initialize bill of materials */
    if (sqlite3_open(package->bom, &bom) != SQLITE_OK) {
        FAIL_SQLITE(bom, NULL, -1);
    }

    if (sqlite3_exec(bom, "PRAGMA synchronous = OFF", NULL, NULL, &err) != SQLITE_OK) {
        FAIL_SQLITE(bom, err, -1);
    }

    if (sqlite3_exec(bom, "PRAGMA journal_mode = MEMORY", NULL, NULL, &err) != SQLITE_OK) {
        FAIL_SQLITE(bom, err, -1);
    }

    /* add the Peachtree Linux package identifier */
    if (chdir(staging)) {
        FAIL_COMMON(-1);
    }

    if ((fd = fopen("signature", "w")) == NULL) {
        FAIL_COMMON(-1);
    }

    /* write out the current package signature we are using */
    fprintf(fd, "%s%s\n", SIGNATURE, SIGNATURE_IN_USE);

    if (fclose(fd)) {
        FAIL_COMMON(-1);
    }

    /* BOM(contents): write the contents of the reloc payload to the contents table */
    if (sqlite3_exec(bom, BOM_CONTENTS_SCHEMA, NULL, 0, &err) != SQLITE_OK) {
        FAIL_SQLITE(bom, err, -1);
    }

    if (sqlite3_prepare(bom, BOM_CONTENTS_INSERT, -1, &stmt, 0) != SQLITE_OK) {
        FAIL_SQLITE(bom, NULL, -1);
    }

    /* populates the manifest list, to avoid self-dependencies */
    if (makelists(package->tree)) {
        FAIL_COMMON(-1);
    }

    if (chdir(pkgroot)) {
        FAIL_COMMON(-1);
    }

    if (sqlite3_exec(bom, "BEGIN TRANSACTION", NULL, NULL, &err) != SQLITE_OK) {
        FAIL_SQLITE(bom, err, -1);
    }

    if (genattrs(package->tree, stmt, plist, package->sigver, package->sumtype, package->debug)) {
        FAIL_COMMON(-1);
    }

    if (sqlite3_exec(bom, "END TRANSACTION", NULL, NULL, &err) != SQLITE_OK) {
        FAIL_SQLITE(bom, err, -1);
    }

    /* Free manifest GSList */
    genattrs_free();

    /* BOM(deps): write the collected dependencies to the deps table */
    if (sqlite3_exec(bom, BOM_DEPS_SCHEMA, NULL, 0, &err) != SQLITE_OK) {
        FAIL_SQLITE(bom, err, -1);
    }

    if (sqlite3_prepare(bom, BOM_DEPS_INSERT, -1, &stmt, 0) != SQLITE_OK) {
        FAIL_SQLITE(bom, NULL, -1);
    }

    if (sqlite3_exec(bom, "BEGIN TRANSACTION", NULL, NULL, &err) != SQLITE_OK) {
        FAIL_SQLITE(bom, err, -1);
    }

    if (writedeps(package, stmt)) {
        FAIL_COMMON(-1);
    }

    if (sqlite3_exec(bom, "END TRANSACTION", NULL, NULL, &err) != SQLITE_OK) {
        FAIL_SQLITE(bom, err, -1);
    }

    if (sqlite3_finalize(stmt) != SQLITE_OK) {
        FAIL_SQLITE(bom, NULL, -1);
    }

    /* BOM(provs): write the SONAMEs in this package to the provs table */
    if (sqlite3_exec(bom, BOM_PROVS_SCHEMA, NULL, 0, &err) != SQLITE_OK) {
        FAIL_SQLITE(bom, err, -1);
    }

    if (sqlite3_prepare(bom, BOM_PROVS_INSERT, -1, &stmt, 0) != SQLITE_OK) {
        FAIL_SQLITE(bom, NULL, -1);
    }

    if (sqlite3_exec(bom, "BEGIN TRANSACTION", NULL, NULL, &err) != SQLITE_OK) {
        FAIL_SQLITE(bom, err, -1);
    }

    if (writeprovs(package, stmt)) {
        FAIL_COMMON(-1);
    }

    if (sqlite3_exec(bom, "END TRANSACTION", NULL, NULL, &err) != SQLITE_OK) {
        FAIL_SQLITE(bom, err, -1);
    }

    if (sqlite3_finalize(stmt) != SQLITE_OK) {
        FAIL_SQLITE(bom, NULL, -1);
    }

    /* the package can provide optional info value */
    /* any missing values get recorded as NULL     */
    if (package->meta != NULL) {
        xasprintf(&t, "%s/pkginfo", package->meta);

        if ((filename = realpath(t, filename)) == NULL) {
            FAIL_COMMON(-1);
        }

        if (stat(filename, &sb) == -1) {
            FAIL_COMMON(-1);
        }

        if (S_ISREG(sb.st_mode)) {
            if ((fd = fopen(filename, "r")) == NULL) {
                fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
                return -1;
            }

            while (getline(&line, &len, fd) != -1) {
                if (!strncmp(line, "desc=", 5)) {
                    if (package->desc != NULL) {
                        free(package->desc);
                        package->desc = NULL;
                    }

                    len = strcspn(line + 5, "\r\n");

                    if ((v = calloc(len + 1, sizeof(*v))) == NULL) {
                        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
                        abort();
                    }

                    if (strncpy(v, line + 5, len) == NULL) {
                        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
                        abort();
                    }

                    if ((package->desc = unquote(v)) == NULL) {
                        return -1;
                    }
                } else if (!strncmp(line, "maintainer=", 11)) {
                    if (package->maintainer != NULL) {
                        free(package->maintainer);
                        package->maintainer = NULL;
                    }

                    len = strcspn(line + 11, "\r\n");

                    if ((v = calloc(len + 1, sizeof(*v))) == NULL) {
                        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
                        abort();
                    }

                    if (strncpy(v, line + 11, len) == NULL) {
                        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
                        abort();
                    }

                    if ((package->maintainer = unquote(v)) == NULL) {
                        return -1;
                    }
                } else if (!strncmp(line, "license=", 8)) {
                    if (package->license != NULL) {
                        free(package->license);
                        package->license = NULL;
                    }

                    len = strcspn(line + 8, "\r\n");

                    if ((v = calloc(len + 1, sizeof(*v))) == NULL) {
                        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
                        abort();
                    }

                    if (strncpy(v, line + 8, len) == NULL) {
                        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
                        abort();
                    }

                    if ((package->license = unquote(v)) == NULL) {
                        return -1;
                    }
                } else if (!strncmp(line, "depspath=", 9)) {
                    if (package->depspath != NULL) {
                        free(package->depspath);
                        package->depspath = NULL;
                    }

                    len = strcspn(line + 9, "\r\n");

                    if ((v = calloc(len + 1, sizeof(*v))) == NULL) {
                        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
                        abort();
                    }

                    if (strncpy(v, line + 9, len) == NULL) {
                        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
                        abort();
                    }

                    if ((package->depspath = unquote(v)) == NULL) {
                        return -1;
                    }
                }

                free(line);
                line = NULL;
            }

            fclose(fd);
        }

        free(filename);
        free(t);
    }

    /* add the size of the package root tree to the pkginfo file */
    if ((package->size = get_tree_size(package->tree)) == -1) {
        fclose(fd);
        FAIL_COMMON(-1);
    }

    /* BOM(info): create the bill of materials info table */
    if (sqlite3_exec(bom, BOM_INFO_SCHEMA, NULL, 0, &err) != SQLITE_OK) {
        FAIL_SQLITE(bom, err, -1);
    }

    if (insert_pkg_info(package, bom, BOM_INFO_INSERT, 0, 0) < 0) {
        FAIL_SQLITE(bom, NULL, -1);
    }

    /* close out the bill of materials */
    sqlite3_close(bom);

    /* copy in any meta information that's present */
    cpmeta(package, DIST_LEGAL, true);

    if (!package->debug) {
        cpmeta(package, DIST_PREINSTALL, false);
        cpmeta(package, DIST_POSTINSTALL, false);
        cpmeta(package, DIST_PREREMOVE, false);
        cpmeta(package, DIST_POSTREMOVE, false);
    }

    return 0;
}

/* MAIN */
int main(int argc, char **argv) {
    int i = 0, long_idx = 0;
    int mode = S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH;
    struct dist package;
    char *outdir = NULL;
    char *tmpdir = NULL;
    struct stat sb;

    prog = basename(argv[0]);
    targetarch = getdefaultarch();

    /* zero out the package structure */
    initpkg(&package);

    /* parse options */
    while ((i = getopt_long(argc, argv, "dt:m:p:v:r:l:R:a:o:z:s:cVh?",
                            long_options, &long_idx)) != -1) {
        switch (i) {
            case 'd':
                package.debug = true;
                break;
            case 't':
                package.tree = strdup(optarg);
                break;
            case 'm':
                package.meta = strdup(optarg);
                break;
            case 'p':
                package.name = strdup(optarg);
                break;
            case 'v':
                package.ver = strdup(optarg);
                break;
            case 'r':
                package.rel = strtol(optarg, NULL, 10);

                if ((package.rel == LONG_MIN || package.rel == LONG_MAX) && errno == ERANGE) {
                    FAIL_COMMON(-1);
                }

                break;
            case 'l':
                if (stat(optarg, &sb) == -1) {
                    fprintf(stderr, "%s cannot be read or does not exist.\n", optarg);
                    return EXIT_FAILURE;
                }

                if (!S_ISREG(sb.st_mode)) {
                    fprintf(stderr, "%s is not a regular file.\n", optarg);
                    return EXIT_FAILURE;
                }

                package.plist = strdup(optarg);
                break;
            case 'R':
                package.repo = strdup(optarg);
                break;
            case 'a':
                /* check target arch and make sure it isn't bogus */
                if (check_targetarch(optarg)) {
                    fprintf(stderr, "Unsupported target architecture: %s\n",
                            optarg);
                    return EXIT_FAILURE;
                }

                free(targetarch);
                targetarch = strdup(optarg);
                break;
            case 'o':
                if (stat(optarg, &sb) == -1) {
                    if (errno == ENOENT) {
                        fprintf(stderr, "%s does not exist.\n", optarg);
                    } else {
                        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
                    }

                    return EXIT_FAILURE;
                }

                if (!S_ISDIR(sb.st_mode)) {
                    fprintf(stderr, "%s is not a directory.\n", optarg);
                    return EXIT_FAILURE;
                }

                outdir = strdup(optarg);
                break;
            case 'z':
                if (!strcasecmp(optarg, "lzma")) {
                    package.compression = ARCHIVE_COMPRESSION_LZMA;
                } else if (!strcasecmp(optarg, "bzip2") ||
                           !strcasecmp(optarg, "bz2")) {
                    package.compression = ARCHIVE_COMPRESSION_BZIP2;
                } else if (!strcasecmp(optarg, "compress") ||
                           !strcasecmp(optarg, "Z")) {
                    package.compression = ARCHIVE_COMPRESSION_COMPRESS;
                } else if (!strcasecmp(optarg, "gzip") ||
                           !strcasecmp(optarg, "gz")) {
                    package.compression = ARCHIVE_COMPRESSION_GZIP;
                } else if (!strcasecmp(optarg, "xz")) {
                    package.compression = ARCHIVE_COMPRESSION_XZ;
                } else if (!strcasecmp(optarg, "none")) {
                    package.compression = ARCHIVE_COMPRESSION_NONE;
                } else {
                    errno = EINVAL;
                    perror("unsupported compression type");
                    exit(errno);
                }

                break;
            case 's':
                if (!strcasecmp(optarg, "md5")) {
                    package.sumtype = MD5SUM;
                } else if (!strcasecmp(optarg, "sha1")) {
                    package.sumtype = SHA1SUM;
                } else if (!strcasecmp(optarg, "sha256")) {
                    package.sumtype = SHA256SUM;
                } else {
                    fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__,
                            strerror(errno));
                    return EXIT_FAILURE;
                }

                break;
            case 'c':
                package.do_cleanup = true;
                break;
            case 'V':
                ver(prog);
                return EXIT_SUCCESS;
            case 'h':
            case '?':
                usage();
                return EXIT_SUCCESS;
            case 0:
                /* long options with no short equivalent */
                if (!strcmp(long_options[long_idx].name, "tmpdir")) {
                    /* if this directory already exists, do not remove it */
                    assert(optarg != NULL);

                    if (stat(optarg, &sb) == -1) {
                        if (errno == ENOENT) {
                            fprintf(stderr, "%s does not exist.\n", optarg);
                        } else {
                            fprintf(stderr, "%s (%d): %s: %s\n", __func__, __LINE__,
                                    strerror(errno), optarg);
                        }

                        return EXIT_FAILURE;
                    }

                    if (S_ISDIR(sb.st_mode)) {
                        opts |= OPTS_SET_TMPDIR;
                    }

                    tmpdir = optarg;
                }

                break;
            default:
                usage();
                return EXIT_FAILURE;
        }
    }

    /* verify we have required options */
    if (checknull(package.tree, REASON_TREE) ||
        checknull(package.name, REASON_PKGNAME) ||
        checknull(package.ver, REASON_PKGVERSION) ||
        checkneg(package.rel, REASON_PKGRELEASE) ||
        checknull(package.plist, REASON_PLIST)) {
        return EXIT_FAILURE;
    }

    /* the metadata directory must be different than the tmp directory */
    if (!strcmp(_PATH_VARTMP, package.meta)) {
        fprintf(stderr, "*** ERROR: meta data directory and tmp directory are the same\n");
        return EXIT_FAILURE;
    }

    /* find a suitable tmp dir if one was not specified */
    if (tmppath == NULL) {
        xasprintf(&tmppath, "%s/%s.%s-%s-%d.%s.XXXXXX",
                  (tmpdir == NULL) ? _PATH_VARTMP : tmpdir, prog,
                  package.name, package.ver, package.rel, targetarch);

        if ((tmppath = mkdtemp(tmppath)) == NULL) {
            fprintf(stderr, "Unable to find a suitable temporary directory, "
                    "please use the --tmpdir option: %s.\n", strerror(errno));
            return EXIT_FAILURE;
        }
    }

    /* make the tmp path if it doesn't exist */
    if (mkdirp(tmppath, mode)) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__,
                strerror(errno));
        return EXIT_FAILURE;
    }

    /* staging directory */
    xasprintf(&staging, "%s/%s", tmppath, STAGING_SUBDIR);

    /* generate some filename strings for later */
    if (genstrings(&package, targetarch, outdir)) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__,
                strerror(errno));
        rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
        return EXIT_FAILURE;
    }

    /* remove package if it exists (just in case, our add_archive_member()
     * function doesn't handle replacement. */
    if (!stat(package.distname, &sb) && S_ISREG(sb.st_mode)) {
        if (remove(package.distname) == -1) {
            perror("remove");
            rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
            return EXIT_FAILURE;
        }
    }

    /* create the package staging directory */
    if (mkstagedir(&package)) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__,
                strerror(errno));
        rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
        return EXIT_FAILURE;
    }

    printf("Creating payloads...");

    if (verbose) {
        printf("\n");
    }

    plist = read_plist(package.plist, package.tree);

    if (add_inodes(package.tree)) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__,
                strerror(errno));
        rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
        return EXIT_FAILURE;
    }

    /* pack up /etc, if it exists */
    if (create_etc(&package)) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__,
                strerror(errno));
        rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
        return EXIT_FAILURE;
    }

    /* copy & preserve -t to the staging reloc tree */
    if (create_reloc(&package) == -1) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__,
                strerror(errno));
        rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
        return EXIT_FAILURE;
    }

    if (plist_checked != plist_length(plist)) {
        fprintf(stderr, "*** WARNING: plist and package root tree mismatch (%s-%s-%d)\n", package.name, package.ver, package.rel);
    }

    if (!verbose) {
        printf("DONE.\n");
    }

    /* add in package meta data */
    printf("Adding meta data...");

    if (dometa(&package)) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
        rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
        return EXIT_FAILURE;
    }

    printf("DONE.\n");

    /* pack everything up */
    printf("Writing %s...", package.distname);

    if (create_dist(&package)) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
        rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
        return EXIT_FAILURE;
    }

    printf("DONE.\n");

    /* cleanup */
    if (cleanup(&package)) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__,
                strerror(errno));
        rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
        return EXIT_FAILURE;
    }

    rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
    free_inodes();
    freepkg(&package);
    free(outdir);
    free(tmppath);
    free_plist(plist);

    return EXIT_SUCCESS;
}
