/* Make a new dist plist file. */

/* {{{ Apache License version 2.0
 */
/*
 * Copyright 2011-2018 David Cantrell <david.l.cantrell@gmail.com>
 *                     Chris Lumens <chris@bangmoney.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/* }}} */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>
#include <ftw.h>
#include <pwd.h>
#include <grp.h>
#include <limits.h>
#include "dist-tools.h"

extern char *optarg;
extern int optind, opterr, optopt;

static char *prog = NULL;
static int trimlen = -1;
static char *user = NULL;
static char *group = NULL;
static uid_t uid = 0;
static gid_t gid = 0;
static strlist_t *plist = NULL;

static struct option long_options[] = {
    {"tree", required_argument, NULL, 't'},
    {"user", required_argument, NULL, 'u'},
    {"group", required_argument, NULL, 'g'},
    {"version", no_argument, NULL, 'V'},
    {"help", no_argument, NULL, 'h'},
    {NULL, 0, NULL, 0}
};

#define REASON_TREE "Missing package root tree (-t)"

/*
 * Usage screen.
 */
static void usage(void) {
    printf("Usage: %s [OPTION]...\n", prog);
    puts("Make a Peachtree Linux dist plist file (used for distmk runs)\n"
         "\nRequired:\n"
         "  -t PATH, --tree=PATH       Source path to package root tree\n"
         "\nOptional:\n"
         "  -u USER, --user=USER       Username for the default owner of"
         " the files\n"
         "  -g GROUP, --group=GROUP    Group name for the default group"
         " owner of the\n"
         "                             files\n"
         "  -V, --version              Display version information\n"
         "  -h, --help                 This help message");
    return;
}

/* Add a line for the plist */
static int _add_plist_entry(const char *fpath, const struct stat *sb,
                            int tflag, struct FTW *ftwbuf) {
    int len = -1;
    const char *vpath = NULL;
    char *symlink = NULL;
    char *filename = NULL;

    if (trimlen == -1) {
        trimlen = strlen(fpath);
        return 0;
    } else {
        vpath = fpath + trimlen;

        while (*vpath == '/') {
            vpath++;
        }
    }

    if (S_ISLNK(sb->st_mode)) {
        if ((symlink = calloc(1, PATH_MAX)) == NULL) {
            fprintf(stderr, "unable to allocate memory in %s (%d)\n",
                    __func__, __LINE__);
            abort();
        }

        len = readlink(fpath, symlink, PATH_MAX-1);

        if (len == -1) {
            fprintf(stderr, "unable to read symlink %s in %s (%d)\n",
                    fpath, __func__, __LINE__);
            abort();
        }

        symlink[len] = '\0';
        xasprintf(&filename, "%s %s(%d) %s(%d) symlink(%s)",
                  vpath, user, uid, group, gid, symlink);
        free(symlink);
    } else {
        xasprintf(&filename, "%s %s(%d) %s(%d) %04o",
                  vpath, user, uid, group, gid, sb->st_mode & S_INTER);
    }

    plist = sorted_strlist_insert(plist, filename);
    return 0;
}

/* Get the struct passwd entry for the specified user. */
static int getuser(void) {
    struct passwd *pwd = NULL;

    if (user == NULL) {
        return -1;
    }

    errno = 0;
    if ((pwd = getpwnam(user)) == NULL) {
        perror("getpwnam");
        return -1;
    }

    uid = pwd->pw_uid;
    gid = pwd->pw_gid;
    return 0;
}

/* Get the struct group entry for the specified group. */
static int getgroup(void) {
    struct group *grp = NULL;

    if (group == NULL) {
        if ((grp = getgrgid(gid)) == NULL) {
            perror("getgrgid");
            return -1;
        }

        group = strdup(grp->gr_name);
    } else {
        if ((grp = getgrnam(group)) == NULL) {
            perror("getgrnam");
            return -1;
        }

        gid = grp->gr_gid;
    }

    return 0;
}

/* Free memory */
static void cleanup(void) {
    free(user);
    free(group);
    free_strlist(plist);
    return;
}

/* MAIN */
int main(int argc, char **argv) {
    int i;
    int flags = FTW_MOUNT | FTW_PHYS;
    char *tree = NULL;
    struct passwd *pw = NULL;
    strlist_entry_t *entry = NULL;

    prog = basename(argv[0]);
    pw = getpwuid(getuid());
    user = strdup(pw->pw_name);

    /* parse options */
    while ((i = getopt_long(argc, argv, "t:u:g:Vh?",
                            long_options, NULL)) != -1) {
        switch (i) {
            case 't':
                tree = strdup(optarg);
                break;
            case 'u':
                free(user);
                user = strdup(optarg);
                break;
            case 'g':
                group = strdup(optarg);
                break;
            case 'V':
                ver(prog);
                return EXIT_SUCCESS;
            case 'h':
            case '?':
            default:
                usage();
                return EXIT_FAILURE;
        }
    }

    /* verify we have required options */
    if (checknull(tree, REASON_TREE)) {
        cleanup();
        return EXIT_FAILURE;
    }

    /* get pwd and grp structures */
    if (getuser() == -1) {
        free(tree);
        cleanup();
        perror("getuser");
        return EXIT_FAILURE;
    }

    if (getgroup() == -1) {
        free(tree);
        cleanup();
        perror("getgroup");
        return EXIT_FAILURE;
    }

    /* add the plist contents */
    plist = init_strlist();

    if (nftw(tree, _add_plist_entry, 25, flags) == -1) {
        free(tree);
        cleanup();
        perror("nftw");
        return EXIT_FAILURE;
    }

    /* sort and print the contents */
    TAILQ_FOREACH(entry, plist, items) {
        printf("%s\n", entry->val);
    }

    free(tree);
    cleanup();
    return EXIT_SUCCESS;
}
