/* Show if an /etc file is owned by a package or not. */

/* {{{ Apache License version 2.0
 */
/*
 * Copyright 2016-2018 David Cantrell <david.l.cantrell@gmail.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/* }}} */

#include <stdio.h>
#include <stdlib.h>
#include <libgen.h>
#include <getopt.h>
#include <errno.h>
#include <string.h>
#include <glob.h>
#include <ftw.h>
#include "dist-tools.h"
#include "db.h"

extern char *optarg;
extern int optind, opterr, optopt;

/* Local types */
typedef struct _owned_entry_t {
    char *path;
    strlist_t *hashlist;
    TAILQ_ENTRY(_owned_entry_t) items;
} owned_entry_t;

typedef TAILQ_HEAD(ownedlist_s, _owned_entry_t) ownedlist_t;

/* Local prototypes */
static void usage(void);
static int get_owned(void);
static int report_unowned(void);
static int _add_owned_entry(const char *, const struct stat *, int, struct FTW *);
static int _report_unowned_entry(const char *, const struct stat *, int, struct FTW *);
static void _print_owner(char *, char *);

/* Local global variables */
static char *prog = NULL;
static char *db_path = NULL;
static pkglist_t *pkglist = NULL;
static char *pkghash = NULL;
static int trimlen = -1;
static char *prepend = NULL;
static ownedlist_t *owned = NULL;

static struct option long_options[] = {
    {"root", required_argument, NULL, 'r'},
    {"version", no_argument, NULL, 'V'},
    {"help", no_argument, NULL, 'h'},
    {NULL, 0, NULL, 0}
};

/* Called by get_owned() below. */
static int _add_owned_entry(const char *fpath, const struct stat *sb,
                            int tflag, struct FTW *ftwbuf) {
    char *path = NULL;
    bool have = false;
    owned_entry_t *ownedentry = NULL;

    /*
     * At level 1, we will see the package subdirectory in db_path.
     * In db_path, these are the package checksum digest.  We can
     * use this value to look up the package name in db_installed.
     */
    if (ftwbuf->level == 1) {
        trimlen = strlen(fpath);
        pkghash = strdup(((char *) fpath) + ftwbuf->base);
    }

    /*
     * Just another safety catch to make sure we are only collecting
     * information about packages with configuration files.
     */
    if (ftwbuf->level == 2 && strcmp(fpath+ftwbuf->base, CONFDIR_PREFIX)) {
        return 0;
    }

    /*
     * Everything at level 2 and below is possibly a configuration file path.
     */
    if (ftwbuf->level >= 2) {
        path = ((char *) fpath) + trimlen;

        TAILQ_FOREACH(ownedentry, owned, items) {
            if (!strcmp(ownedentry->path, path)) {
                have = true;
                break;
            }
        }

        if (have) {
            ownedentry->hashlist = sorted_strlist_insert(ownedentry->hashlist, pkghash);
        } else {
            if ((ownedentry = calloc(1, sizeof(*ownedentry))) == NULL) {
                fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
                return -1;
            }

            ownedentry->path = strdup(path);
            ownedentry->hashlist = init_strlist();
            ownedentry->hashlist = sorted_strlist_insert(ownedentry->hashlist, pkghash);
            TAILQ_INSERT_TAIL(owned, ownedentry, items);
        }
    }

    return 0;
}

/* Called by report_unowned() below. */
static int _report_unowned_entry(const char *fpath, const struct stat *sb,
                                 int tflag, struct FTW *ftwbuf) {
    bool found = false;
    owned_entry_t *entry = NULL;

    if (prepend != NULL && strlen(prepend) > 1) {
        fpath += strlen(prepend);
    }

    /* anything not in the owned hash would be unowned */
    TAILQ_FOREACH(entry, owned, items) {
        if (!strcmp(entry->path, fpath)) {
            found = true;
            break;
        }
    }

    if (!found) {
        printf("%s\n", fpath);
    }

    return 0;
}

/* Prints all of the packages that own a particular configuration file. */
static void _print_owner(char *pkghash, char *etcfile) {
    struct dist *pkg = NULL;

    pkg = find_pkg_by_hash(pkglist, pkghash);

    if (pkg != NULL) {
        printf("%s: %s-%s-%d%s.%s\n", etcfile, pkg->name, pkg->ver, pkg->rel,
                                      (pkg->debug) ? ".debug" : "", pkg->arch);
    }

    return;
}

/* Usage screen displayed when the user passes --help as an argument. */
static void usage(void) {
    printf("Usage:\n");
    printf("  %s [option]... [filespec...]\n", prog);
    puts("\n"
         "Options:\n"
         "  -r PATH, --root=PATH       Use PATH for the root (/) directory\n"
         "  -V, --version              Show distinfo version\n"
         "  -h, --help                 This help message\n");
    return;
}

/* Collect all the etc files owned by the package database. */
static int get_owned(void) {
    int flags = FTW_MOUNT | FTW_PHYS;

    if ((owned = calloc(1, sizeof(*owned))) == NULL) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
        abort();
    }

    TAILQ_INIT(owned);

    if (nftw(db_path, _add_owned_entry, 25, flags) == -1) {
        fprintf(stderr, "%s: %s line %d: %s\n", __FILE__, __func__, __LINE__, strerror(errno));
        return -1;
    }

    return 0;
}

/* Report etc paths that are not owned by an installed package. */
static int report_unowned(void) {
    int flags = FTW_MOUNT | FTW_PHYS;
    char *etcpath = NULL;

    etcpath = get_path(prepend, CONFDIR_PREFIX);

    if (nftw(etcpath, _report_unowned_entry, 25, flags) == -1) {
        fprintf(stderr, "%s: %s line %d: %s\n", __FILE__, __func__, __LINE__, strerror(errno));
        return -1;
    }

    free(etcpath);
    return 0;
}

/* MAIN */
int main(int argc, char **argv) {
    int i = 0, j = 0;
    int long_idx = 0;
    char *db_installed = NULL;
    char *etcfile = NULL;
    int success = EXIT_SUCCESS;
    struct stat sb;
    bool have = false;
    owned_entry_t *ownedentry = NULL;
    strlist_entry_t *hashentry = NULL;
    glob_t found;

    prog = basename(argv[0]);

    /* parse options */
    while ((i = getopt_long(argc, argv, "r:Vh?",
                            long_options, &long_idx)) != EOF) {
        switch (i) {
            case 'r':
                if (*optarg != '/') {
                    printf("Root directory must be an absolute pathname\n");
                    return EXIT_FAILURE;
                }

                prepend = strdup(optarg);
                break;
            case 'V':
                ver(prog);
                return EXIT_SUCCESS;
            case 'h':
            case '?':
                usage();
                return EXIT_SUCCESS;
            default:
                usage();
                return EXIT_FAILURE;
        }
    }

    /* use / if the user did not specify a root path */
    if (prepend == NULL) {
        prepend = strdup("/");
    }

    /* handle an alternative root path for installation */
    db_path = get_path(prepend, DB_PATH);
    db_installed = get_path(prepend, DB_FILE);

    /* collect information about installed packages */
    if (stat(db_installed, &sb) == -1 || !S_ISREG(sb.st_mode)) {
        fprintf(stderr, "This system lacks a Peachtree Linux package database.\n");
        return EXIT_FAILURE;
    }

    pkglist = scanpkg(db_installed, NULL);

    if (pkglist == NULL) {
        return EXIT_FAILURE;
    }

    /* collect pkg database information on etc ownership */
    if (get_owned()) {
        fprintf(stderr, "%s: %s line %d: %s\n", __FILE__, __func__, __LINE__, strerror(errno));
        return EXIT_FAILURE;
    }

    /* do stuff */
    if (argc == optind) {
        /* no filespecs given, report on entire system */
        if (report_unowned()) {
            fprintf(stderr, "%s: %s line %d: %s\n", __FILE__, __func__, __LINE__, strerror(errno));
            return EXIT_FAILURE;
        }
    } else {
        /* iterate over filespecs given */
        for (i = optind; i < argc; i++) {
            /* handle any potential wildcards */
            j = glob(argv[i], GLOB_BRACE|GLOB_TILDE, NULL, &found);

            if (j == 0) {
                /* matches found */
                while (j < found.gl_pathc) {
                    etcfile = found.gl_pathv[j];
                    have = false;

                    TAILQ_FOREACH(ownedentry, owned, items) {
                        if (!strcmp(ownedentry->path, etcfile)) {
                            have = true;
                            break;
                        }
                    }

                    if (have) {
                        TAILQ_FOREACH(hashentry, ownedentry->hashlist, items) {
                            _print_owner(hashentry->val, etcfile);
                        }
                    } else {
                        printf("%s: not owned\n", etcfile);
                    }

                    j++;
                }
            } else if (j == GLOB_NOMATCH) {
                fprintf(stderr, "*** %s not found on system\n", argv[i]);
                continue;
            } else {
                fprintf(stderr, "*** ERROR: failed expanding %s: %s\n", argv[i], strerror(errno));
                continue;
            }

            globfree(&found);
        }
    }

    free_pkglist(pkglist);
    free(db_path);
    free(db_installed);
    free(prepend);

    return success;
}
