/* struct finfo helper functions */

/* {{{ */
/*
 * Copyright 2014-2018 David Cantrell <david.l.cantrell@gmail.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/* }}} */

#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <endian.h>
#include "dist-tools.h"

/*
 * Build a struct finfo from the current row in an sqlite3_stmt.
 * If 'id' is <0, read the id from the sqlite3 row.
 * If 'id' is >=0, use this value for the package id.
 */
int read_finfo(int id, enum signature sigver, sqlite3_stmt *stmt,
               struct finfo *fi, char *prepend) {
    mode_t mode = 0;

    assert(stmt != NULL);
    assert(fi != NULL);

    /* finfo needs a struct stat, so build one sort of */
    if ((fi->attr = calloc(1, sizeof(struct stat))) == NULL) {
        fprintf(stderr, "%s: %d: failed allocating struct stat\n", __FILE__, __LINE__);
        return -1;
    }

    if (id < 0) {
        fi->packageid = sqlite3_column_int(stmt, 0);
    } else {
        fi->packageid = id;
    }

    fi->filename = strdup((char *) sqlite3_column_text(stmt, 1));
    fi->owner = strdup((char *) sqlite3_column_text(stmt, 2));
    fi->attr->st_uid = (uid_t) sqlite3_column_int(stmt, 3);
    fi->group = strdup((char *) sqlite3_column_text(stmt, 4));
    fi->attr->st_gid = (gid_t) sqlite3_column_int(stmt, 5);

    memset(&(fi->attr->st_mode), 0, sizeof(fi->attr->st_mode));

    if (sigver == SIGVER_21) {
        /*
         * This is a known bug in v2.1 packages.  If you build a noarch
         * package, the host you install it on must match the endianness
         * of the build host.  That's why for v2.2 packages, we store the
         * mode in big endian format in the packages.
         */
        memcpy(&(fi->attr->st_mode), sqlite3_column_blob(stmt, 6), sqlite3_column_bytes(stmt, 6));
    } else if (sigver >= SIGVER_22) {
        memcpy(&mode, sqlite3_column_blob(stmt, 6), sqlite3_column_bytes(stmt, 6));
        mode = be32toh(mode);
        memcpy(&(fi->attr->st_mode), &mode, sizeof(mode));
    } else {
        fprintf(stderr, "%s (%d): unknown package signature %d\n", __func__, __LINE__, sigver);
        abort();
    }

    fi->attr->st_size = (off_t) sqlite3_column_int(stmt, 7);
    fi->attr->st_mtime = sqlite3_column_int(stmt, 8);
    fi->type = sqlite3_column_int(stmt, 9);

    if (fi->type == TYPE_HARD || fi->type == TYPE_SOFT) {
        fi->linkname = strdup((char *) sqlite3_column_text(stmt, 10));
    } else {
        fi->linkname = NULL;
    }

    if (fi->type == TYPE_REG) {
        fi->checksum = strdup((char *) sqlite3_column_text(stmt, 11));
    } else {
        fi->checksum = NULL;
    }

    fi->fullpath = get_path(prepend, fi->filename);

    return 0;
}

/*
 * Helper to make sure all of a struct finfo is properly freed.
 * Useful as a callback for glib data structures.
 */
void free_finfo(struct finfo *info) {
    if (info == NULL) {
        return;
    }

    free(info->filename);
    free(info->fullpath);
    free(info->checksum);
    free(info->attr);
    free(info->linkname);
    free(info->owner);
    free(info->group);

    info->filename = NULL;
    info->fullpath = NULL;
    info->checksum = NULL;
    info->attr = NULL;
    info->linkname = NULL;
    info->owner = NULL;
    info->group = NULL;

    info = NULL;

    return;
}

/*
 * Deep copy of a struct finfo.  Caller must free returned struct.
 */
struct finfo *dup_finfo(struct finfo *info) {
    struct finfo *ret = NULL;

    assert(info != NULL);

    if ((ret = calloc(1, sizeof(struct finfo))) == NULL) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
        abort();
    }

    ret->packageid = info->packageid;
    ret->type = info->type;
    ret->has_checksum = info->has_checksum;

    if (info->filename != NULL) {
        ret->filename = strdup(info->filename);
    }

    if (info->fullpath != NULL) {
        ret->fullpath = strdup(info->fullpath);
    }

    if (info->checksum != NULL) {
        ret->checksum = strdup(info->checksum);
    }

    if (info->owner != NULL) {
        ret->owner = strdup(info->owner);
    }

    if (info->group != NULL) {
        ret->group = strdup(info->group);
    }

    if (info->linkname != NULL) {
        ret->linkname = strdup(info->linkname);
    }

    if ((ret->attr = calloc(1, sizeof(struct stat))) == NULL) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
        abort();
    }

    memcpy(ret->attr, info->attr, sizeof(struct stat));

    return ret;
}
