/*
 * Walk a directory tree and read attributes and generate checksums.
 * Also generate package dependencies.
 */

/* {{{ Apache License version 2.0
 */
/*
 * Copyright 2003-2018 David Cantrell <david.l.cantrell@gmail.com>
 *                     David Shea <david@reallylongword.org>
 *                     Chris Lumens <chris@bangmoney.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/* }}} */

#include <stdlib.h>
#include <stdbool.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <ftw.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <libelf.h>
#include <gelf.h>
#include "dist-tools.h"
#include "plist.h"
#include "strlist.h"
#include "signature.h"

extern char **environ;

/* We use this struct to pass in more info to the nftw() callback. */
struct attrsinfo {
    sqlite3_stmt *stmt;
    plist_t *plist;
    enum signature sigver;
    enum checksum sumtype;
    bool debug;
};

static strlist_t *deplist = NULL;
static strlist_t *provlist = NULL;
static strlist_t *manifest = NULL;
static strlist_t *symlinks = NULL;
static int trimlen = -1;
static struct attrsinfo ai;

/* Local prototypes */
static void insert_name(const char *, sqlite3_stmt *);
static bool elf_reader(const char *, const int, strlist_t **);
static int _add_attrs_entry(const char *, const struct stat *, int, struct FTW *);

/* Insert an SONAME element in to an SQLite table */
static void insert_name(const char *soname, sqlite3_stmt *stmt) {
    if (sqlite3_bind_text(stmt, 1, soname, strlen(soname), SQLITE_TRANSIENT) != SQLITE_OK) {
        fprintf(stderr, "%s (%d): failure binding SONAME '%s'\n", __func__, __LINE__, soname);
        abort();
    }

    if (sqlite3_step(stmt) != SQLITE_DONE) {
        fprintf(stderr, "%s (%d): failure inserting SONAME '%s'\n", __func__, __LINE__, soname);
        abort();
    }

    if (sqlite3_reset(stmt) != SQLITE_OK) {
        fprintf(stderr, "%s (%d): failure resetting statement\n", __func__, __LINE__);
        abort();
    }

    if (sqlite3_clear_bindings(stmt) != SQLITE_OK) {
        fprintf(stderr, "%s (%d): failure clearing bindings\n", __func__, __LINE__);
        abort();
    }

    return;
}

/*
 * ELF object reading support function.  This function should
 * be called when the current working directory is the location
 * of 'filename'.  This function does not change directories.
 */
static bool elf_reader(const char *filename, const int type, strlist_t **list) {
    int fd = -1;
    Elf *elf = NULL;
    GElf_Ehdr ehdr;
    Elf_Scn *scn = NULL;
    strlist_entry_t *entry = NULL;

    if (filename == NULL) {
        return false;
    }

    /* skip non-ELF files */
    if (!is_elf(filename)) {
        return true;
    }

    elf_version(EV_CURRENT);

    if ((fd = open(filename, O_RDONLY)) == -1) {
        fprintf(stderr, "%s (%d): %s (%s)\n", __func__, __LINE__, strerror(errno), filename);
        return false;
    }

    elf = elf_begin(fd, ELF_C_READ_MMAP, NULL);

    if (gelf_getehdr(elf, &ehdr) == NULL) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
        return false;
    }

    /* only collect DT_NEEDED from ELF dynamic shared objects */
    if (ehdr.e_type != ET_EXEC && ehdr.e_type != ET_DYN) {
        if (elf_end(elf) != 0) {
            fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
            abort();
        }

        if (close(fd) == -1) {
            fprintf(stderr, "%s (%d): %s (%s)\n", __func__, __LINE__, strerror(errno), filename);
            return false;
        }

        return true;
    }

    scn = NULL;

    while ((scn = elf_nextscn(elf, scn)) != NULL) {
        char *name = NULL;
        int ndx, maxndx;
        GElf_Shdr shdr;
        Elf_Data *data = NULL;
        GElf_Dyn dyn;

        if (gelf_getshdr(scn, &shdr) == NULL) {
            continue;
        }

        if (shdr.sh_type != SHT_DYNAMIC) {
            continue;
        }

        if ((data = elf_getdata(scn, data)) == NULL || data->d_size == 0) {
            /* no data to read, go to next section */
            continue;
        }

        maxndx = data->d_size / shdr.sh_entsize;

        for (ndx = 0; ndx < maxndx; ++ndx) {
            if (gelf_getdyn(data, ndx, &dyn) == NULL) {
                continue;
            }

            if (dyn.d_tag == type) {
                /*
                 * This took a while to figure out since the elfutils
                 * documentation doesn't really explain anything.  The
                 * important thing to know is that all of the sections
                 * in an ELF file just contain offsets and addresses to
                 * useful data like strings such as "libc.so.6".
                 *
                 * Here we iterate over each section and get the GElf_Dyn
                 * object from the section.  If it's a DT_NEEDED dyn
                 * object, then it represents a shared library we need
                 * at runtime.  If it's DT_SONAME, it's the name of this
                 * library (there should really only be one SONAME per
                 * ELF object, but we won't assume that).
                 *
                 * To get the library name, we use elf_strptr() here to
                 * extract the string from the big string table.  The
                 * string table is pointed to by shdr.sh_link and the
                 * offset to the library name is given by dyn.d_un.d_ptr.
                 *
                 * And this is a standard?
                 *
                 * Also, we strdup() here because we need a copy of the
                 * string in the strlist_t.  It will be free()'ed later.
                 */
                name = elf_strptr(elf, shdr.sh_link, (size_t) (dyn.d_un.d_ptr));

                /* we need to initialize the list on first use */
                if (*list == NULL) {
                    *list = init_strlist();
                }

                if (!is_str_in_list(*list, name)) {
                    if ((entry = calloc(1, sizeof(*entry))) == NULL) {
                        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
                        abort();
                    }

                    entry->val = strdup(name);
                    TAILQ_INSERT_TAIL(*list, entry, items);
                }
            }
        }
    }

    if (elf_end(elf) != 0) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
        return false;
    }

    if (close(fd) == -1) {
        fprintf(stderr, "%s (%d): %s (%s)\n", __func__, __LINE__, strerror(errno), filename);
        return false;
    }

    return true;
}

/*
 * Adds an individual entry to the manifest and symlinks lists.
 */
static int _add_lists_entry(const char *fpath, const struct stat *sb,
                            int tflag, struct FTW *ftwbuf) {
    strlist_t *list = NULL;

    if (S_ISREG(sb->st_mode)) {
        list = manifest;
    } else if (S_ISLNK(sb->st_mode)) {
        list = symlinks;
    } else {
        return 0;
    }

    if (list == NULL) {
        list = init_strlist();
    }

    list = sorted_strlist_insert(list, fpath);
    return 0;
}

/*
 * Compute attribute values for the specified entry.
 * NOTE: Relies on the global 'ai' struct, initialized by genattrs().
 */
static int _add_attrs_entry(const char *fpath, const struct stat *sb, int tflag, struct FTW *ftwbuf) {
    int ret = 0;
    const char *vpath = NULL;
    struct finfo info;
    plist_entry_t *plistentry = NULL;
    plist_entry_t *iterator = NULL;
    strlist_entry_t *strentry = NULL;

    if (trimlen == -1) {
        trimlen = strlen(fpath);
        return 0;
    } else {
        vpath = fpath + trimlen;

        while (*vpath == '/') {
            vpath++;
        }
    }

    /* initialize finfo as other to avoid hard link checks */
    info.type = TYPE_OTHER;
    info.checksum = NULL;
    info.linkname = NULL;
    info.filename = strdup(vpath);
    info.fullpath = strdup(fpath);
    info.owner = NULL;
    info.group = NULL;
    info.has_checksum = true;

    if ((info.attr = calloc(1, sizeof(struct stat))) == NULL) {
        FAIL_COMMON(-1);
    }

    memcpy(info.attr, sb, sizeof(struct stat));

    /* grab the attributes */
    ret = gen_contents(&info, ai.sumtype);

    /* override attributes from stat(2) with plist values */
    TAILQ_FOREACH(iterator, ai.plist, items) {
        if (!strcmp(iterator->filename, vpath)) {
            plistentry = iterator;
            break;
        }
    }

    if (plistentry != NULL) {
        if (!plistentry->is_symlink) {
            info.attr->st_mode = plistentry->mode;
        }

        info.owner = strdup(plistentry->owner);
        info.attr->st_uid = plistentry->uid;
        info.group = strdup(plistentry->group);
        info.attr->st_gid = plistentry->gid;

        /* we could have a checksum already computed by gen_contents() */
        if (!plistentry->has_checksum) {
            free(info.checksum);
            info.checksum = strdup("*");
        }

        info.has_checksum = plistentry->has_checksum;
    }

    /* gather dependency information */
    if (info.type == TYPE_REG && !ai.debug) {
        if (!elf_reader(vpath, DT_NEEDED, &deplist)) {
            perror("Error checking file dependencies");
            ret = -1;
        }

        /* remove dependencies provided by this package */
        if (length_strlist(deplist) > 0) {
            TAILQ_FOREACH(strentry, deplist, items) {
                if (is_str_in_list(manifest, strentry->val) ||
                    is_str_in_list(symlinks, strentry->val)) {
                    TAILQ_REMOVE(deplist, strentry, items);
                    free(strentry->val);
                    strentry->val = NULL;
                    free(strentry);
                }
            }
        }

        /* collect SONAMEs provided by this package */
        if (!elf_reader(vpath, DT_SONAME, &provlist)) {
            perror("Error checking file provides");
            ret = -1;
        }
    }

    /* write out infos */
    if (write_one_entry(-1, &info, ai.stmt, ai.sigver)) {
        fprintf(stderr, "Unable to write info for %s: %s\n", vpath, strerror(errno));
        ret = -1;
    }

    free(info.checksum);
    free(info.linkname);
    free(info.filename);
    free(info.fullpath);
    free(info.attr);
    free(info.owner);
    free(info.group);

    return ret;
}

/*
 * Exposed call for programs that want to build the manifest and
 * symlinks lists for a package staging root.
 */
int makelists(const char *dir) {
    int flags = FTW_MOUNT | FTW_PHYS;

    assert(dir != NULL);

    if (nftw(dir, _add_lists_entry, 25, flags) == -1) {
        FAIL_COMMON(-1);
    }

    return 0;
}

/* Write the list of dependencies to the package */
int writedeps(struct dist *package, sqlite3_stmt *stmt) {
    strlist_entry_t *entry = NULL;

    assert(stmt != NULL);

    if (package->debug || length_strlist(deplist) == 0) {
        return 0;
    }

    if (deplist != NULL) {
        TAILQ_FOREACH(entry, deplist, items) {
            insert_name(entry->val, stmt);
        }
    }

    return 0;
}

/* Write the list of SONAMEs provided by this package */
int writeprovs(struct dist *package, sqlite3_stmt *stmt) {
    strlist_entry_t *entry = NULL;

    assert(stmt != NULL);

    if (package->debug || length_strlist(provlist) == 0) {
        return 0;
    }

    if (provlist != NULL) {
        TAILQ_FOREACH(entry, provlist, items) {
            insert_name(entry->val, stmt);
        }
    }

    return 0;
}

/*
 * Walk the tree and generate attributes for all entries.
 * Since we use nftw(), we take the input we are given and put it in
 * the global attrsinfo struct.
 */
int genattrs(char *tree, sqlite3_stmt *stmt, plist_t *plist, enum signature sigver, enum checksum sumtype, bool debug) {
    int flags = FTW_MOUNT | FTW_PHYS;

    assert(tree != NULL);

    ai.stmt = stmt;
    ai.plist = plist;
    ai.sigver = sigver;
    ai.sumtype = sumtype;
    ai.debug = debug;
    trimlen = -1;

    if (nftw(tree, _add_attrs_entry, 25, flags) == -1) {
        FAIL_COMMON(-1);
    }

    return 0;
}

/*
 * genattrs_free - Free up lists we have.
 */
void genattrs_free(void) {
    free_strlist(manifest);
    free_strlist(symlinks);
    return;
}
