/*
 * Iterates over a TAILQ and list pair describing a package, writes new
 * checksums, attrs, and manifest based on the actual state of the package
 *
 * The basic idea is to manage everything in a TAILQ of struct finfo structs.
 *
 * write_one_entry will write the bom.db contents table and/or installed.db
 * contents table entry for a single finfo struct.
 */

/* {{{ Apache License version 2.0
 */
/*
 * Copyright 2003-2018 David Shea <david@reallylongword.org>
 *                     David Cantrell <david.l.cantrell@gmail.com>
 *                     Chris Lumens <chris@bangmoney.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/* }}} */

#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <stdbool.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <endian.h>
#include <limits.h>
#include "dist-tools.h"
#include "db.h"

extern char *prepend;

/* No need for directory or hardlink lists
 * Directory order is maintained by keeping directories in a tree,
 *    sorted by length
 * Hard link order is maintained by fixdist mapping all hard links in a set
 *    to a single regular file, and that regular file being written first
 *    in the manifest
 * Any metadata file by setting one of the skip_* arguments to non-zero
 */
int write_contents(int rowid, enum signature sigver, sqlite3 *db, finfolist_t *list) {
    int errs = 0;
    char *err = NULL;
    finfolist_entry_t *entry = NULL;
    sqlite3_stmt *contents_stmt = NULL;

    assert(db != NULL);

    /* statement to insert in to the contents table */
    if (sqlite3_prepare(db, DB_CONTENTS_GENDIST, -1, &contents_stmt, 0) != SQLITE_OK) {
        FAIL_SQLITE(db, NULL, -1);
    }

    if (sqlite3_exec(db, "BEGIN TRANSACTION", NULL, NULL, &err) != SQLITE_OK) {
        FAIL_SQLITE(db, err, -1);
    }

    /* Iterate over each finfo and write a new row */
    if (list != NULL) {
        TAILQ_FOREACH(entry, list, items) {
            errs = write_one_entry(rowid, entry->info, contents_stmt, sigver);
        }
    }

    if (sqlite3_exec(db, "END TRANSACTION", NULL, NULL, &err) != SQLITE_OK) {
        FAIL_SQLITE(db, err, -1);
    }

    if (sqlite3_finalize(contents_stmt) != SQLITE_OK) {
        fprintf(stderr, "%s (%d): %s (%d)\n", __func__, __LINE__, sqlite3_errmsg(db), sqlite3_errcode(db));
        errs = -1;
    }

    return errs;
}

/* set any sqlite3_stmt pointer to NULL to skip */
int write_one_entry(const int id, struct finfo *info, sqlite3_stmt *dbstmt, enum signature sigver) {
    int offset = 0;
    struct passwd *pwd = NULL;
    struct group *grp = NULL;
    char *modestr = NULL;
    int success = 0;
    mode_t mode = 0;

    /* skip configuration files */
    if (!strncmp(info->filename, CONFDIR_PREFIX, strlen(CONFDIR_PREFIX))) {
        return success;
    }

    /* longest 32-bit UID is 10 chars */
    if ((pwd = getpwuid(info->attr->st_uid)) == NULL) {
        xasprintf(&(info->owner), "%u", (uid_t) (info->attr->st_uid));
    } else {
        info->owner = strdup(pwd->pw_name);
    }

    if ((grp = getgrgid(info->attr->st_gid)) == NULL) {
        xasprintf(&(info->group), "%u", (gid_t) (info->attr->st_gid));
    } else {
        info->group = strdup(grp->gr_name);
    }

    /* info table in bom.db or contents table in installed.db */
    if (dbstmt != NULL) {
        if (id >= 0) {
            if (sqlite3_bind_int(dbstmt, 1, id) != SQLITE_OK) {
                FAIL_SQLITE_STMT(dbstmt, -1);
            }

            offset = 1;
        }

        if (sqlite3_bind_text(dbstmt, offset + 1, info->filename, strlen(info->filename), SQLITE_TRANSIENT) != SQLITE_OK) {
            FAIL_SQLITE_STMT(dbstmt, -1);
        }

        if (sqlite3_bind_text(dbstmt, offset + 2, info->owner, strlen(info->owner), SQLITE_TRANSIENT) != SQLITE_OK) {
            FAIL_SQLITE_STMT(dbstmt, -1);
        }

        if (sqlite3_bind_int(dbstmt, offset + 3, (uid_t) (info->attr->st_uid)) != SQLITE_OK) {
            FAIL_SQLITE_STMT(dbstmt, -1);
        }

        if (sqlite3_bind_text(dbstmt, offset + 4, info->group, strlen(info->group), SQLITE_TRANSIENT) != SQLITE_OK) {
            FAIL_SQLITE_STMT(dbstmt, -1);
        }

        if (sqlite3_bind_int(dbstmt, offset + 5, (gid_t) (info->attr->st_gid)) != SQLITE_OK) {
            FAIL_SQLITE_STMT(dbstmt, -1);
        }

        if (sigver == SIGVER_21) {
            /*
             * Known bug: noarch packages require the endianness of the
             * installation host to match the build host.  Fixed in v2.2
             * packages.
             */
            if (sqlite3_bind_blob(dbstmt, offset + 6, &(info->attr->st_mode), sizeof(info->attr->st_mode), SQLITE_TRANSIENT) != SQLITE_OK) {
                FAIL_SQLITE_STMT(dbstmt, -1);
            }
        } else if (sigver >= SIGVER_22) {
            mode = htobe32(info->attr->st_mode);
            if (sqlite3_bind_blob(dbstmt, offset + 6, &mode, sizeof(mode), SQLITE_TRANSIENT) != SQLITE_OK) {
                FAIL_SQLITE_STMT(dbstmt, -1);
            }
        } else {
            fprintf(stderr, "%s (%d): unknown package signature %d\n", __func__, __LINE__, sigver);
            FAIL_SQLITE_STMT(dbstmt, -1);
        }

        if (sqlite3_bind_int(dbstmt, offset + 7, info->attr->st_size) != SQLITE_OK) {
            FAIL_SQLITE_STMT(dbstmt, -1);
        }

        if (sqlite3_bind_int(dbstmt, offset + 8, info->attr->st_mtime) != SQLITE_OK) {
            FAIL_SQLITE_STMT(dbstmt, -1);
        }

        if (sqlite3_bind_int(dbstmt, offset + 9, info->type) != SQLITE_OK) {
            FAIL_SQLITE_STMT(dbstmt, -1);
        }

        if (info->type == TYPE_SOFT || info->type == TYPE_HARD) {
            if (sqlite3_bind_text(dbstmt, offset + 10, info->linkname, strlen(info->linkname), SQLITE_TRANSIENT) != SQLITE_OK) {
                FAIL_SQLITE_STMT(dbstmt, -1);
            }
        } else {
            if (sqlite3_bind_null(dbstmt, offset + 10) != SQLITE_OK) {
                FAIL_SQLITE_STMT(dbstmt, -1);
            }
        }

        if (info->checksum == NULL) {
            if (sqlite3_bind_null(dbstmt, offset + 11) != SQLITE_OK) {
                FAIL_SQLITE_STMT(dbstmt, -1);
            }
        } else {
            if (sqlite3_bind_text(dbstmt, offset + 11, info->checksum, strlen(info->checksum), SQLITE_TRANSIENT) != SQLITE_OK) {
                FAIL_SQLITE_STMT(dbstmt, -1);
            }
        }

        if (sqlite3_step(dbstmt) != SQLITE_DONE) {
            FAIL_SQLITE_STMT(dbstmt, -1);
        }

        if (sqlite3_reset(dbstmt) != SQLITE_OK) {
            FAIL_SQLITE_STMT(dbstmt, -1);
        }

        if (sqlite3_clear_bindings(dbstmt) != SQLITE_OK) {
            FAIL_SQLITE_STMT(dbstmt, -1);
        }
    }

    free(modestr);
    return success;
}

/* Determine the dist-tools file type of a specific file. */
int get_file_type(struct finfo *info, int *errs) {
    int type = -1;
    char *hardtest = NULL;

    assert(info != NULL);

    /* check if this entry is a hard link to an existing file */
    hardtest = lookup_inode(info->attr->st_ino);

    /* just silently change problems here, since some previous function
     * should have handled the bitching */
    if (hardtest != NULL && strcmp(info->filename, hardtest)) {
        type = TYPE_HARD;
        free(info->linkname);
        info->linkname = NULL;

        if ((info->linkname = strdup(hardtest)) == NULL) {
            fprintf(stderr, "%s: Memory allocation failure\n", __func__);
            errno = 0;

            if (errs != NULL) {
                (*errs) = -1;
            }
        }
    } else if (S_ISREG(info->attr->st_mode)) {
        type = TYPE_REG;
    } else if (S_ISDIR(info->attr->st_mode)) {
        type = TYPE_DIR;
    } else if (S_ISLNK(info->attr->st_mode)) {
        type = TYPE_SOFT;
    } else if (S_ISCHR(info->attr->st_mode)) {
        type = TYPE_CHAR;
    } else if (S_ISBLK(info->attr->st_mode)) {
        type = TYPE_BLOCK;
    } else if (S_ISFIFO(info->attr->st_mode)) {
        type = TYPE_FIFO;
    } else if (S_ISSOCK(info->attr->st_mode)) {
        type = TYPE_SOCK;
    } else {
        type = TYPE_OTHER;
    }

    return type;
}

/* prepares all the info structs for writing to a file */
int gen_contents(struct finfo *info, enum checksum sumtype) {
    int errs = 0;
    int len = -1;
    char *hardtest = NULL;

    assert(info != NULL);
    assert(sumtype > 0);

    /* if we need the fullpath, get that */
    if (info->fullpath == NULL) {
        /* the filenames in the bom.db files lack leading slashes */
        if (strlen(prepend) == 1 && !strcmp(prepend, "/")) {
            xasprintf(&(info->fullpath), "/%s", info->filename);
        } else {
            xasprintf(&(info->fullpath), "%s/%s", prepend, info->filename);
        }

        /* if we lacked fullpath, reset the attr struct */
        if (lstat(info->fullpath, info->attr) == -1) {
            fprintf(stderr, "%s: Could not stat %s: %s\n", __func__,
                    info->fullpath, strerror(errno));
            abort();
        }
    }

    /* get file type */
    if ((info->type = get_file_type(info, &errs)) == -1) {
        fprintf(stderr, "%s (%d): invalid file type for %s\n", __FILE__, __LINE__, info->filename);
        errs = -1;
    }

    /* only compute checksums for real files (skip hard links) */
    if (info->type == TYPE_REG && info->checksum == NULL) {
        info->checksum = checksum(info->fullpath, sumtype);

        if (info->checksum == NULL) {
            fprintf(stderr,
                    "%s: Unable to calculate message checksum for %s: %s\n",
                    __func__, info->fullpath, strerror(errno));
            errs = -1;
            errno = 0;
        }
    }

    if (info->type == TYPE_SOFT && info->linkname == NULL) {
        if ((hardtest = calloc(1, PATH_MAX)) == NULL) {
            fprintf(stderr, "%s: Memory allocation failure\n", __func__);
            errs = -1;
            errno = 0;
        }

        if ((hardtest != NULL) && ((len = readlink(info->fullpath, hardtest, PATH_MAX-1)) == -1)) {
            fprintf(stderr, "%s: Error reading link %s: %s\n", __func__,
                    info->fullpath, strerror(errno));
            errs = -1;
            errno = 0;
        }

        hardtest[len >= PATH_MAX ? PATH_MAX - 1 : len] = '\0';

        if ((info->linkname = strdup(hardtest)) == NULL) {
            fprintf(stderr, "%s: Memory allocation failure\n", __func__);
            errs = -1;
            errno = 0;
        }

        free(hardtest);
    }

    return errs;
}

/* vim: set softtabstop=3 shiftwidth=3: */
