/* dist-tools database definitions header file */

/* {{{ Apache License version 2.0
 */
/*
 * Copyright 2003-2017 David Cantrell <david.l.cantrell@gmail.com>
 *                     David Shea <david@reallylongword.org>
 *                     Chris Lumens <chris@bangmoney.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/* }}} */

#ifndef _DIST_TOOLS_DB_H
#define _DIST_TOOLS_DB_H

#include <stdio.h>
#include <sqlite3.h>

/* SQL schemas */
#define BOM_INFO_SCHEMA                                          \
    "CREATE TABLE IF NOT EXISTS info                             \
        (name TEXT NOT NULL ON CONFLICT FAIL,                    \
         ver TEXT NOT NULL ON CONFLICT FAIL,                     \
         rel INTEGER NOT NULL ON CONFLICT FAIL CHECK (rel >= 0), \
         arch TEXT NOT NULL ON CONFLICT FAIL,                    \
         repo TEXT,                                              \
         size BIG INTEGER NOT NULL ON CONFLICT FAIL,             \
         debug BOOLEAN,                                          \
         desc TEXT,                                              \
         maintainer TEXT,                                        \
         license TEXT,                                           \
         depspath TEXT)"

#define BOM_DEPS_SCHEMA \
    "CREATE TABLE IF NOT EXISTS deps (name TEXT NOT NULL ON CONFLICT FAIL)"

#define BOM_PROVS_SCHEMA \
    "CREATE TABLE IF NOT EXISTS provs (name TEXT NOT NULL ON CONFLICT FAIL)"

#define BOM_CONTENTS_SCHEMA \
    "CREATE TABLE IF NOT EXISTS contents                           \
        (path TEXT NOT NULL ON CONFLICT FAIL,                      \
         ownername TEXT NOT NULL ON CONFLICT FAIL,                 \
         uid INTEGER NOT NULL ON CONFLICT FAIL CHECK (uid >= 0),   \
         groupname TEXT NOT NULL ON CONFLICT FAIL,                 \
         gid INTEGER NOT NULL ON CONFLICT FAIL CHECK (gid >= 0),   \
         mode BLOB NOT NULL ON CONFLICT FAIL,                      \
         size BIG INTEGER,                                         \
         mtime BIG INTEGER,                                        \
         filetype INTEGER CHECK (filetype >= 0 AND filetype <= 9), \
         linkdest TEXT,                                            \
         digest TEXT)"

#define DB_PACKAGES_SCHEMA                                       \
    "CREATE TABLE IF NOT EXISTS packages                         \
        (id INTEGER PRIMARY KEY,                                 \
         name TEXT NOT NULL ON CONFLICT FAIL,                    \
         ver TEXT NOT NULL ON CONFLICT FAIL,                     \
         rel INTEGER NOT NULL ON CONFLICT FAIL CHECK (rel >= 0), \
         arch TEXT NOT NULL ON CONFLICT FAIL,                    \
         repo TEXT,                                              \
         size BIG INTEGER NOT NULL ON CONFLICT FAIL,             \
         debug BOOLEAN,                                          \
         desc TEXT,                                              \
         maintainer TEXT,                                        \
         license TEXT,                                           \
         depspath TEXT,                                          \
         digest TEXT NOT NULL)"

#define DB_CONTENTS_SCHEMA                                         \
    "CREATE TABLE IF NOT EXISTS contents                           \
        (packageid INTEGER NOT NULL,                               \
         path TEXT NOT NULL ON CONFLICT FAIL,                      \
         ownername TEXT NOT NULL ON CONFLICT FAIL,                 \
         uid INTEGER NOT NULL ON CONFLICT FAIL CHECK (uid >= 0),   \
         groupname TEXT NOT NULL ON CONFLICT FAIL,                 \
         gid INTEGER NOT NULL ON CONFLICT FAIL CHECK (gid >= 0),   \
         mode BLOB NOT NULL ON CONFLICT FAIL,                      \
         size BIG INTEGER,                                         \
         mtime BIG INTEGER,                                        \
         filetype INTEGER CHECK (filetype >= 0 AND filetype <= 9), \
         linkdest TEXT,                                            \
         digest TEXT,                                              \
         FOREIGN KEY(packageid) REFERENCES packages(id))"

/* SQL statements */
#define BOM_INFO_INSERT     "INSERT INTO info VALUES (?,?,?,?,?,?,?,?,?,?,?)"
#define BOM_DEPS_INSERT     "INSERT INTO deps VALUES (?)"
#define BOM_PROVS_INSERT    "INSERT INTO provs VALUES (?)"
#define BOM_CONTENTS_INSERT "INSERT INTO contents VALUES (?,?,?,?,?,?,?,?,?,?,?)"
#define BOM_CONTENTS_ATTACH "ATTACH DATABASE '%s' AS bom"
#define BOM_CONTENTS_DETACH "DETACH DATABASE bom"
#define BOM_CONTENTS_SELECT "SELECT * FROM bom.contents"

#define DB_CONTENTS_INSERT  "INSERT INTO contents SELECT '%d',* FROM bom.contents"
#define DB_CONTENTS_GENDIST "INSERT INTO contents VALUES (?,?,?,?,?,?,?,?,?,?,?,?)"
#define DB_PACKAGES_INSERT  "INSERT INTO packages VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?)"
#define DB_PACKAGES_UPDATE  "UPDATE packages SET name=?, ver=?, rel=?, arch=?, repo=?, size=?, debug=?, desc=?, maintainer=?, license=?, depspath=?, digest=? WHERE id='%d'"
#define DB_CONTENTS_DELETE  "DELETE FROM contents WHERE packageid='%d'"

/* used during 'distadd -u' operations */
/* selects entries that have been removed in the newer version of a package */
#define DB_CONTENTS_GONE    "SELECT packageid,filetype,path,digest,linkdest FROM contents WHERE packageid=? AND path NOT IN (SELECT path FROM bom.contents)"

/* selects entries that have changed in the newer version of a package */
#define DB_CONTENTS_CHANGED "SELECT c.packageid,c.filetype,c.path,c.digest,c.linkdest FROM contents AS c, bom.contents AS b WHERE c.packageid=? AND (c.path = b.path) AND ((c.filetype != b.filetype) OR (c.filetype = %d AND c.linkdest != b.linkdest) OR (c.filetype = %d AND c.digest != b.digest))"

#define INFO_SELECT         "SELECT * FROM info"
#define CONTENTS_SELECT     "SELECT * FROM contents"
#define DEPS_SELECT         "SELECT * FROM deps"
#define PROVS_SELECT        "SELECT * FROM provs"

/* Used by 'distinfo -q' operations */
#define DB_INFO_ETC_QUERY   "SELECT name,ver,rel,arch,digest FROM packages WHERE debug='0'"
#define DB_INFO_QUERY       "SELECT packages.name,packages.ver,packages.rel,packages.arch,packages.debug FROM packages,contents WHERE (packages.id=contents.packageid AND contents.path=?) ORDER BY packages.name"

/* For 'distinfo -f' on installed packages */
#define DB_INFO_FILELIST    "SELECT path FROM contents WHERE packageid=?"

/* For 'distinfo -f' on package files */
#define BOM_INFO_FILELIST   "SELECT path FROM contents"

/* For 'distinfo --provs' on package files */
#define BOM_INFO_PROVS      "SELECT * FROM provs"

/* For 'distinfo --deps' on package files */
#define BOM_INFO_DEPS       "SELECT * FROM deps"

/* For 'etcmerge' getting a list of all packages */
#define ETCMERGE_ALLPKGS    "SELECT name,ver,rel,arch,debug,digest FROM packages ORDER BY name"

/* For 'distchk' and 'distrm' getting package contents */
#define PACKAGE_CONTENTS    "SELECT * FROM contents WHERE packageid=%d"

/* Macros */
#define FAIL_SQLITE(db, err, errcode)                        \
    fprintf(stderr, "%s (%d): %s (%d)\n", __func__, __LINE__,\
            sqlite3_errmsg((db)), sqlite3_errcode((db)));    \
    sqlite3_free((err));                                     \
    return (errcode);

#define FAIL_SQLITE_STMT(stmt, errcode)                      \
    fprintf(stderr, "%s (%d): %s (%d)\n", __func__, __LINE__,\
            sqlite3_errmsg(sqlite3_db_handle((stmt))),       \
            sqlite3_errcode(sqlite3_db_handle((stmt))));     \
    return (errcode);

#endif /* _DIST_TOOLS_DB_H */
