/* Package list functions. */

/* {{{ Apache License version 2.0
 */
/*
 * Copyright 2018 David Cantrell <david.l.cantrell@gmail.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/* }}} */

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <sys/queue.h>
#include "dist-tools.h"

static int _compare_pkglist_entry(const void *a, const void *b) {
    pkglist_entry_t *p1 = *(pkglist_entry_t **) a;
    pkglist_entry_t *p2 = *(pkglist_entry_t **) b;

    return strcmp(p1->pkg->name, p2->pkg->name);
}

/*
 * Allocate memory for and initialize a new pkglist_t.
 */
pkglist_t *init_pkglist(void) {
    pkglist_t *list = NULL;

    if ((list = calloc(1, sizeof(*list))) == NULL) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
        abort();
    }

    TAILQ_INIT(list);
    return list;
}

/*
 * Given an allocated pkglist_t, free all the memory associated with it.
 */
void free_pkglist(pkglist_t *list) {
    pkglist_entry_t *entry = NULL;

    if (list == NULL) {
        return;
    }

    while (!TAILQ_EMPTY(list)) {
        entry = TAILQ_FIRST(list);
        TAILQ_REMOVE(list, entry, items);
        freepkg(entry->pkg);
        entry->pkg = NULL;
        free(entry);
    }

    free(list);
    list = NULL;

    return;
}

/*
 * Return the entry count for the given pkglist.
 */
int length_pkglist(pkglist_t *list) {
    int c = 0;
    pkglist_entry_t *entry = NULL;

    if (list == NULL) {
        return 0;
    }

    TAILQ_FOREACH(entry, list, items) {
        c++;
    }

    return c;
}

/*
 * Sort the specified pkglist_t in ascending order.  This works by first
 * making an array copy of the pkglist_t, running that through qsort(),
 * and then creating a new pkglist_t with the entries in the sorted order.
 * This sounds expensive, but it's just arrays and lists of pointers.
 */
pkglist_t *sort_pkglist(pkglist_t *list) {
    int i;
    int len;
    pkglist_t *sorted = NULL;
    pkglist_entry_t **array = NULL;
    pkglist_entry_t *entry = NULL;

    if ((len = length_pkglist(list)) == 0) {
        return sorted;
    }

    /* create an array of the list contents */
    sorted = init_pkglist();

    if ((array = calloc(len, sizeof(pkglist_entry_t *))) == NULL) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
        return NULL;
    }

    i = 0;
    TAILQ_FOREACH(entry, list, items) {
        array[i] = entry;
        i++;
    }

    /* sort the array */
    qsort(array, len, sizeof(pkglist_entry_t *), _compare_pkglist_entry);

    /* create new list to return */
    for (i = 0; i < len; i++) {
        TAILQ_INSERT_TAIL(sorted, array[i], items);
    }

    /* we don't need this anymore */
    free(array);

    return sorted;
}

/*
 * Given a string contacting a package digest and a list of packages,
 * return the package that matches the digest string.
 */
struct dist *find_pkg_by_hash(pkglist_t *list, const char *hash) {
    pkglist_entry_t *entry = NULL;

    assert(hash != NULL);

    TAILQ_FOREACH(entry, list, items) {
        if (!strcmp(entry->pkg->pkgdigest, hash)) {
            return entry->pkg;
        }
    }

    return NULL;
}
