/* File and directory removal helper functions. */

/* {{{ Apache License version 2.0
 */
/*
 * Copyright 2003-2018 David Shea <david@reallylongword.org>
 *                     David Cantrell <david.l.cantrell@gmail.com>
 *                     Chris Lumens <chris@bangmoney.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/* }}} */

#include <stdint.h>
#include <assert.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "dist-tools.h"
#include "opts.h"

/* Global variables set by whatever controlling program is in charge. */
extern char *prepend;
extern uint64_t opts;

/*
 * Callback when iterating a TAILQ of struct finfo objects.  Used to
 * safely remove files.
 */
int checkunlink(struct finfo *info) {
    int err = 0;
    int bad = 0;

    assert(info != NULL);

    if (info->type == TYPE_DIR) {
        return err;
    }

    bad = checkfile(info, prepend, (opts & OPTS_STRICT), OPTS_NO_REPAIR);

    if ((opts & OPTS_QUICK) || (opts & OPTS_FORCE) || !bad) {
        if (unlink(info->fullpath)) {
            fprintf(stderr, "%s: Unable to unlink %s: %m\n", __func__, info->fullpath);
            err = -1;
        }
    }

    return err;
}

/*
 * Callback when iterating a TAILQ of struct finfo objects.  Used to
 * safely remove directories.  Note, the caller needs to ensure that
 * directory removal happens in the correct order.  This function
 * cannot remove an empty tree of subdirectories, only empty directories.
 */
int checkrmdir(struct finfo *info) {
    int errs = 0;
    int bad = 0;
    char *efile = NULL;

    assert(info != NULL);

    bad = checkfile(info, prepend, (opts & OPTS_STRICT), OPTS_NO_REPAIR);
    efile = joinpath(prepend, info->filename, NULL);

    if ((opts & OPTS_QUICK) || (opts & OPTS_FORCE) || !bad) {
        if (rmdir(efile)) {
            if (errno == ENOTEMPTY && (opts & OPTS_VERBOSE)) {
                printf("%s not empty, not removing\n", efile);
            } else if (errno == EBUSY) {
                fprintf(stderr, "%s still in use, not removing\n", efile);
                errs = -1;
            } else if (errno != ENOTEMPTY && errno != EBUSY) {
                fprintf(stderr, "%s: Unable to remove %s: %s\n", __func__,
                        efile, strerror(errno));
                errs = -1;
            }
        }
    }

    free(efile);
    return errs;
}
