/*
 * Takes a package format string of the form:
 *     NAME
 *     NAME.ARCH
 *     NAME.debug
 *     NAME.debug.ARCH
 *     NAME%VER
 *     NAME-VER
 *     NAME-VER.ARCH
 *     NAME-VER.debug
 *     NAME-VER.debug.ARCH
 *     NAME-VER-REL
 *     NAME-VER-REL.ARCH
 *     NAME-VER-REL.debug
 *     NAME-VER-REL.debug.ARCH
 * finds all versions of an installed package, and fills a list with struct
 * dists.  Returns the list.
 */

/* {{{ Apache License version 2.0
 */
/*
 * Copyright 2003-2018 David Shea <david@reallylongword.org>
 *                     David Cantrell <david.l.cantrell@gmail.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/* }}} */

#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include "dist-tools.h"
#include "db.h"
#include "pkglist.h"

/*
 * This is the main function that finds package database entries.  Users
 * and callers can specify packages using this format:
 *     NAME
 *     NAME.ARCH
 *     NAME.debug
 *     NAME.debug.ARCH
 *     NAME%VER
 *     NAME-VER
 *     NAME-VER.ARCH
 *     NAME-VER.debug
 *     NAME-VER.debug.ARCH
 *     NAME-VER-REL
 *     NAME-VER-REL.ARCH
 *     NAME-VER-REL.debug
 *     NAME-VER-REL.debug.ARCH
 *
 * The db_installed variable is the path to installed.db, which is generally
 * set by main() in whatever program is running.
 *
 * The pkglist variable is what will be filled with the returned packages.
 * Caller must free this list, but they should know that since they own it.
 *
 * Returns the list of matching packages, which could be NULL.  In which
 * case there were no matches.  Only zuul.
 */
pkglist_t *scanpkg(const char *db_installed, const char *pkgstr) {
    char *query = NULL;
    char *compare = NULL;
    sqlite3 *db = NULL;
    sqlite3_stmt *stmt = NULL;
    char *err = NULL;
    struct dist tmppkg;
    bool first = false;
    pkglist_entry_t *entry = NULL;
    pkglist_t *pkglist = NULL;

    /* open the installed package database */
    if (sqlite3_open(db_installed, &db) != SQLITE_OK) {
        FAIL_SQLITE(db, NULL, NULL);
    }

    if (sqlite3_exec(db, "PRAGMA synchronous = OFF", NULL, NULL, &err) != SQLITE_OK) {
        FAIL_SQLITE(db, err, NULL);
    }

    if (sqlite3_exec(db, "PRAGMA journal_mode = MEMORY", NULL, NULL, &err) != SQLITE_OK) {
        FAIL_SQLITE(db, err, NULL);
    }

    /* handle input in various forms */
    if (pkgstr != NULL) {
        if (index(pkgstr, '%') || index(pkgstr, '-') || index(pkgstr, '.')) {
            /*
             * support these formats:
             * NAME.ARCH
             * NAME.debug
             * NAME.debug.ARCH
             * NAME%VER
             * NAME-VER
             * NAME-VER.ARCH
             * NAME-VER.debug
             * NAME-VER.debug.ARCH
             * NAME-VER-REL
             * NAME-VER-REL.ARCH
             * NAME-VER-REL.debug
             * NAME-VER-REL.debug.ARCH
             *
             * This is complicated, but let me explain.  Rather than parse the
             * input string, iterate over each package and build candidate input
             * strings and compare those with what we were actually given.  If
             * a match is found, add that package to the list and continue.  The
             * idea here is to avoid parsing errors on the input string at the
             * expense of a slightly longer run time.
             */
            first = true;
            query = strdup("SELECT * FROM packages WHERE id IN (");

            if (sqlite3_prepare(db, "SELECT * FROM packages", -1, &stmt, 0) != SQLITE_OK) {
                FAIL_SQLITE(db, NULL, NULL);
            }

            while (sqlite3_step(stmt) == SQLITE_ROW) {
                /* clear out compare from previous loop iterations */
                free(compare);
                compare = NULL;

                /* start by getting this package's information */
                initpkg(&tmppkg);

                if (getpkginfo(&tmppkg, PKG_DBENTRY, NULL, NULL, stmt)) {
                    return NULL;
                }

                if (!tmppkg.debug) {
                    /* NAME-VER-REL.ARCH */
                    xasprintf(&compare, "%s-%s-%d.%s", tmppkg.name, tmppkg.ver, tmppkg.rel, tmppkg.arch);

                    if (!strcmp(compare, pkgstr)) {
                        if (first) {
                            first = false;
                        } else {
                            query = strappend(query, ",");
                        }

                        query = strappend_int(query, tmppkg.id);
                        continue;
                    }

                    free(compare);
                    compare = NULL;

                    /* NAME-VER-REL */
                    xasprintf(&compare, "%s-%s-%d", tmppkg.name, tmppkg.ver, tmppkg.rel);

                    if (!strcmp(compare, pkgstr)) {
                        if (first) {
                            first = false;
                        } else {
                            query = strappend(query, ",");
                        }

                        query = strappend_int(query, tmppkg.id);
                        continue;
                    }

                    free(compare);
                    compare = NULL;

                    /* NAME-VER.ARCH */
                    xasprintf(&compare, "%s-%s.%s", tmppkg.name, tmppkg.ver, tmppkg.arch);

                    if (!strcmp(compare, pkgstr)) {
                        if (first) {
                            first = false;
                        } else {
                            query = strappend(query, ",");
                        }

                        query = strappend_int(query, tmppkg.id);
                        continue;
                    }

                    free(compare);
                    compare = NULL;

                    /* NAME-VER */
                    xasprintf(&compare, "%s-%s", tmppkg.name, tmppkg.ver);

                    if (!strcmp(compare, pkgstr)) {
                        if (first) {
                            first = false;
                        } else {
                            query = strappend(query, ",");
                        }

                        query = strappend_int(query, tmppkg.id);
                        continue;
                    }

                    free(compare);
                    compare = NULL;

                    /* NAME%VER */
                    xasprintf(&compare, "%s%%%s", tmppkg.name, tmppkg.ver);

                    if (!strcmp(compare, pkgstr)) {
                        if (first) {
                            first = false;
                        } else {
                            query = strappend(query, ",");
                        }

                        query = strappend_int(query, tmppkg.id);
                        continue;
                    }

                    free(compare);
                    compare = NULL;

                    /* NAME.ARCH */
                    xasprintf(&compare, "%s.%s", tmppkg.name, tmppkg.arch);

                    if (!strcmp(compare, pkgstr)) {
                        if (first) {
                            first = false;
                        } else {
                            query = strappend(query, ",");
                        }

                        query = strappend_int(query, tmppkg.id);
                        continue;
                    }

                    free(compare);
                    compare = NULL;

                    /* NAME */
                    compare = strdup(tmppkg.name);

                    if (!strcmp(compare, pkgstr)) {
                        if (first) {
                            first = false;
                        } else {
                            query = strappend(query, ",");
                        }

                        query = strappend_int(query, tmppkg.id);
                        continue;
                    }

                    free(compare);
                    compare = NULL;
                } else if (tmppkg.debug) {
                    /* NAME-VER-REL.debug.ARCH */
                    xasprintf(&compare, "%s-%s-%d.debug.%s", tmppkg.name, tmppkg.ver, tmppkg.rel, tmppkg.arch);

                    if (!strcmp(compare, pkgstr)) {
                        if (first) {
                            first = false;
                        } else {
                            query = strappend(query, ",");
                        }

                        query = strappend_int(query, tmppkg.id);
                        continue;
                    }

                    free(compare);
                    compare = NULL;

                    /* NAME-VER-REL.debug */
                    xasprintf(&compare, "%s-%s-%d.debug", tmppkg.name, tmppkg.ver, tmppkg.rel);

                    if (!strcmp(compare, pkgstr)) {
                        if (first) {
                            first = false;
                        } else {
                            query = strappend(query, ",");
                        }

                        query = strappend_int(query, tmppkg.id);
                        continue;
                    }

                    free(compare);
                    compare = NULL;

                    /* NAME-VER.debug.ARCH */
                    xasprintf(&compare, "%s-%s%s.%s", tmppkg.name, tmppkg.ver,
                              (tmppkg.debug ? ".debug" : ""), tmppkg.arch);

                    if (!strcmp(compare, pkgstr)) {
                        if (first) {
                            first = false;
                        } else {
                            query = strappend(query, ",");
                        }

                        query = strappend_int(query, tmppkg.id);
                        continue;
                    }

                    free(compare);
                    compare = NULL;

                    /* NAME-VER.debug */
                    xasprintf(&compare, "%s-%s%s", tmppkg.name, tmppkg.ver,
                              (tmppkg.debug ? ".debug" : ""));

                    if (!strcmp(compare, pkgstr)) {
                        if (first) {
                            first = false;
                        } else {
                            query = strappend(query, ",");
                        }

                        query = strappend_int(query, tmppkg.id);
                        continue;
                    }

                    free(compare);
                    compare = NULL;

                    /* NAME.debug.ARCH */
                    xasprintf(&compare, "%s%s.%s", tmppkg.name,
                              (tmppkg.debug ? ".debug" : ""), tmppkg.arch);

                    if (!strcmp(compare, pkgstr)) {
                        if (first) {
                            first = false;
                        } else {
                            query = strappend(query, ",");
                        }

                        query = strappend_int(query, tmppkg.id);
                        continue;
                    }

                    free(compare);
                    compare = NULL;

                    /* NAME.debug */
                    xasprintf(&compare, "%s%s", tmppkg.name,
                              (tmppkg.debug ? ".debug" : ""));

                    if (!strcmp(compare, pkgstr)) {
                        if (first) {
                            first = false;
                        } else {
                            query = strappend(query, ",");
                        }

                        query = strappend_int(query, tmppkg.id);
                        continue;
                    }

                    free(compare);
                    compare = NULL;
                }
            }

            if (sqlite3_finalize(stmt) != SQLITE_OK) {
                FAIL_SQLITE(db, NULL, NULL);
            }

            query = strappend(query, ")");
        } else {
            /* assume package name as a last resort */
            xasprintf(&query, "SELECT * FROM packages WHERE name='%s'", pkgstr);
        }
    } else {
        /* user specified no package, gather entire database */
        xasprintf(&query, "SELECT * FROM packages");
    }

    /* run the query and for each row, build a struct dist */
    if (sqlite3_prepare(db, query, -1, &stmt, 0) != SQLITE_OK) {
        free(query);
        FAIL_SQLITE(db, NULL, NULL);
    }

    while (sqlite3_step(stmt) == SQLITE_ROW) {
        /* the pkglist may need initializing */
        if (pkglist == NULL) {
            pkglist = init_pkglist();
        }

        /* init pkglist entry */
        if ((entry = calloc(1, sizeof(pkglist_entry_t))) == NULL) {
            fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
            return NULL;
        }

        /* init the struct dist for this entry */
        if ((entry->pkg = calloc(1, sizeof(struct dist))) == NULL) {
            fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
            abort();
        }

        if (getpkginfo(entry->pkg, PKG_DBENTRY, NULL, NULL, stmt)) {
            return NULL;
        }

        TAILQ_INSERT_TAIL(pkglist, entry, items);
    }

    if (sqlite3_finalize(stmt) != SQLITE_OK) {
        FAIL_SQLITE(db, NULL, NULL);
    }

    free(query);
    sqlite3_close(db);
    return pkglist;
}
