/* Generic string list functions. */

/* {{{ Apache License version 2.0
 */
/*
 * Copyright 2018 David Cantrell <david.l.cantrell@gmail.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/* }}} */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <sys/queue.h>
#include "strlist.h"

/*
 * Allocate memory for and initialize a new strlist_t.
 */
strlist_t *init_strlist(void) {
    strlist_t *list = NULL;

    if ((list = calloc(1, sizeof(*list))) == NULL) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
        abort();
    }

    TAILQ_INIT(list);
    return list;
}

/*
 * Given an allocated strlist_t, free all the memory associated with it.
 */
void free_strlist(strlist_t *list) {
    strlist_entry_t *entry = NULL;

    if (list == NULL) {
        return;
    }

    while (!TAILQ_EMPTY(list)) {
        entry = TAILQ_FIRST(list);
        TAILQ_REMOVE(list, entry, items);
        free(entry->val);
        entry->val = NULL;
        free(entry);
    }

    free(list);

    return;
}

/*
 * Returns true if the specified string is in the given list.
 */
bool is_str_in_list(strlist_t *list, const char *str) {
    strlist_entry_t *entry = NULL;

    assert(str != NULL);

    if (list == NULL) {
        return false;
    }

    TAILQ_FOREACH(entry, list, items) {
        if (!strcmp(entry->val, str)) {
            return true;
        }
    }

    return false;
}

/*
 * Return the entry count for the given strlist.
 */
int length_strlist(strlist_t *list) {
    int c = 0;
    strlist_entry_t *entry = NULL;

    if (list == NULL) {
        return 0;
    }

    TAILQ_FOREACH(entry, list, items) {
        c++;
    }

    return c;
}

/*
 * Insert a strlist entry in sorted order.  Tosses duplicate strings.
 */
strlist_t *sorted_strlist_insert(strlist_t *list, const char *s) {
    int i = 0;
    int len = 0;
    strlist_entry_t *entry = NULL;
    strlist_entry_t *new = NULL;

    assert(list != NULL);
    assert(s != NULL);

    /* The new entry. */
    if ((new = calloc(1, sizeof(*new))) == NULL) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
        return NULL;
    } else {
        new->val = strdup(s);
    }

    len = length_strlist(list);

    /* Try to insert in sorted order. */
    if (len > 0) {
        TAILQ_FOREACH(entry, list, items) {
            i = strcmp(new->val, entry->val);

            if (i < 0) {
                TAILQ_INSERT_AFTER(list, entry, new, items);
                return list;
            } else if (i > 0) {
                TAILQ_INSERT_BEFORE(entry, new, items);
                return list;
            } else if (i == 0) {
                /* duplicate, let's ignore it */
                free(new->val);
                free(new);
                return list;
            }
        }
    }

    /* Handle the empty list state */
    TAILQ_INSERT_TAIL(list, new, items);

    return list;
}

/*
 * Given a string and a delimiter string, tokenize it and store
 * the resulting strings in a strlist.  Return the strlist.
 */
strlist_t *split_strlist(char *in, char *delim) {
    char *str = NULL;
    char *end = NULL;
    char *tok = NULL;
    strlist_t *list = NULL;
    strlist_entry_t *entry = NULL;

    assert(in != NULL);

    str = strdup(in);
    end = str;
    list = init_strlist();

    while ((tok = strsep(&end, delim)) != NULL) {
        if ((entry = calloc(1, sizeof(strlist_entry_t))) == NULL) {
            fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
            free(str);
            free_strlist(list);
            return NULL;
        }

        entry->val = strdup(tok);
        TAILQ_INSERT_TAIL(list, entry, items);
    }

    free(str);
    return list;
}
