/*
 * Split a string in to an array of substrings, using the string of
 * delimiters as the breaking points.
 */

/* {{{ Apache License version 2.0
 */
/*
 * Copyright 2018 David Cantrell <david.l.cantrell@gmail.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/* }}} */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include "dist-tools.h"

/*
 * Tokenize a string based on tokens listed in delim, return the tokens
 * as a char array.  The caller needs to free the char array because it
 * will contain pointers to copies of the passed in string.  This is
 * because the strsep() function has to modify the original string.
 */
char **strsplit(char *s, char *delim) {
    int i = 0;
    char *token = NULL;
    char *walk = s;
    char *head = NULL;
    char **ret = NULL;

    assert(s != NULL);
    assert(delim != NULL);

    /* compute length of ret array */
    while (*walk++ != '\0') {
        if (index(delim, *walk) != NULL) {
            i++;
        }
    }

    i++;     /* make sure the array of strings has a NULL last element */

    /* allocate return array */
    if ((ret = calloc(i, sizeof(char *))) == NULL) {
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, strerror(errno));
        abort();
    }

    /* split up the string */
    head = walk = strdup(s);
    i = 0;

    while ((token = strsep(&walk, delim)) != NULL) {
        xasprintf(&ret[i], "%s", token);
        i++;
    }

    ret[i] = NULL;
    free(head);

    return ret;
}

/*
 * Return number of elements in an array of strings.
 */
int strvlen(char **s) {
    int len = 0;

    if (s == NULL) {
        return len;
    }

    while (*s++ != NULL) {
        len++;
    }

    return len;
}

/*
 * Free an array of strings.  Assumes the array elements must be freed.
 */
void strvfree(char **s) {
    int i = 0;
    int len = strvlen(s);

    if (len == 0) {
        return;
    }

    while (i++ < len) {
        free(s[i]);
        s[i] = NULL;
    }

    free(s);
    return;
}
