#!/bin/sh
# Verify we can install a package to a specific root tree

CWD="$(pwd)"

title=" >>> $(basename ${0}) <<< "
line="-$(echo ${title} | sed -e 's|.|-|g')-"
echo
echo "+${line}+"
echo "|${title}|"
echo "+${line}+"
echo

LEAKCHECK_CMD=
[ "${LEAKCHECK}" = "y" ] && LEAKCHECK_CMD="valgrind --leak-check=full --show-reachable=yes "

DISTMK="$(readlink -f ../distmk)"
DISTADD="$(readlink -f ../distadd)"
ETCMERGE="$(readlink -f ../etcmerge)"
TREE="$(mktemp -d)"
META="$(mktemp -d)"
OUTPUT="$(mktemp -d)"
TMPDIST="$(mktemp -d)"
FAILS=0
EXPECTED_FAILS=0

[ ! -x ${DISTMK} ] && ${MAKE} -C ../ distmk
[ ! -x ${DISTADD} ] && ${MAKE} -C ../ distadd
[ ! -x ${ETCMERGE} ] && ${MAKE} -C ../ etcmerge

# create a tree to package up
mkdir -p ${TREE}/etc
mkdir -p ${TREE}/usr/sbin
echo "System ready" > ${TREE}/etc/issue
cp -p ${DISTMK} ${TREE}/usr/sbin/distmk

# create plist
cat << EOF > ${META}/plist
etc root(0) root(0) 0755
etc/issue root(0) root(0) 0644
usr root(0) root(0) 0755
usr/sbin root(0) root(0) 0755
usr/sbin/distmk root(0) root(0) 0755
EOF

echo "Big giant legal file." > ${META}/legal

cat << EOF > ${META}/pkginfo
desc="Test program"
maintainer="J. Random Packager <someone@example.net>"
license=WTFPL
EOF

# create the package
if [ "${LEAKCHECK}" = "y" ]; then
    ${LEAKCHECK_CMD} \
    ${DISTMK} -t ${TREE} -p "test" -v "1.0" -r 1 -m ${META} -l ${META}/plist \
              -R peachtree -o ${OUTPUT} --tmpdir=${TMPDIST}
else
    ${DISTMK} -t ${TREE} -p "test" -v "1.0" -r 1 -m ${META} -l ${META}/plist \
              -R peachtree -o ${OUTPUT} --tmpdir=${TMPDIST} >/dev/null 2>&1
fi
PKG="test-1.0-1.$(uname -m).dist"

# install to / should fail if we are not root
if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distadd PKG'..."
    ${LEAKCHECK_CMD} ${DISTADD} -d ${OUTPUT}/${PKG}
else
    echo -n "checking 'distadd PKG'..."
    if [ $(id -u) -ne 0 ]; then
        echo "SKIPPING (cannot perform test as non-root)"
    else
        EXPECTED_FAILS=$((${EXPECTED_FAILS} + 1))
        out="$(${DISTADD} -d ${OUTPUT}/${PKG} 2>&1)"
        r=$?
        out="$(echo "${out}" | head -n 1)"
        if [ ${r} -eq 1 -a "${out}" = "copyfile: Unable to open /usr/sbin/distmk.distnew for writing: Permission denied" ]; then
            FAILS=$((${FAILS} + 1))
            echo "passed."
        else
            echo "FAILED."
        fi
    fi
fi

# install to a staging root should work fine
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distadd -r PATH PKG'..."
    ${LEAKCHECK_CMD} ${DISTADD} -d -r ${OUTPUT}/r ${OUTPUT}/${PKG}
else
    echo -n "checking 'distadd -r PATH PKG'..."
    out="$(${DISTADD} -d -r ${OUTPUT}/r ${OUTPUT}/${PKG} 2>&1)"
    r=$?
    if [ ${r} -eq 1 ]; then
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    else
        echo "passed."
    fi
fi

# installing a package we already have should fail
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distadd -r PATH PKG' on already installed PKG..."
    ${LEAKCHECK_CMD} ${DISTADD} -d -r ${OUTPUT}/r ${OUTPUT}/${PKG}
    ${LEAKCHECK_CMD} ${DISTADD} -d -r ${OUTPUT}/r ${OUTPUT}/${PKG}
else
    echo -n "checking 'distadd -r PATH PKG' on already installed PKG..."
    ${DISTADD} -d -r ${OUTPUT}/r ${OUTPUT}/${PKG} >/dev/null 2>&1
    out="$(${DISTADD} -d -r ${OUTPUT}/r ${OUTPUT}/${PKG} 2>&1)"
    r=$?
    out="$(echo "${out}" | grep "test already installed")"
    if [ ${r} -eq 1 -a ! -z "${out}" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# force installing a package we already have should work
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distadd -f -r PATH PKG' on already installed PKG..."
    ${LEAKCHECK_CMD} ${DISTADD} -d -r ${OUTPUT}/r ${OUTPUT}/${PKG}
    ${LEAKCHECK_CMD} ${DISTADD} -d -f -r ${OUTPUT}/r ${OUTPUT}/${PKG}
else
    echo -n "checking 'distadd -f -r PATH PKG' on already installed PKG..."
    spec="$(basename ${PKG} .dist)"
    ${DISTADD} -d -r ${OUTPUT}/r ${OUTPUT}/${PKG} >/dev/null 2>&1
    out="$(${DISTADD} -d -f -r ${OUTPUT}/r ${OUTPUT}/${PKG} 2>&1)"
    r=$?
    out="$(echo "${out}" | tail -n 1)"
    p1=$(sqlite ${OUTPUT}/r/var/db/dist/installed.db "SELECT * FROM packages" | wc -l)
    if [ ${r} -eq 0 -a ${p1} -eq 2 -a "${out}" = "Force installed package ${spec} successfully" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# installing a package and generating metadata
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distadd -g' on new package..."
    ${LEAKCHECK_CMD} ${DISTADD} -g -d -r ${OUTPUT}/r ${OUTPUT}/${PKG}
else
    echo -n "checking 'distadd -g' on new package..."
    out="$(${DISTADD} -g -d -r ${OUTPUT}/r ${OUTPUT}/${PKG} 2>&1)"
    r=$?
    p1=$(sqlite ${OUTPUT}/r/var/db/dist/installed.db "SELECT * FROM packages" 2>/dev/null | wc -l)
    p2=$(sqlite ${OUTPUT}/r/var/db/dist/installed.db "SELECT * FROM contents" 2>/dev/null | wc -l)
    if [ ${r} -eq 0 -a ${p1} -eq 1 -a ${p2} -eq 3 -a -x ${OUTPUT}/r/usr/sbin/distmk ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# invoking etcmerge to install initial configuration files
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distadd -e' for initial config file install..."
    env ETCMERGE=${ETCMERGE} ${LEAKCHECK_CMD} ${DISTADD} -d -e -r ${OUTPUT}/r ${OUTPUT}/${PKG}
else
    echo -n "checking 'distadd -e' for initial config file install..."
    out="$(env ETCMERGE=${ETCMERGE} ${DISTADD} -d -e -r ${OUTPUT}/r ${OUTPUT}/${PKG} 2>&1)"
    r=$?
    out="$(echo "${out}" | grep "Installing new" | grep "/etc/issue")"
    if [ ${r} -eq 0 -a ! -z "${out}" -a -r ${OUTPUT}/r/etc/issue ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

rm -rf ${TREE}
rm -rf ${META}
rm -rf ${TMPDIST}
rm -rf ${OUTPUT}

if [ "${LEAKCHECK}" = "y" ]; then
    exit 0
else
    # expected $EXPECTED_FAILS failures
    [ ${FAILS} -eq ${EXPECTED_FAILS} ] && exit 0 || exit 1
fi
