#!/bin/sh
# Verify standard integrity checks work.

CWD="$(pwd)"

title=" >>> $(basename ${0}) <<< "
line="-$(echo ${title} | sed -e 's|.|-|g')-"
echo
echo "+${line}+"
echo "|${title}|"
echo "+${line}+"
echo

LEAKCHECK_CMD=
[ "${LEAKCHECK}" = "y" ] && LEAKCHECK_CMD="valgrind --leak-check=full --show-reachable=yes "

DISTMK=$(readlink -f ../distmk)
DISTADD=$(readlink -f ../distadd)
DISTCHK=$(readlink -f ../distchk)
TREE="$(mktemp -d)"
META="$(mktemp -d)"
OUTPUT="$(mktemp -d)"
TMPDIST="$(mktemp -d)"
FAILS=0
EXPECTED_FAILS=0

[ ! -x ${DISTMK} ] && ${MAKE} -C ../ distmk
[ ! -x ${DISTADD} ] && ${MAKE} -C ../ distadd
[ ! -x ${DISTCHK} ] && ${MAKE} -C ../ distchk

# create a tree to package up
mkdir -p ${TREE}/usr/sbin
cp -p ${DISTMK} ${TREE}/usr/sbin/distmk

# create plist
USER="$(id -un)"
tUID="$(id -ur)"
GROUP="$(id -gn)"
tGID="$(id -gr)"

echo "usr ${USER}(${tUID}) ${GROUP}(${tGID}) 0755" > ${META}/plist
echo "usr/sbin ${USER}(${tUID}) ${GROUP}(${tGID}) 0755" >> ${META}/plist
echo "usr/sbin/distmk ${USER}(${tUID}) ${GROUP}(${tGID}) 0755" >> ${META}/plist

echo "Big giant legal file." > ${META}/legal

cat << EOF > ${META}/pkginfo
desc="Test program"
maintainer="J. Random Packager <someone@example.net>"
license=WTFPL
EOF

# create the package
if [ "${LEAKCHECK}" = "y" ]; then
    ${LEAKCHECK_CMD} \
    ${DISTMK} -t ${TREE} -p "test" -v "1.0" -r 1 -m ${META} -l ${META}/plist \
              -R peachtree -o ${OUTPUT} --tmpdir=${TMPDIST}
else
    ${DISTMK} -t ${TREE} -p "test" -v "1.0" -r 1 -m ${META} -l ${META}/plist \
              -R peachtree -o ${OUTPUT} --tmpdir=${TMPDIST} >/dev/null 2>&1
fi

PKG="test-1.0-1.$(uname -m).dist"

# check installed package
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distchk PKG'..."
    ${LEAKCHECK_CMD} ${DISTADD} -d -r ${OUTPUT}/r ${OUTPUT}/${PKG}
else
    echo -n "checking 'distchk PKG'..."
    ${DISTADD} -d -r ${OUTPUT}/r ${OUTPUT}/${PKG} >/dev/null 2>&1
    out="$(${DISTCHK} -r ${OUTPUT}/r test 2>&1 | tail -n 1)"
    r=$?
    if [ ${r} -eq 0 -a "${out}" = "OK" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# check installed package with expected file type failure
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distchk PKG' with expected file type failure..."
    ${LEAKCHECK_CMD} ${DISTADD} -d -r ${OUTPUT}/r ${OUTPUT}/${PKG}
else
    echo -n "checking 'distchk PKG' with expected file type failure..."
    ${DISTADD} -d -r ${OUTPUT}/r ${OUTPUT}/${PKG} >/dev/null 2>&1
    rm -f ${OUTPUT}/r/usr/sbin/distmk
    mkfifo ${OUTPUT}/r/usr/sbin/distmk
    out="$(${DISTCHK} -r ${OUTPUT}/r test 2>&1)"
    r=$?
    out="$(echo "${out}" | grep "distmk is not a regular file")"
    if [ ${r} -eq 1 -a ! -z "${out}" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# check installed package with expected file mode failure
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distchk PKG' with expected file mode failure..."
    ${LEAKCHECK_CMD} ${DISTADD} -d -r ${OUTPUT}/r ${OUTPUT}/${PKG}
else
    echo -n "checking 'distchk PKG' with expected file mode failure..."
    ${DISTADD} -d -r ${OUTPUT}/r ${OUTPUT}/${PKG} >/dev/null 2>&1
    chmod 0644 ${OUTPUT}/r/usr/sbin/distmk
    out="$(${DISTCHK} -r ${OUTPUT}/r test 2>&1)"
    r=$?
    out="$(echo "${out}" | grep "distmk (0644) does not match expected (0755)")"
    if [ ${r} -eq 1 -a ! -z "${out}" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# check installed package with expected file checksum failure
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distchk PKG' with expected file checksum failure..."
    ${LEAKCHECK_CMD} ${DISTADD} -d -r ${OUTPUT}/r ${OUTPUT}/${PKG}
else
    echo -n "checking 'distchk PKG' with expected file checksum failure..."
    ${DISTADD} -d -r ${OUTPUT}/r ${OUTPUT}/${PKG} >/dev/null 2>&1
    echo "12345" >> ${OUTPUT}/r/usr/sbin/distmk
    out="$(${DISTCHK} -r ${OUTPUT}/r test 2>&1)"
    r=$?
    out="$(echo "${out}" | grep "bad checksum" | grep "/usr/sbin/distmk")"
    if [ ${r} -eq 1 -a ! -z "${out}" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

rm -rf ${TREE}
rm -rf ${META}
rm -rf ${TMPDIST}
rm -rf ${OUTPUT}

if [ "${LEAKCHECK}" = "y" ]; then
    exit 0
else
    # expected $EXPECTED_FAILS failures
    [ ${FAILS} -eq ${EXPECTED_FAILS} ] && exit 0 || exit 1
fi
