#!/bin/sh
# Verify we can get information about installed packages

CWD="$(pwd)"

title=" >>> $(basename ${0}) <<< "
line="-$(echo ${title} | sed -e 's|.|-|g')-"
echo
echo "+${line}+"
echo "|${title}|"
echo "+${line}+"
echo

LEAKCHECK_CMD=
[ "${LEAKCHECK}" = "y" ] && LEAKCHECK_CMD="valgrind --leak-check=full --show-reachable=yes "

DISTMK=$(readlink -f ../distmk)
DISTADD=$(readlink -f ../distadd)
DISTINFO=$(readlink -f ../distinfo)
TREE="$(mktemp -d)"
META="$(mktemp -d)"
OUTPUT="$(mktemp -d)"
TMPDIST="$(mktemp -d)"
FAILS=0
EXPECTED_FAILS=0

[ ! -x ${DISTMK} ] && ${MAKE} -C ../ distmk
[ ! -x ${DISTADD} ] && ${MAKE} -C ../ distadd
[ ! -x ${DISTINFO} ] && ${MAKE} -C ../ distinfo

# create a tree to package up
mkdir -p ${TREE}/usr/sbin
cp -p ${DISTMK} ${TREE}/usr/sbin/distmk

# create plist
cat << EOF > ${META}/plist
usr root(0) root(0) 0755
usr/sbin root(0) root(0) 0755
usr/sbin/distmk root(0) root(0) 0755
EOF

echo "Big giant legal file." > ${META}/legal

cat << EOF > ${META}/pkginfo
desc="Test program"
maintainer="J. Random Packager <someone@example.net>"
license=WTFPL
EOF

# create the package
if [ "${LEAKCHECK}" = "y" ]; then
    ${LEAKCHECK_CMD} \
    ${DISTMK} -t ${TREE} -p "test" -v "1.0" -r 1 -m ${META} -l ${META}/plist \
              -R peachtree -o ${OUTPUT} --tmpdir=${TMPDIST}
else
    ${DISTMK} -t ${TREE} -p "test" -v "1.0" -r 1 -m ${META} -l ${META}/plist \
              -R peachtree -o ${OUTPUT} --tmpdir=${TMPDIST} >/dev/null 2>&1
fi

PKG="test-1.0-1.$(uname -m).dist"

# list info about a package file
if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distinfo PKG' on a package file..."
    ${LEAKCHECK_CMD} ${DISTINFO} ${OUTPUT}/${PKG}
else
    echo -n "checking 'distinfo PKG' on a package file..."
    out="$(${DISTINFO} ${OUTPUT}/${PKG} 2>&1)"
    r=$?
    f1="$(echo "${out}" | awk '{ print $1; }')"
    f2="$(echo "${out}" | awk '{ print $2; }')"
    if [ ${r} -eq 0 -a "${f1}" = "peachtree" -a "${f2}" = "test-1.0-1.$(uname -m)" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# list all packages
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distinfo' to list all installed packages..."
    ${LEAKCHECK_CMD} ${DISTADD} -d -r ${OUTPUT}/r ${OUTPUT}/${PKG}
    ${LEAKCHECK_CMD} ${DISTINFO} -r ${OUTPUT}/r
else
    echo -n "checking 'distinfo' to list all installed packages..."
    ${DISTADD} -d -r ${OUTPUT}/r ${OUTPUT}/${PKG} >/dev/null 2>&1
    out="$(${DISTINFO} -r ${OUTPUT}/r 2>&1)"
    r=$?
    f1="$(echo "${out}" | awk '{ print $1; }')"
    f2="$(echo "${out}" | awk '{ print $2; }')"
    if [ ${r} -eq 0 -a "${f1}" = "peachtree" -a "${f2}" = "test-1.0-1.$(uname -m)" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# list a non-existent package
if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distinfo PKG' for a non-existent PKG..."
    ${LEAKCHECK_CMD} ${DISTINFO} -r ${OUTPUT}/r test2
else
    echo -n "checking 'distinfo PKG' for a non-existent PKG..."
    out="$(${DISTINFO} -r ${OUTPUT}/r test2 2>&1)"
    r=$?
    if [ ${r} -eq 0 -a -z "${out}" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# list package version
if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distinfo -v'..."
    ${LEAKCHECK_CMD} ${DISTINFO} -v -r ${OUTPUT}/r test
else
    echo -n "checking 'distinfo -v'..."
    out="$(${DISTINFO} -v -r ${OUTPUT}/r test 2>&1)"
    r=$?
    lines=$(echo "${out}" | wc -l)
    r1="$(echo "${out}" | grep "Name:" | cut -c17-)"
    r2="$(echo "${out}" | grep "Version:" | cut -c17-)"
    r3="$(echo "${out}" | grep "Release:" | cut -c17-)"
    if [ ${r} -eq 0 -a ${lines} -eq 3 -a "${r1}" = "test" -a "${r2}" = "1.0" -a "${r3}" = "1" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# list package maintainer
if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distinfo -m'..."
    ${LEAKCHECK_CMD} ${DISTINFO} -m -r ${OUTPUT}/r test
else
    echo -n "checking 'distinfo -m'..."
    out="$(${DISTINFO} -m -r ${OUTPUT}/r test 2>&1)"
    r=$?
    lines=$(echo "${out}" | wc -l)
    r1="$(echo "${out}" | grep "Name:" | cut -c17-)"
    r2="$(echo "${out}" | grep "Maintainer:" | cut -c17-)"
    if [ ${r} -eq 0 -a ${lines} -eq 2 -a "${r1}" = "test" -a "${r2}" = "J. Random Packager <someone@example.net>" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# list package description
if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distinfo -d'..."
    ${LEAKCHECK_CMD} ${DISTINFO} -d -r ${OUTPUT}/r test
else
    echo -n "checking 'distinfo -d'..."
    out="$(${DISTINFO} -d -r ${OUTPUT}/r test 2>&1)"
    r=$?
    lines=$(echo "${out}" | wc -l)
    r1="$(echo "${out}" | grep "Name:" | cut -c17-)"
    r2="$(echo "${out}" | tail -n 1)"
    if [ ${r} -eq 0 -a ${lines} -eq 3 -a "${r1}" = "test" -a "${r2}" = "Test program" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# list package size
if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distinfo -s'..."
    ${LEAKCHECK_CMD} ${DISTINFO} -s -r ${OUTPUT}/r test
else
    echo -n "checking 'distinfo -s'..."
    out="$(${DISTINFO} -s -r ${OUTPUT}/r test 2>&1)"
    r=$?
    lines=$(echo "${out}" | wc -l)
    r1="$(echo "${out}" | grep "Name:" | cut -c17-)"
    r2="$(echo "${out}" | grep "Installed size:" | grep "bytes")"
    if [ ${r} -eq 0 -a ${lines} -eq 2 -a "${r1}" = "test" -a ! -z "${r2}" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# list package pkgsize (on installed package)
if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distinfo -p' on an installed package..."
    ${LEAKCHECK_CMD} ${DISTINFO} -p -r ${OUTPUT}/r test
else
    echo -n "checking 'distinfo -p' on an installed package..."
    out="$(${DISTINFO} -p -r ${OUTPUT}/r test 2>&1)"
    r=$?
    lines=$(echo "${out}" | wc -l)
    r1="$(echo "${out}" | grep "Name:" | cut -c17-)"
    if [ ${r} -eq 0 -a ${lines} -eq 1 -a "${r1}" = "test" -a ! -z "${r2}" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# list package pkgsize (on a package file)
if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distinfo -p PKG' on a package file..."
    ${LEAKCHECK_CMD} ${DISTINFO} -p ${OUTPUT}/${PKG}
else
    echo -n "checking 'distinfo -p PKG' on a package file..."
    out="$(${DISTINFO} -p ${OUTPUT}/${PKG} 2>&1)"
    r=$?
    lines=$(echo "${out}" | wc -l)
    r1="$(echo "${out}" | grep "Package:" | cut -c17-)"
    r2="$(echo "${out}" | grep "Name:" | cut -c17-)"
    r3="$(echo "${out}" | grep "Package size:" | grep "bytes")"
    if [ ${r} -eq 0 -a ${lines} -eq 3 -a "$(basename ${r1})" = "${PKG}" -a "${r2}" = "test" -a ! -z "${r3}" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# list package arch
if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distinfo -a'..."
    ${LEAKCHECK_CMD} ${DISTINFO} -a -r ${OUTPUT}/r test
else
    echo -n "checking 'distinfo -a'..."
    out="$(${DISTINFO} -a -r ${OUTPUT}/r test 2>&1)"
    r=$?
    lines=$(echo "${out}" | wc -l)
    r1="$(echo "${out}" | grep "Name:" | cut -c17-)"
    r2="$(echo "${out}" | grep "Architecture:" | cut -c17-)"
    if [ ${r} -eq 0 -a ${lines} -eq 2 -a "${r1}" = "test" -a "${r2}" = "$(uname -m)" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# list package files
if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distinfo -f'..."
    ${LEAKCHECK_CMD} ${DISTINFO} -f -r ${OUTPUT}/r test
else
    echo -n "checking 'distinfo -f'..."
    out="$(${DISTINFO} -f -r ${OUTPUT}/r test 2>&1)"
    r=$?
    lines=$(echo "${out}" | wc -l)
    r1="$(echo "${out}" | grep "Name:" | cut -c17-)"
    r2="$(echo "${out}" | grep -E "^/usr$")"
    r3="$(echo "${out}" | grep -E "^/usr/sbin$")"
    r4="$(echo "${out}" | grep -E "^/usr/sbin/distmk$")"
    if [ ${r} -eq 0 -a ${lines} -eq 4 -a "${r1}" = "test" -a "${r2}" = "/usr" -a "${r3}" = "/usr/sbin" -a "${r4}" = "/usr/sbin/distmk" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# list package license
if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distinfo -l'..."
    ${LEAKCHECK_CMD} ${DISTINFO} -l -r ${OUTPUT}/r test
else
    echo -n "checking 'distinfo -l'..."
    out="$(${DISTINFO} -l -r ${OUTPUT}/r test 2>&1)"
    r=$?
    lines=$(echo "${out}" | wc -l)
    r1="$(echo "${out}" | grep "Name:" | cut -c17-)"
    r2="$(echo "${out}" | grep "License:" | cut -c17-)"
    if [ ${r} -eq 0 -a ${lines} -eq 2 -a "${r1}" = "test" -a "${r2}" = "WTFPL" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# list package legal file
if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distinfo -L'..."
    ${LEAKCHECK_CMD} ${DISTINFO} -L -r ${OUTPUT}/r test
else
    echo -n "checking 'distinfo -L'..."
    out="$(${DISTINFO} -L -r ${OUTPUT}/r test 2>&1)"
    r=$?
    lines=$(echo "${out}" | wc -l)
    r1="$(echo "${out}" | grep "Name:" | cut -c17-)"
    r2="$(echo "${out}" | tail -n 1)"
    if [ ${r} -eq 0 -a ${lines} -eq 3 -a "${r1}" = "test" -a "${r2}" = "Big giant legal file." ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# list package legal file from a package file
if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distinfo -L PKG' on a package file..."
    ${LEAKCHECK_CMD} ${DISTINFO} -L ${OUTPUT}/${PKG}
else
    echo -n "checking 'distinfo -L PKG' on a package file..."
    out="$(${DISTINFO} -L ${OUTPUT}/${PKG} 2>&1)"
    r=$?
    lines=$(echo "${out}" | wc -l)
    r1="$(echo "${out}" | grep "Name:" | cut -c17-)"
    r2="$(echo "${out}" | tail -n 1)"
    r3="$(echo "${out}" | grep "Package:" | cut -c17-)"
    if [ ${r} -eq 0 -a ${lines} -eq 4 -a "${r1}" = "test" -a "${r2}" = "Big giant legal file." -a "$(basename "${r3}")" = "${PKG}" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# list package repository name
if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distinfo -R'..."
    ${LEAKCHECK_CMD} ${DISTINFO} -R -r ${OUTPUT}/r test
else
    echo -n "checking 'distinfo -R'..."
    out="$(${DISTINFO} -R -r ${OUTPUT}/r test 2>&1)"
    r=$?
    lines=$(echo "${out}" | wc -l)
    r1="$(echo "${out}" | grep "Name:" | cut -c17-)"
    r2="$(echo "${out}" | grep "Repository:" | cut -c17-)"
    if [ ${r} -eq 0 -a ${lines} -eq 2 -a "${r1}" = "test" -a "${r2}" = "peachtree" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# is this a debug package or not?
if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distinfo -g'..."
    ${LEAKCHECK_CMD} ${DISTINFO} -g -r ${OUTPUT}/r test
else
    echo -n "checking 'distinfo -g'..."
    out="$(${DISTINFO} -g -r ${OUTPUT}/r test 2>&1)"
    r=$?
    lines=$(echo "${out}" | wc -l)
    r1="$(echo "${out}" | grep "Name:" | cut -c17-)"
    r2="$(echo "${out}" | grep "Debug package:" | cut -c17-)"
    if [ ${r} -eq 0 -a ${lines} -eq 2 -a "${r1}" = "test" -a "${r2}" = "no" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# list all package info
if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distinfo -A'..."
    ${LEAKCHECK_CMD} ${DISTINFO} -A -r ${OUTPUT}/r test
else
    echo -n "checking 'distinfo -A'..."
    out="$(${DISTINFO} -A -r ${OUTPUT}/r test 2>&1)"
    r=$?
    lines=$(echo "${out}" | wc -l)
    r1="$(echo "${out}" | grep "Name:" | cut -c17-)"
    r2="$(echo "${out}" | grep "Version:" | cut -c17-)"
    r3="$(echo "${out}" | grep "Release:" | cut -c17-)"
    r4="$(echo "${out}" | grep "License:" | cut -c17-)"
    r5="$(echo "${out}" | grep "Architecture:" | cut -c17-)"
    r6="$(echo "${out}" | grep "Maintainer:" | cut -c17-)"
    r7="$(echo "${out}" | grep "Installed size:" | grep "bytes")"
    r8="$(echo "${out}" | tail -n 2 | head -n 1)"
    if [ ${r} -eq 0 -a "${r1}" = "test" -a "${r2}" = "1.0" -a "${r3}" = "1" -a "${r4}" = "WTFPL" -a "${r5}" = "$(uname -m)" -a "${r6}" = "J. Random Packager <someone@example.net>" -a ! -z "${r7}" -a "${r8}" = "Test program" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# query for package that owns a file (expect success)
if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distinfo -q FILE' with expected success..."
    ${LEAKCHECK_CMD} ${DISTINFO} -r ${OUTPUT}/r -q /usr/sbin/distmk
else
    echo -n "checking 'distinfo -q FILE' with expected success..."
    out="$(${DISTINFO} -r ${OUTPUT}/r -q /usr/sbin/distmk 2>&1)"
    r=$?
    lines=$(echo "${out}" | wc -l)
    r1="$(echo "${out}" | cut -d ' ' -f 2)"
    if [ ${r} -eq 0 -a ${lines} -eq 1 -a "${r1}" = "$(basename ${PKG} .dist)" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# query for package that owns a file (expect failure)
if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distinfo -q FILE' with expected failure..."
    ${LEAKCHECK_CMD} ${DISTINFO} -r ${OUTPUT}/r -q /usr/sbin/thing
else
    echo -n "checking 'distinfo -q FILE' with expected failure..."
    out="$(${DISTINFO} -r ${OUTPUT}/r -q /usr/sbin/thing 2>&1)"
    r=$?
    lines=$(echo "${out}" | wc -l)
    if [ ${r} -eq 0 -a ${lines} -eq 1 -a "${out}" = "/usr/sbin/thing does not exist." ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

# list info in shell variable output (make sure this is the last test)
if [ "${LEAKCHECK}" = "y" ]; then
    echo "checking 'distinfo -A -S' for shell output..."
    ${LEAKCHECK_CMD} ${DISTINFO} -A -S -r ${OUTPUT}/r test > ${OUTPUT}/infovars 2>&1
else
    echo -n "checking 'distinfo -A -S' for shell output..."
    ${DISTINFO} -A -S -r ${OUTPUT}/r test > ${OUTPUT}/infovars 2>&1
    r=$?
    . ${OUTPUT}/infovars
    if [ ${r} -eq 0 -a "${PKG}" = "test" -a "${VER}" = "1.0" -a "${REL}" = "1" -a "${LICENSE}" = "WTFPL" -a "${REPO}" = "peachtree" -a "${ARCH}" = "$(uname -m)" -a "${MAINTAINER}" = "J. Random Packager <someone@example.net>" -a "${DESC}" = "Test program" -a ${SIZE} -gt 0 -a "${DEBUG}" = "n" ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi
fi

rm -rf ${TREE}
rm -rf ${META}
rm -rf ${TMPDIST}
rm -rf ${OUTPUT}

if [ "${LEAKCHECK}" = "y" ]; then
    exit 0
else
    # expected $EXPECTED_FAILS failures
    [ ${FAILS} -eq ${EXPECTED_FAILS} ] && exit 0 || exit 1
fi
