#!/bin/sh
# Verify we can create a package with distmk

title=" >>> $(basename ${0}) <<< "
line="-$(echo ${title} | sed -e 's|.|-|g')-"
echo
echo "+${line}+"
echo "|${title}|"
echo "+${line}+"
echo

LEAKCHECK_CMD=
[ "${LEAKCHECK}" = "y" ] && LEAKCHECK_CMD="valgrind --leak-check=full --show-reachable=yes "

DISTMK=$(readlink -f ../distmk)
TREE="$(mktemp -d)"
META="$(mktemp -d)"
OUTPUT="$(mktemp -d)"
TMPDIST="$(mktemp -d)"
FAILS=0

[ ! -x ${DISTMK} ] && ${MAKE} -C ../ distmk

# create a tree to package up
mkdir -p ${TREE}/etc
mkdir -p ${TREE}/usr/sbin
echo "System ready" > ${TREE}/etc/issue
cp -p ${DISTMK} ${TREE}/usr/sbin/distmk

# create plist
cat << EOF > ${META}/plist
etc root(0) root(0) 0755
etc/issue root(0) root(0) 0644
usr root(0) root(0) 0755
usr/sbin root(0) root(0) 0755
usr/sbin/distmk root(0) root(0) 0755
EOF

echo "Big giant legal file." > ${META}/legal

cat << EOF > ${META}/pkginfo
desc="Test program"
maintainer="J. Random Packager <someone@example.net>"
license=WTFPL
EOF

# create the package
if [ "${LEAKCHECK}" = "y" ]; then
    ${LEAKCHECK_CMD} ${DISTMK} -t ${TREE} -p "test" -v "1.0" -r 1 -m ${META} -l ${META}/plist -R peachtree -o ${OUTPUT} --tmpdir=${TMPDIST}
else
    out="$(${DISTMK} -t ${TREE} -p "test" -v "1.0" -r 1 -m ${META} -l ${META}/plist -R peachtree -o ${OUTPUT} --tmpdir=${TMPDIST} 2>&1)"
    r=$?
    echo -n "checking distmk packaging ability..."
    FAILS=$((${FAILS} + ${r}))
    [ ${r} -ne 0 ] && echo "FAILED." || echo "passed."

    # check the package contents
    pkg="$(/bin/ls -1 ${OUTPUT}/*.dist 2>/dev/null)"
    cd ${OUTPUT}
    mkdir a
    cd a
    cpio -id < ${pkg} >/dev/null 2>&1

    echo -n "checking signature..."
    if [ ! "$(cat signature 2>/dev/null)" = "Peachtree Linux PACKAGE v2.2" ]; then
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    else
        echo "passed."
    fi

    echo -n "checking bom.db contents table..."
    sqlite bom.db "SELECT path FROM CONTENTS" | sort > contents
    grep -v "^etc" ${META}/plist | cut -d ' ' -f 1 | sort > ${META}/plist.sorted
    cmp -s ${META}/plist.sorted contents
    r=$?
    FAILS=$((${FAILS} + ${r}))
    [ ${r} -eq 1 ] && echo "FAILED." || echo "passed."

    echo -n "checking etc.payload..."
    xz -dc etc.payload | tar -xf -
    if [ ! "$(cat etc/issue 2>/dev/null)" = "System ready" ]; then
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    else
        echo "passed."
    fi

    echo -n "checking reloc.payload..."
    xz -dc reloc.payload | tar -xf -
    if [ -f usr/sbin/distmk ]; then
        echo "passed."
    else
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    fi

    echo "checking digests..."
    sqlite bom.db "SELECT path,digest FROM contents WHERE digest NOT NULL" | \
    while read line ; do
        path="$(echo "$line" | cut -d '|' -f 1)"
        digest="$(echo "$line" | cut -d '|' -f 2)"

        echo -n "    $path..."
        computed="$(sha1sum $path | cut -d ' ' -f 1)"

        if [ "$digest" = "$computed" ]; then
            echo "passed."
        else
            echo "FAILED."
            FAILS=$((${FAILS} + 1))
        fi
    done

    echo -n "checking pkginfo..."
    info="$(sqlite bom.db "SELECT * FROM info")"
    name="$(echo $info | cut -d '|' -f 1)"
    version="$(echo $info | cut -d '|' -f 2)"
    release="$(echo $info | cut -d '|' -f 3)"
    desc="$(echo $info | cut -d '|' -f 8)"
    maintainer="$(echo $info | cut -d '|' -f 9)"
    license="$(echo $info | cut -d '|' -f 10)"
    repo="$(echo $info | cut -d '|' -f 5)"
    if [ ! "${name}" = "test" -o ! "${version}" = "1.0" -o ! "${release}" = "1" -o ! "${desc}" = "Test program" -o ! "${maintainer}" = "J. Random Packager <someone@example.net>" -o ! "${repo}" = "peachtree" -o ! "${license}" = "WTFPL" ]; then
        FAILS=$((${FAILS} + 1))
        echo "FAILED."
    else
        echo "passed."
    fi
fi

rm -rf ${TREE}
rm -rf ${META}
rm -rf ${TMPDIST}
rm -rf ${OUTPUT}

if [ "${LEAKCHECK}" = "y" ]; then
    exit 0
else
    # expected zero failures
    [ ${FAILS} -gt 0 ] && exit 1 || exit 0
fi
