#!/bin/sh
# Simple script to extract x509 certificates and create a JRE cacerts file.
# Adapted from:
# http://www.linuxfromscratch.org/blfs/view/svn/general/openjdk.html

# The date binary is not reliable on 32bit systems for dates after 2038
mydate() {
    local y=$(echo $1 | cut -d ' ' -f 4)
    local M=$(echo $1 | cut -d ' ' -f 1)
    local d=$(echo $1 | cut -d ' ' -f 2)
    local m

    if [ ${d} -lt 10 ]; then
        d="0${d}"
    fi

    case $M in
        Jan) m="01" ;;
        Feb) m="02" ;;
        Mar) m="03" ;;
        Apr) m="04" ;;
        May) m="05" ;;
        Jun) m="06" ;;
        Jul) m="07" ;;
        Aug) m="08" ;;
        Sep) m="09" ;;
        Oct) m="10" ;;
        Nov) m="11" ;;
        Dec) m="12" ;;
    esac

    certdate="${y}${m}${d}"
}

showhelp() {
    echo "$(basename ${0}) creates a valid cacerts file for use with IcedTea."
    echo
    echo "        -f  --cafile        The path to a file containing PEM formated CA"
    echo "                            certificates.  May not be used with -d/--cadir."
    echo "        -d  --cadir         The path to a diectory of PEM formatted CA"
    echo "                            certificates.  May not be used with -f/--cafile."
    echo "        -o  --outfile       The path to the output file."
    echo
    echo "        -k  --keytool       The path to the java keytool utility."
    echo
    echo "        -s  --openssl       The path to the openssl utility."
    echo
    echo "        -h  --help          Show this help message and exit."
    echo
}

# Initialize empty variables so that the shell does not polute the script
CAFILE=""
CADIR=""
OUTFILE=""
OPENSSL=""
KEYTOOL=""
certdate=""
date=""
today=$(date +%Y%m%d)

# Process command line arguments
if [ -z "${1}" ]; then
    showhelp
    exit 1
fi

ARGS=$(getopt -o f:d:o:k:s:h --long cafile:,cadir:,outfile:,keytool:,openssl:,help -n "mkcacerts" -- "$@")
if [ $? -ne 0 ]; then
    echo "Terminating..." >&2
    exit 1
fi

eval set -- "${ARGS}"

while true ; do
    case "${1}" in
        -f|--cafile)
            CAFILE="${2}"
            shift 2 ;;
        -d|--cadir)
            CADIR="${2}"
            shift 2 ;;
        -o|--outfile)
            OUTFILE="${2}"
            shift 2 ;;
        -k|--keytool)
            KEYTOOL="${2}"
            shift 2 ;;
        -s|--openssl)
            OPENSSL="${2}"
            shift 2 ;;
        -h|--help)
            showhelp
            exit 0 ;;
        --)
            shift
            break ;;
        *)
            showhelp
            exit 1 ;;
    esac
done

# Handle common errors
if [ -z "${CAFILE}" -a -z "${CADIR}" ]; then
    echo "ERROR!  You must provide an x509 certificate store!"
    echo "\'$(basename ${0}) --help\' for more info."
    exit 1
fi

if [ ! -z "${CAFILE}" -a ! -z "${CADIR}" ]; then
    echo "ERROR!  You cannot provide two x509 certificate stores!"
    echo "\'$(basename ${0}) --help\' for more info."
    exit 1
fi

if [ -z "${KEYTOOL}" ]; then
    echo "ERROR!  You must provide a valid keytool program!"
    echo "\'$(basename ${0}) --help\' for more info."
    exit 1
fi

if [ -z "${OPENSSL}" ]; then
    echo "ERROR!  You must provide a valid path to openssl!"
    echo "\'$(basename ${0}) --help\' for more info."
    exit 1
fi

if [ -z "${OUTFILE}" ]; then
    echo "ERROR!  You must provide a valid output file!"
    echo "\'$(basename ${0}) --help\' for more info."
    exit 1
fi

# Get on with the work

# If using a CAFILE, split it into individual files in a temp directory
if [ ! -z "${CAFILE}" ]; then
    TEMPDIR=$(mktemp -d)
    CADIR="${TEMPDIR}"

    # Get a list of staring lines for each cert
    CERTLIST=$(grep -n "^-----BEGIN" "${CAFILE}" | cut -d ':' -f 1)

    # Get a list of ending lines for each cert
    ENDCERTLIST=$(grep -n "^-----END" "${CAFILE}" | cut -d ':' -f 1)

    # Start a loop
    for certbegin in $(echo "${CERTLIST}") ; do
        for certend in $(echo "${ENDCERTLIST}") ; do
            [ "${certend}" -gt "${certbegin}" ] && break
        done

        sed -n "${certbegin},${certend}p" "${CAFILE}" > "${CADIR}/${certbegin}.pem"
        keyhash=$(${OPENSSL} x509 -noout -in "${CADIR}/${certbegin}.pem" -hash)
        echo "Generated PEM file with hash:  ${keyhash}."
    done
fi

# Write the output file
for cert in $(find "${CADIR}" -type f -name "*.pem" -o -name "*.crt") ; do
    # Make sure the certificate date is valid...
    d=$(${OPENSSL} x509 -enddate -in "${cert}" -noout | sed 's/^notAfter=//')
    mydate "${d}"

    if [ "${certdate}" -lt "${today}" ]; then
        echo "${cert} expired on ${certdate}! Skipping..."
        unset d certdate
        continue
    fi

    unset d certdate
    ls "${cert}"
    tempfile=$(mktemp)
    certbegin=$(grep -n "^-----BEGIN" "${cert}" | cut -d ':' -f 1)
    certend=$(grep -n "^-----END" "${cert}" | cut -d ':' -f 1)
    sed -n "${certbegin},${certend}p" "${cert}" > "${tempfile}"
    echo yes | env LC_ALL=C "${KEYTOOL}" -import -alias $(basename "${cert}") -keystore \
                   "${OUTFILE}" -storepass 'changeit' -file "${tempfile}"
    rm "${tempfile}"
done

[ "${TEMPDIR}x" = "x" ] || rm -rf "${TEMPDIR}"
