#!/bin/sh
#
# build.sh: Implement build stage.
#
# {{{ Apache License version 2.0
#
# Copyright 2011-2018 David Cantrell <david.l.cantrell@gmail.com>
#                     Chris Lumens <chris@bangmoney.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# }}}

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $WRKDIR               The top level work directory.
#    $WRKSRC               The extracted source directory.
#    $WRKBLD               The source build directory.
#    $WRKSYSROOT           The --sysroot tree for this build.
#    $PREFIX               default: /usr
#    $SYSCONFDIR           default: /etc
#    $BINDIR               default: /usr/bin
#    $LIBDIR               default: /usr/lib
#    $LIBEXECDIR           default: /usr/libexec
#    $DATADIR              default: /usr/share
#    $INCLUDEDIR           default: /usr/include
#    $SEPARATE_BUILD       (y/n) Use source build directories?
#    $CONFIGURE_STYLE      The method used to configure $WRKSRC for builds.
#    $PKG                  The package name.
#    $VER                  The package version.
#    $PACKAGE              The full package name.
#    $CFLAGS               C compiler flags.
#    $CXXFLAGS             C++ compiler flags.
#    $LDFLAGS              Linker flags.
#    $PYTHON               Python interpreter to use.
#    $BUILD_ENV            Environment variables passed to $(BUILD_COMMAND).
#    $BUILD_TARGET         The name of the build target (default=all).
#    $BUILD_ARGS           Extra arguments to pass to $(BUILD_COMMAND).
#    $BUILD_COMMAND        The make command.
#    $PARALLEL_JOBS        Number of make(1) jobs to run (the -j option).
#    $TOPSRC               The top level distribution source directory.
#
# For CONFIGURE_STYLE=gnu
#    $CONFIGURE_ENV        Environment variables passed to ./configure
#    $CONFIGURE_OPTS       Extra options passed to ./configure
#    $AUTORECONF           Run 'autoreconf' on the package (y/n).
# This script is called by the build system scripts and should not be called
# directly.

[ "${SKIP_BUILD}" = "y" ] && exit 0

if [ -z "${WRKDIR}" ] || [ ! -d "${WRKDIR}" ]; then
    exit 1
fi

[ -z "${PACKAGE}" ] && exit 2

if [ -z "${WRKSRC}" ] || ( [ ! -d "${WRKSRC}" ] && [ ! -d "${WRKDIR}/src" ] ); then
    exit 3
fi

CWD="$(pwd)"
LOG="${WRKDIR}/build.log"
ERR="${WRKDIR}/build.err"
JOPT=

if [ ! -z "${PARALLEL_JOBS}" ] && [ ${PARALLEL_JOBS} -gt 1 ]; then
    JOPT="-j ${PARALLEL_JOBS}"
fi

if [ -d "${WRKSRC}" ]; then
    echo ">>> Building ${PACKAGE}..."

    if [ "${SEPARATE_BUILD}" = "y" ]; then
        [ -d "${WRKBLD}" ] || mkdir -p "${WRKBLD}"
        cd "${WRKBLD}"
    else
        cd "${WRKSRC}"
    fi

    case "${CONFIGURE_STYLE}" in
        gnu)
            # Regenerate autoconf files if requested
            if [ "${AUTORECONF}" = "y" ]; then
                [ "${SEPARATE_BUILD}" = "y" ] && cd "${WRKSRC}"

                [ -d m4 ] || mkdir -p m4

                ( env ${CONFIGURE_ENV} \
                  autoreconf -v -f -i -B ${WRKSYSROOT}${DATADIR}/aclocal \
                  2>&1 1>&3 | tee ${ERR} \
                ) 3>&1 1>&2 | tee ${LOG} || exit $?

                # If there are resulting po/Makefile.in.in files, check to
                # see if it has datarootdir and if it doesn't, add it.  This
                # suppresses a harmless warning message, but makes our log
                # files nicer.
                find ${WRKSRC} -name "Makefile.in.in" | grep "po/Makefile.in.in" | \
                while read mfile ; do
                    grep -q ^datarootdir ${mfile} >/dev/null 2>&1
                    if [ $? -eq 1 ]; then
                        sed -i -e '/^datadir/a datarootdir = @datarootdir@' ${mfile}
                    fi
                done

                [ "${SEPARATE_BUILD}" = "y" ] && cd "${WRKBLD}"
            fi

            # Run ./configure
            ( env CFLAGS="${CFLAGS}" \
                  CPPFLAGS="${CFLAGS}" \
                  CXXFLAGS="${CXXFLAGS}" \
                  LDFLAGS="${LDFLAGS}" \
                  ${CONFIGURE_ENV} \
              ${WRKSRC}/configure ${CONFIGURE_OPTS} \
              2>&1 1>&3 | tee -a ${ERR} \
            ) 3>&1 1>&2 | tee -a ${LOG} || exit $?

            # Check for circular symlinks and remove if they are present.
            # This works around an issue noticed on a few packages in
            # August 2015 building on 2.9.19.  autoconf wants to create
            # these links
            for f in GNUmakefile ; do
                readlink -f "${WRKSRC}/${f}" >/dev/null 2>&1
                [ $? -eq 1 ] && rm -f "${WRKSRC}/${f}"
            done

            # Run make
            ( env ${BUILD_ENV} \
              ${BUILD_COMMAND} ${JOPT} ${BUILD_TARGET} ${BUILD_ARGS} \
              2>&1 1>&3 | tee -a ${ERR} \
            ) 3>&1 1>&2 | tee -a ${LOG} || exit $?

            ;;
        noconfigure|manual)
            if [ "${CONFIGURE_STYLE}" = "manual" ]; then
                echo "*** WARNING: CONFIGURE_STYLE=manual is deprecated, use CONFIGURE_STYLE=noconfigure"
            fi

            ( env ${BUILD_ENV} ${BUILD_COMMAND} ${JOPT} ${BUILD_TARGET} ${BUILD_ARGS} \
              2>&1 1>&3 | tee ${ERR} \
            ) 3>&1 1>&2 | tee ${LOG} || exit $?

            ;;
        haskell)
            cp ${TOPSRC}/base/haskell/haskell.mk ${WRKSRC}/Makefile
            cd ${WRKSRC}

            ( env ${BUILD_ENV} \
                  PREFIX="${PREFIX}" \
                  BINDIR="${BINDIR}" \
                  LIBDIR="${LIBDIR}" \
                  DATADIR="${DATADIR}" \
                  PACKAGE="${PACKAGE}" \
                  CONFIGURE_OPTS="${CONFIGURE_OPTS}" \
              ${BUILD_COMMAND} ${JOPT} build ${BUILD_ARGS} \
              2>&1 1>&3 | tee ${ERR} \
            ) 3>&1 1>&2 | tee ${LOG} || exit $?

            ;;
        cmake)
            ( env ${BUILD_ENV} \
                  CMAKE_C_FLAGS="${CFLAGS}" \
                  CMAKE_CXX_FLAGS="${CXXFLAGS}" \
                  CMAKE_C_LINK_FLAGS="${LDFLAGS}" \
                  CMAKE_CXX_LINK_FLAGS="${LDFLAGS}" \
              cmake ${CONFIGURE_OPTS} ${WRKSRC} &&
              env ${BUILD_ENV} ${BUILD_COMMAND} ${JOPT} ${BUILD_TARGET} ${BUILD_ARGS} \
              2>&1 1>&3 | tee ${ERR} \
            ) 3>&1 1>&2 | tee ${LOG} || exit $?

            ;;
        python)
            ( env ${BUILD_ENV} ${PYTHON} setup.py \
              ${BUILD_TARGET} ${CONFIGURE_OPTS} ${BUILD_ARGS} \
              2>&1 1>&3 | tee ${ERR} \
            ) 3>&1 1>&2 | tee ${LOG} || exit $?

            ;;
        qmake)
            # qmake is a terrible program when it comes to integrating it
            # with build systems.  We need to write a qt.conf file that
            # defines the Qt prefix to the sysroot location.  This file
            # needs to live in the same directory as qmake, which is the
            # /usr/bin directory in the sysroot.  Dumb, right?
            #
            # Figuring out these variable names was not trivial.  I finally
            # came across this page:
            # https://doc.qt.io/qt-5/qt-conf.html

            qtconf="${WRKSYSROOT}${BINDIR}/qt.conf"
            echo "[Paths]" > ${qtconf}
            echo "Sysroot = ${WRKSYSROOT}" >> ${qtconf}
            echo "Prefix = ${WRKSYSROOT}${X11_LIBDIR}/qt" >> ${qtconf}
            echo "Headers = ${WRKSYSROOT}${X11_INCLUDEDIR}/qt" >> ${qtconf}
            echo "Documentation = ${WRKSYSROOT}${DATADIR}/doc/qt" >> ${qtconf}
            echo "Libraries = ${WRKSYSROOT}${X11_LIBDIR}" >> ${qtconf}
            echo "LibraryExecutables = ${WRKSYSROOT}${X11_LIBEXECDIR}/qt" >> ${qtconf}
            echo "Binaries = ${WRKSYSROOT}${X11_BINDIR}" >> ${qtconf}
            echo "Settings = ${WRKSYSROOT}${SYSCONFDIR}/qt" >> ${qtconf}

            [ -f "${WRKSYSROOT}${X11_BINDIR}/qmake" ] && QMAKE="${WRKSYSROOT}${X11_BINDIR}/qmake" || QMAKE="qmake"

            # Run qmake, which is sort of like configure.
            ( env ${BUILD_ENV} \
                  ${CONFIGURE_ENV} \
              ${QMAKE} ${CONFIGURE_OPTS} \
              2>&1 1>&3 | tee ${ERR} \
            ) 3>&1 1>&2 | tee ${LOG} || exit $?

            # Now run 'make'
            ( env ${BUILD_ENV} ${BUILD_COMMAND} ${JOPT} ${BUILD_TARGET} ${BUILD_ARGS} \
              2>&1 1>&3 | tee -a ${ERR} \
            ) 3>&1 1>&2 | tee -a ${LOG} || exit $?

            ;;
        meson)
            ( cd "${WRKSRC}"
              env ${CONFIGURE_ENV} meson ${CONFIGURE_OPTS} ${WRKBLD} &&
              env ${BUILD_ENV} ninja -C ${WRKBLD} ${BUILD_TARGET} ${BUILD_ARGS} \
              2>&1 1>&3 | tee ${ERR} \
            ) 3>&1 1>&2 | tee ${LOG} || exit $?

            ;;
        cargo)
            ( cd "${WRKBLD}"
              mkdir -p .cargo
              env ${CONFIGURE_ENV} cargo vendor -v > .cargo/config
              env ${BUILD_ENV} cargo build -v --release --all-targets ${BUILD_ARGS} \
              2>&1 1>&3 | tee ${ERR} \
            ) 3>&1 1>&2 | tee ${LOG} || exit $?

            ;;
        *)
            echo "*** Do not understand CONFIGURE_STYLE=${CONFIGURE_STYLE}"
            exit 4
            ;;
    esac
fi

[ -s "${LOG}" ] || rm -f "${LOG}"
[ -s "${ERR}" ] || rm -f "${ERR}"
