#!/bin/zsh
#
# buildreqs.sh: A new-style way of getting BUILD_DEPENDS for a package.
#               This script gets the direct build requirements for the
#               package being built, and then recursively checks the
#               build requirements of all those, building up a complete
#               set of everything needed.  The results are printed in
#               sorted order on stdout.
#
# {{{ Apache License version 2.0
#
# Copyright 2013-2018 Chris Lumens <chris@bangmoney.org>
#                     David Cantrell <david.l.cantrell@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# }}}

zmodload zsh/mapfile
set -o pipefail

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 1
fi

if [ -z "${WRKDIR}" ] || [ ! -d "${WRKDIR}" ]; then
    exit 2
fi

[ -z "${TARGET_MACH}" ] && exit 3

CWD="$(pwd)"

# No build dependencies - not a fatal error
[ ! -f "${CWD}/buildreqs" ] && exit 0

TMPDIR="$(mktemp -d)"

typeset -a worklist
typeset -a packages

# Helper function to see if a dependency is in the worklist
_contains() {
    dep="${1}"
    [ -z "${dep}" ] && return 1

    for d in $(echo ${worklist[*]}) ; do
        [ "${d}" = "${dep}" ] && return 0
    done

    return 1
}

# Helper function to iterate each package's buildreqs.
gather_buildreqs() {
    CURRENT="${1}"
    BUILDREQS="${TOPSRC}/${CURRENT}/buildreqs"

    while read dep ; do
        if ! _contains ${dep} ; then
            if [ "${CWD}" = "${TOPSRC}/${dep}" ]; then
                echo >&2
                echo "*** Circular dependency found!" >&2
                echo "*** ${CURRENT} requires $(echo "${CWD}" | sed -e "s|${TOPSRC}/||g")" >&2
                echo >&2
                return 2
            fi

            mach="$(make --no-print-directory -C "${TOPSRC}/${dep}" show var=TARGET_MACH)"
            if [ "${mach}" = "${TARGET_MACH}" ] || [ "${mach}" = "noarch" ]; then
                echo "${dep}"
            fi
        fi
    done < "${BUILDREQS}"
}

# Start the work list with all the direct build dependencies of the package
# we're building.
worklist=("${(f)${mapfile[${CWD}/buildreqs]}}")
buildreqs=(${worklist})

while true; do
    # Is the work list empty?  If so, we're done and the contents of the
    # packages list is everything that needs to be installed (potentially
    # with lots of dupes, though).
    len=${#worklist}
    if [ -z "${len}" ] || [ ${len} -eq 0 ]; then
        break
    fi

    # Handle any blank entries
    if [ -z ${worklist[1]} ]; then
        worklist[1]=()
        continue
    fi

    # Skip reqs that do not apply to this architecture nor noarch
    mach="$(make --no-print-directory -C "${TOPSRC}/${worklist[1]}" show var=TARGET_MACH)"
    if [ ! "${mach}" = "${TARGET_MACH}" ] && [ ! "${mach}" = "noarch" ]; then
        worklist[1]=()
        continue
    fi

    # Add the first package in the work list to the packages list.
    packages+=("${worklist[1]}")

    # Then add all the build dependencies of that package to the end of the
    # work list.  Note that not all packages have build dependencies.  eglibc,
    # for example...
    if [ -f ${TOPSRC}/${worklist[1]}/buildreqs ]; then
        gather_buildreqs ${worklist[1]} > ${TMPDIR}/deps
        RETVAL=$?

        if [ ${RETVAL} -ne 0 ]; then
            rm -rf ${TMPDIR}
            exit ${RETVAL}
        fi

        # reduce the package's buildreqs list to exclude implicit deps
        i=1
        while [ ${i} -le ${#buildreqs} ]; do
            grep -q "^${buildreqs[$i]}$" ${TMPDIR}/deps >/dev/null 2>&1
            [ $? -eq 0 ] && buildreqs[$i]=()
            i=$(($i + 1))
        done

        worklist+=("${(f)${mapfile[${TMPDIR}/deps]}}")
    fi

    # And remove the first element of the work list.
    worklist[1]=()
done

rm -rf ${TMPDIR}

# Our packages list now probably has lots of dupes in it.  Let's filter those
# out so we don't try to install the same packages over and over again.  Once
# again, it's probably glibc that's going to be here a dozen times.
if [ "${REDUCE}" = "y" ]; then
    ( IFS=$'\n' ; echo "${buildreqs[*]}" ) | sort -u > ${CWD}/buildreqs
else
    ( IFS=$'\n' ; echo "${packages[*]}" ) | sort -u
fi
