#!/bin/sh
#
# check-upstream.sh: Implement check-upstream target.
#
# {{{ Apache License version 2.0
#
# Copyright 2016-2018 David Cantrell <david.l.cantrell@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# }}}

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $TOPSRC               The top level project directory.
# Variables provided for optional purposes when an upstream file is
# not present in the source tree or is missing values in order to take
# the build system defaults:
#    $PKG                  The package name.
#    $VER                  The package version.
#    $UPSTREAM_URLS        The list of download URLs.
#    $UPSTREAM_GITREPO     The upstream project's git source repository.
#    $UPSTREAM_REGEXP      The version checking regular expression.
#    $UPSTREAM_PREFIX      The prefix trimming regular expression.
#    $UPSTREAM_SUFFIX      The suffix trimming regular expression.
#    $HTML_OUTPUT          (y/n) Set to y to output an HTML page.
# This script is called by the build system scripts and should not be called
# directly.

# Gather all version information.
find_latest() {
    ret="${UPSTREAMS}/ret"
    tmpfile="$(mktemp -p ${UPSTREAMS})"

    [ -f "${ret}" ] && rm -f "${ret}"

    # The default is always just checking the download URLs, so check any
    # optional methods first.
    if [ ! -z "${gitrepo}" ]; then
        # Check a git source repo tags for latest release
        git ls-remote --tags "${gitrepo}" | grep -E "${regexp}" > "${tmpfile}"
    elif [ ! -z "${urls}" ]; then
        # Check one or more HTTP, HTTPS, or FTP sites for the latest version.
        for site in ${urls} ; do
            curlout="$(mktemp -p ${UPSTREAMS})"
            curl_opts="-A Mozilla -s -f -L --retry 3 --ftp-pasv --stderr /dev/null"
            protocol="$(echo "${site}" | cut -d ':' -f 1)"

            if [ "${protocol}" = "ftp" ]; then
                # make sure the url ends with a slash, otherwise curl gets confused
                trail="$(echo "${site}" | rev | cut -c1)"
                [ "${trail}" = "/" ] || site="${site}/"

                # force an FTP NLST command to get filenames only
                curl_opts="${curl_opts} -l"
            elif [ "${protocol}" = "http" ] || [ "${protocol}" = "https" ]; then
                # perform an insecure download with curl (may want to change this
                # in the future?)
                curl_opts="${curl_opts} -k"
            else
                echo "*** ERROR: Unsupported upstream URL protocol for ${PKGPATH}: ${protocol}" >&2
                continue
            fi

            # collect upstream info and strip HTML tags
            curl ${curl_opts} -o ${curlout} ${site} 2>/dev/null
            t="$(file -b --mime-type ${curlout})"

            if [ "${t}" = "application/x-gzip" ]; then
                # some web servers give you gzip compressed web pages
                mv "${curlout}" "${curlout}.gz"
                gzip -d "${curlout}.gz"
            fi

            sed -r 's/<[^>]*>//g' ${curlout} | \
                sed -r 's/<[^>]*>$//g' | \
                tr -s ' ' | \
                grep -vE '^[ \t]*$' | \

            # grab all lines that match the upstream regular expression
            grep -E "${regexp}" > "${tmpfile}"
            rm -f "${curlout}"
        done
    fi

    # if we an exclude expression, filter out those lines
    if [ ! -z "${exclude}" ]; then
        grep -vE "${exclude}" "${tmpfile}" > "${tmpfile}.new"
        mv "${tmpfile}.new" "${tmpfile}"
    fi

    # trim the prefix and suffix
    if [ ! -z "${prefix}" ]; then
        echo "${prefix}" | grep -q '/' >/dev/null 2>&1

        if [ $? -eq 0 ]; then
            while read line ; do
                echo "${line}" | sed -r "s#^${prefix}##g"
            done < "${tmpfile}" > "${tmpfile}.new"
        else
            while read line ; do
                echo "${line}" | sed -r "s/^${prefix}//g"
            done < "${tmpfile}" > "${tmpfile}.new"
        fi

        mv "${tmpfile}.new" "${tmpfile}"
    fi

    if [ ! -z "${suffix}" ]; then
        echo "${suffix}" | grep -q '/' >/dev/null 2>&1

        if [ $? -eq 0 ]; then
            while read line ; do
                echo "${line}" | sed -r "s#${suffix}\$##g"
            done < "${tmpfile}" > "${tmpfile}.new"
        else
            while read line ; do
                echo "${line}" | sed -r "s/${suffix}$//g"
            done < "${tmpfile}" > "${tmpfile}.new"
        fi

        mv "${tmpfile}.new" "${tmpfile}"
    fi

    # skip odd minor versions (common open source numbering system to
    # indicate unstable releases).
    if [ ! -z "${skip_odd}" ] && [ ${skip_odd} -eq 1 ]; then
        while read line ; do
            parts=$(echo "${line}" | tr '.' '\n' | wc -l)
            [ ${parts} -le 1 ] && echo "${line}"
            minor="$(echo "${line}" | cut -d '.' -f 2)"
            isodd=$(expr ${minor} % 2 2>/dev/null)
            [ $? -eq 2 ] && continue
            [ -z "${isodd}" ] && continue
            [ ${isodd} -eq 0 ] && echo "${line}"
        done < "${tmpfile}" > "${tmpfile}.new"
        mv "${tmpfile}.new" "${tmpfile}"
    fi

    # save these versions
    cat "${tmpfile}" >> "${ret}"

    # remove duplicates, natural sort, and get latest
    latest="$(sort "${ret}" | uniq | sort -V | tail -n 1)"

    # clean up
    rm -f "${ret}"
    rm -f "${tmpfile}"

    echo "${latest}"
}

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 1
fi

CWD="$(pwd)"
UPSTREAMS="$(mktemp -d)"
UPSTREAM_FILE="${CWD}/upstream"
PKGPATH="$(echo "${CWD}" | sed -e "s|${TOPSRC}/||g")"

# create a directory to hold transformed upstream file fields
[ -d "${UPSTREAMS}" ] || mkdir -p "${UPSTREAMS}"

if [ -f "${UPSTREAM_FILE}" ]; then
    # transform upstream file to shell scripts
    output=
    have_ver=0
    have_urls=0
    have_gitrepo=0
    have_regexp=0
    have_prefix=0
    have_suffix=0

    while read line ; do
        # skip blank lines and comments
        [ -z "${line}" ] && continue
        [ "$(echo "${line}" | cut -c1)" = "#" ] && continue

        # do we have a new block to start?
        start="$(echo "${line}" | cut -c1)"
        end="$(echo "${line}" | rev | cut -c1)"

        if [ "${start}" = "[" ] && [ "${end}" = "]" ]; then
            # write out any missing values for the previous block
            if [ ! -z "${output}" ]; then
                [ ${have_ver} -eq 0 ] && echo "ver=\"${VER}\"" >> "${output}"
                [ ${have_urls} -eq 0 ] && echo "urls=\"${UPSTREAM_URLS}\"" >> "${output}"
                [ ${have_gitrepo} -eq 0 ] && echo "gitrepo=\"${UPSTREAM_GITREPO}\"" >> "${output}"
                [ ${have_regexp} -eq 0 ] && echo "regexp=\"${UPSTREAM_REGEXP}\"" >> "${output}"
                [ ${have_prefix} -eq 0 ] && echo "prefix=\"${UPSTREAM_PREFIX}\"" >> "${output}"
                [ ${have_suffix} -eq 0 ] && echo "suffix=\"${UPSTREAM_SUFFIX}\"" >> "${output}"
            fi

            # reset checks for default overrides
            have_ver=0
            have_urls=0
            have_gitrepo=0
            have_regexp=0
            have_prefix=0
            have_suffix=0

            # new block in the upstream file
            label="$(echo "${line}" | cut -d '[' -f 2 | cut -d ']' -f 1)"
            output="${UPSTREAMS}/$(echo "${label}" | tr ' ' '_').sh"
            echo "label=\"${label}\"" > "${output}"
            continue
        fi

        # lines in each block are in 'key: value' format
        key="$(echo "${line}" | cut -d ' ' -f 1 | cut -d ':' -f 1)"
        value="$(echo "${line}" | cut -d ' ' -f 2)"

        # convert booleans to ints for later use
        if [ "${key}" = "no_upstream" ] || [ "${key}" = "skip_odd" ]; then
            case ${value} in
                [Yy][Ee][Ss]|Y|y|[Tt][Rr][Uu][Ee]|T|t)
                    value="1" ;;
                [Nn][Oo]|[Ff][Aa][Ll][Ss][Ee]|N|n|F|f)
                    value="0" ;;
            esac
        fi

        # write out the line
        echo "$key=\"$value\"" >> "${output}"

        # note any default overrides we had
        [ "${key}" = "ver" ] && have_ver=1
        [ "${key}" = "urls" ] && have_urls=1
        [ "${key}" = "gitrepo" ] && have_gitrepo=1
        [ "${key}" = "regexp" ] && have_regexp=1
        [ "${key}" = "prefix" ] && have_prefix=1
        [ "${key}" = "suffix" ] && have_suffix=1
    done < "${UPSTREAM_FILE}"

    # write out any missing values for the previous block
    # this catches upstream files with one block, the loop
    # above handles additional blocks in the upstream file
    if [ ! -z "${output}" ]; then
        [ ${have_ver} -eq 0 ] && echo "ver=\"${VER}\"" >> "${output}"
        [ ${have_urls} -eq 0 ] && echo "urls=\"${UPSTREAM_URLS}\"" >> "${output}"
        [ ${have_gitrepo} -eq 0 ] && echo "gitrepo=\"${UPSTREAM_GITREPO}\"" >> "${output}"
        [ ${have_regexp} -eq 0 ] && echo "regexp=\"${UPSTREAM_REGEXP}\"" >> "${output}"
        [ ${have_prefix} -eq 0 ] && echo "prefix=\"${UPSTREAM_PREFIX}\"" >> "${output}"
        [ ${have_suffix} -eq 0 ] && echo "suffix=\"${UPSTREAM_SUFFIX}\"" >> "${output}"
    fi
else
    # default case for packages without an 'upstream' file
    output="${UPSTREAMS}/${PKG}.sh"
    echo "label=\"${PKG}\"" >> "${output}"
    echo "ver=\"${VER}\"" >> "${output}"
    echo "urls=\"${URLS}\"" >> "${output}"
    echo "gitrepo=\"${UPSTREAM_GITREPO}\"" >> "${output}"
    echo "regexp=\"${UPSTREAM_REGEXP}\"" >> "${output}"
    echo "prefix=\"${UPSTREAM_PREFIX}\"" >> "${output}"
    echo "suffix=\"${UPSTREAM_SUFFIX}\"" >> "${output}"
fi

# iterate over each upstream source
for upstream in ${UPSTREAMS}/*.sh ; do
    # optional booleans that may be overridden by the upstream data
    no_upstream=0
    skip_odd=0

    # Reset variables before each section runs
    label=
    ver=
    urls=
    gitrepo=
    regexp=
    prefix=
    suffix=
    exclude=

    # source in this block's values
    . "${upstream}"

    # some packages have no upstream source, we are it
    if [ ! -z "${no_upstream}" ] && [ ${no_upstream} -eq 1 ]; then
        continue
    fi

    # look for the latest upstream version
    latest="$(find_latest)"
    if [ -z "${latest}" ]; then
        if [ "${HTML_OUTPUT}" = "y" ]; then
            echo "<tr><td>${PKGPATH}</td><td style=\"background-color: red;\"><p style=\"color:white\">UNKNOWN</p></td><td>${label}</td></tr>"
        else
            echo "Unable to determine the latest version of ${label} for ${PKGPATH}"
        fi
    elif [ ! "${latest}" = "${ver}" ]; then
        if [ "${HTML_OUTPUT}" = "y" ]; then
            echo "<tr><td>${PKGPATH}</td><td>${latest}</td><td>${label}</td></tr>"
        else
            echo "${PKGPATH} is at ${latest} (${label})"
        fi
    fi
done

rm -rf "${UPSTREAMS}"
