#!/bin/sh
#
# checksum.sh: Verify checksums of $DISTFILES
#
# {{{ Apache License version 2.0
#
# Copyright 2011-2017 David Cantrell <david.l.cantrell@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# }}}

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $WRKDIR               The top level work directory.
#    $PKGDISTFILES         The package specific distfiles subdirectory.
# This script is called by the build system scripts and should not be called
# directly.

if [ -z "${WRKDIR}" ] || [ ! -d "${WRKDIR}" ]; then
    echo "*** WRKDIR is empty in $(basename $0)." >&2
    exit 1
fi

CWD="$(pwd)"

if [ -f "${WRKDIR}/archives" ]; then
    if [ -z "${PKGDISTFILES}" ] || [ ! -d "${PKGDISTFILES}" ]; then
        echo "*** Missing package distfiles, verify the 'fetch' target worked." >&2
        exit 2
    fi

    grep -E -v "^#" "${WRKDIR}/archives" | grep -E -v "^$" | \
    while read first second third ; do
        if [ -z "${third}" ]; then
            # archives file is in openssl(1) dgst format
            checkcmd="openssl"
            checkopts="dgst -$(echo ${first} | cut -d '(' -f 1 | tr '[A-Z]' '[a-z]')"
        else
            # archives file is in legacy Peachtree Linux format, use GNU coreutils
            checkcmd="$(echo ${first} | tr '[A-Z]' '[a-z]')sum"
        fi

        ${checkcmd} --version >/dev/null 2>&1
        if [ $? -eq 127 ]; then
            echo "*** Unable to find ${checkcmd} program." >&2
            exit 3
        fi

        if [ -z "${third}" ]; then
            # archives file is in openssl(1) dgst format
            filename="$(echo ${first} | cut -d '(' -f 2 | cut -d ')' -f 1)"
            computed="$(${checkcmd} ${checkopts} ${PKGDISTFILES}/${filename} | cut -d ' ' -f 2)"
        else
            # archives file is in legacy Peachtree Linux format
            filename="${third}"
            computed="$(${checkcmd} "${PKGDISTFILES}/${third}" | cut -d ' ' -f 1)"
        fi

        if [ -f "${PKGDISTFILES}/${filename}" ]; then
            echo -n ">>> Checksumming ${filename}..."

            if [ "${computed}" = "${second}" ]; then
                echo "ok."
            else
                echo "FAILED."
                rm -f "${PKGDISTFILES}/${filename}"
            fi
        fi
    done
fi
