#!/bin/sh
#
# extract.sh: Implement extract stage.
#
# {{{ Apache License version 2.0
#
# Copyright 2011-2017 David Cantrell <david.l.cantrell@gmail.com>
#                     Chris Lumens <chris@bangmoney.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# }}}

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $PKGDISTFILES            The package specifc distfiles subdirectory.
#    $WRKDIR                  The top level work directory.
#    $WRKDIST                 The name of the extracted source directory.
#    $GPG                     The gpg command and common options to use.
#    $KEYRING                 The full path to the keyring file to use.
#    $TRUSTDB                 The full path to the GPG trust database file.
#    $GPG_SIGNATURE_ENDINGS   File endings for the GPG signatures.
# This script is called by the build system scripts and should not be called
# directly.

if [ -z "${WRKDIR}" ] || [ ! -d "${WRKDIR}" ]; then
    exit 1
fi

if [ -z "${PKGDISTFILES}" ] || [ ! -d "${PKGDISTFILES}" ] && [ ! -d "src" ]; then
    exit 0
fi

if [ ! -z "${GPG}" ] && [ -z "${GPG_SIGNATURE_ENDINGS}" ]; then
    exit 3
fi

if [ ! -z "${KEYRING}" ] && [ -z "${GPG_SIGNATURE_ENDINGS}" ]; then
    exit 4
fi

if [ ! -z "${TRUSTDB}" ] && [ -z "${GPG_SIGNATURE_ENDINGS}" ]; then
    exit 5
fi

CWD="$(pwd)"
GPGCMD="${GPG} --keyring ${KEYRING} --trustdb-name ${TRUSTDB}"
RMLIST="${WRKDIR}/intermediates"
VERIFY="${WRKDIR}/verify"
DO_VERIFY=0

# Handle most common file endings for compressed archives.
newarchive_name() {
    archive="$(basename "${1}")"
    mimetype="${2}"

    if [ -z "${archive}" ] && [ -z "${mimetype}" ]; then
        return
    fi

    case "${mimetype}" in
        application/x-gzip)
            newarchive="$(basename "${archive}" .gz)" ;;
        application/x-bzip2)
            newarchive="$(basename "${archive}" .bz2)" ;;
        application/x-xz)
            newarchive="$(basename "${archive}" .xz)" ;;
        *)
            newarchive="$(basename "${archive}")" ;;
    esac

    # Handle annoying DOS 8.3ish style file endings.
    # NOTE:  We're making assumptions about file type here.
    if [ "${newarchive}" = "${archive}" ]; then
        ext="$(echo "${archive}" | rev | cut -d '.' -f 1 | rev)"

        case "${ext}" in
            tgz)
                newarchive="$(basename "${archive}" .tgz).tar" ;;
            tbz2)
                newarchive="$(basename "${archive}" .tbz2).tar" ;;
            txz)
                newarchive="$(basename "${archive}" .txz).tar" ;;
            crate)
                newarchive="$(basename "${archive}" .crate).tar" ;;
            *)
                newarchive="$(basename "${archive}")" ;;
        esac
    fi

    echo "${newarchive}"
}

# Remove intermediate archive files, but only if they are in WRKDIR.
clean_old_archive() {
    archive="$(basename "${1}")"

    if [ -f "${WRKDIR}/${archive}" ]; then
        # If there are signed source archives, keep all the intermediates
        # and remove them all at the end.
        if [ -f "${KEYRING}" ]; then
            echo "${WRKDIR}/${archive}" >> "${RMLIST}"
        else
            rm -f "${WRKDIR}/${archive}"
        fi
    fi
}

# Handle unpacking on gzip, bzip2, xz, and other files.  The command
# specified must support the -d (for decompress) and -c (send output
# to stdout) options.
# Arguments:
#     $1    The name of the variable in the caller's space to put
#           the name of the new archive for the loop.  This is a
#           string that matches the variable name without '$' at the
#           front.  Think of it as a pointer from a C function?
#     $2    The extract command to run (e.g., gzip).
#     $3    The file name to extract.
#     $4    The MIME type of the file we are extracting.
extract_z() {
    local _ret=${1}
    shift

    cmd="${1}"
    archive="${2}"
    t="$3"
    newarchive="${WRKDIR}/$(newarchive_name "${archive}" "${t}")"

    if [ ! -f "${newarchive}" ]; then
        echo ">>> Extracting: ${cmd} -dc ${short}"
        ${cmd} -dc "${archive}" > "${newarchive}"
        clean_old_archive "${archive}"
    fi

    eval "${_ret}=\${newarchive}"
}

# Handle unpacking a ZIP file.
extract_zip() {
    [ ${DO_VERIFY} -eq 1 ] && return 0

    short="${1}"
    archive="${2}"

    echo ">>> Extracting: unzip ${short}"

    # Determine if the zip file is a "zip bomb", one that
    # just extracts all files to the current directory.  If
    # it is, make sure we create a subdirectory to extract
    # everything to.
    zipbomb="$(unzip -l "${archive}" 2>/dev/null | head -n 4 | tail -n 1 | rev | cut -d ' ' -f 1 | rev | cut -d '/' -f 1)"

    # Default extract location.
    extractdir="${WRKDIR}"

    if [ ! "${zipbomb}" = "${WRKDIST}" ]; then
        subdir="$(basename "${archive}" .zip)"
        mkdir -p ${WRKDIR}/${subdir}
        extractdir="${WRKDIR}/${subdir}"
    fi

    unzip -q -d "${extractdir}" "${archive}" || rm -f "${archive}"
    clean_old_archive "${archive}"
    echo "${extractdir}"
}

verify_intermediates() {
    [ ${DO_VERIFY} -eq 1 ] || return 0

    archive="${1}"

    if [ -f "${VERIFY}" ]; then
        basearchive="$(basename "${archive}")"
        grep -q "^${basearchive}=" "${VERIFY}" >/dev/null 2>&1

        if [ $? -eq 0 ]; then
            signature="$(grep "^${basearchive}=" "${VERIFY}" | cut -d '=' -f 2)"
            ${GPGCMD} --verify ${signature} ${archive}

            # gpg verification failed, abort now
            if [ ${?} -ne 0 ]; then
                rm -f ${KEYRING}~
                exit 6
            fi
        fi
    fi
}

verify_signature() {
    [ ${DO_VERIFY} -eq 1 ] || return 0

    signature="${1}"
    ending="${2}"

    archive="$(basename ${signature} ${ending})"

    if [ -f "${PKGDISTFILES}/${archive}" ]; then
        ${GPGCMD} --verify ${signature} ${PKGDISTFILES}/${archive}

        # gpg verification failed, abort now
        if [ ${?} -ne 0 ]; then
            rm -f ${KEYRING}~
            exit 7
        fi
    else
        # This is a signature that matches a later extraction
        # point of the archive.  For example, the kernel is
        # distributed as a .tar.xz, but the signature is only
        # on the .tar file. Save these for later loop iterations.
        echo "${archive}=${signature}" >> "${VERIFY}"
        sort "${VERIFY}" | uniq > "${VERIFY}.new"
        mv "${VERIFY}.new" "${VERIFY}"
    fi
}

run_extract() {
    if [ ${DO_VERIFY} -eq 1 ] && [ ! -f "${KEYRING}" ]; then
        return 0
    fi

    if [ -f "${WRKDIR}/archives" ]; then
        if [ -z "${PKGDISTFILES}" ] || [ ! -d "${PKGDISTFILES}" ]; then
            exit 8
        fi

        grep -vE "(^#|^$)" "${WRKDIR}/archives" | \
        while read first second third ; do
            if [ -z "${third}" ]; then
                # archives file is in openssl(1) dgst format
                archive="$(echo ${first} | cut -d '(' -f 2 | cut -d ')' -f 1)"
            else
                # archives file is in legacy Peachtree Linux format
                archive="${third}"
            fi

            archive="$(readlink -f "${PKGDISTFILES}/${archive}")"
            verify_intermediates "${archive}"

            while [ 1 ]; do
                [ -f "${archive}" ] || break
                t="$(file -b --mime-type "${archive}")"
                short="$(basename "${archive}")"

                case "${t}" in
                    application/x-gzip)
                        extract_z archive "gzip" "${archive}" "${t}"
                        ;;
                    application/x-bzip2)
                        extract_z archive "bzip2" "${archive}" "${t}"
                        ;;
                    application/x-xz)
                        extract_z archive "xz" "${archive}" "${t}"
                        ;;
                    application/x-tar)
                        [ ${DO_VERIFY} -eq 1 ] && break

                        echo ">>> Extracting: tar -xf ${short}"

                        # Determine if the tar file is a "tar bomb", one that
                        # just extracts all files to the current directory.  If
                        # it is, make sure we create a subdirectory to extract
                        # everything to.
                        tarbomb="$(tar -tvf "${archive}" 2>/dev/null | head -n 1 | cut -c-1)"

                        # Add in an additional check to see if the first part of
                        # file path equals $WRKDIST.  Crude check.
                        subdir="$(tar -tvf "${archive}" 2>/dev/null | head -n 1 | rev | cut -d ' ' -f 1 | rev | cut -d '/' -f 1)"

                        # Default extract location.
                        extractdir="${WRKDIR}"

                        # Safety name.  For packages with multiple source archives, we
                        # will not match WRKDIST for each archive.  We need to figure
                        # that out before performing the tarbomb test.
                        safety="$(basename "${archive}" .tar)"
                        echo "${safety}" | grep -q -E "^${WRKDIST}" >/dev/null 2>&1
                        iswrkdist=${?}

                        if [ ! "${tarbomb}" = "d" ] && \
                           ( ( [ ${iswrkdist} -eq 0 ] && [ ! "${subdir}" = "${WRKDIST}" ] ) || \
                             ( [ ${iswrkdist} -eq 1 ] && [ ! "${subdir}" = "${safety}" ] ) ); then
                            extractdir="${WRKDIR}/${safety}"
                            mkdir -p "${extractdir}"
                        fi

                        tar -C "${extractdir}" -xf "${archive}" || rm -f "${archive}"
                        clean_old_archive "${archive}"
                        archive="${extractdir}"
                        ;;
                    application/zip)
                        archive="$(extract_zip "${short}" "${archive}")"
                        ;;
                    application/octet-stream|binary)
                        ending="$(echo "${archive}" | rev | cut -d '.' -f 1 | rev | tr [A-Z] [a-z])"

                        if [ "${ending}" = "bz2" ]; then
                            # Incorrectly detected bzip2 compressed archive
                            extract_z archive "bzip2" "${archive}" "${t}"
                        elif [ "$ending" = "xz" ]; then
                            # Incorrectly detected xz compressed archive
                            extract_z archive "xz" "${archive}" "${t}"
                        elif [ "$ending" = "gz" ]; then
                            # Incorrectly detected gzip compressed archive
                            extract_z archive "gzip" "${archive}" "${t}"
                        elif [ "${ending}" = "sign" ] || [ "${ending}" = "sig" ]; then
                            if [ ${DO_VERIFY} -eq 1 ] &&  [ -f "${KEYRING}" ]; then
                                # Assume this is a GPG signature
                                verify_signature "${archive}" ".${ending}"
                            fi

                            break
                        else
                            # Could be an incorrectly detected .zip file
                            zipinfo -h "${archive}" >/dev/null 2>&1

                            if [ ${?} -eq 0 ]; then
                                archive="$(extract_zip "${short}" "${archive}")"
                            else
                                break
                            fi
                        fi

                        ;;
                    application/pgp-signature)
                        if [ ${DO_VERIFY} -eq 1 ] && [ -f "${KEYRING}" ]; then
                            for signature in ${GPG_SIGNATURE_ENDINGS} ; do
                                # see if this is our signature
                                echo "${archive}" | grep -q "${signature}$" >/dev/null 2>&1
                                [ ${?} -eq 0 ] && break
                            done

                            verify_signature "${archive}" "${signature}"
                        fi

                        break

                        ;;
                    *)
                        break
                        ;;
                esac

                # see if there is any signature verification to do
                verify_intermediates "${archive}"
            done
        done
    fi
}

rm -f "${RMLIST}" "${VERIFY}"

# Do two passes across the package's archives file.  The first pass handles
# uncompressing and checking any GPG signatures.  This pass is skipped if
# the package does not have a GPGKEYS variable.  The second pass is the same
# loop with all extraction enabled.

DO_VERIFY=1
run_extract || exit 9

DO_VERIFY=0
run_extract || exit 10

# Handle source that comes from our own source tree.
if [ -d "${CWD}/src" ]; then
    if [ -d "${WRKDIR}/${WRKDIST}" ]; then
        echo "*** ${WRKDIR}/${WRKDIST} already exists." >&2
        exit 11
    fi

    cp -pr "${CWD}/src" "${WRKDIR}/${WRKDIST}"
fi

if [ -f "${RMLIST}" ]; then
    cat "${RMLIST}" | xargs rm -f
fi

rm -f "${RMLIST}" "${VERIFY}" "${KEYRING}~"
