#!/bin/sh
#
# fakeinst.sh: Implement fakeinst stage.
#
# {{{ Apache License version 2.0
#
# Copyright 2011-2018 David Cantrell <david.l.cantrell@gmail.com>
#                     Chris Lumens <chris@bangmoney.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# }}}

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $PKG                  Package name.
#    $VER                  Package version.
#    $UPSTREAM_PKG         Upstream version of the package name.
#    $WRKDIR               The top level work directory.
#    $WRKDIST              The name of the extracted source directory.
#    $WRKINST              The staging root directory.
#    $WRKSYSROOT           The --sysroot tree for this build.
#    $WRKDEBUG             A copy of $WRKINST with non-stripped ELF files.
#    $DEBUGDIR             The directory in $WRKDEBUG.
#    $NO_DEBUG             Create a debug package (y/n)?
#    $WRKSRC               The extracted source directory.
#    $WRKBLD               The source build directory.
#    $SEPARATE_BUILD       (y/n) Use source build directories?
#    $CONFIGURE_STYLE      The method used to configure $WRKSRC for builds.
#    $PYTHON               Python interpreter to use.
#    $PACKAGE              The full package name.
#    $FAKEINST_ENV         Environment variables passed to fakeinst target.
#    $FAKEINST_TARGET      Name of fakeinst target (default=install).
#    $FAKEINST_ARGS        Extra arguments to $(FAKEINST_COMMAND).
#    $FAKEINST_COMMAND     The make command, usually.
#    $MAKE                 The make command driving the build system.
#    $TOPSRC               The top level distribution source directory.
#    $IGNORE_REQS          List of package build deps to ignore (optional).
#    $TARGET_NAME          Target arch name (e.g., vax-peachtree-linux).
#    $TARGET_MACH          arch(1) or 'uname -m' value (or 'noarch').
#    $BINDIR               Install path for binaries (e.g., /usr/bin).
#    $DATADIR              Directory for architecture-independent data files.
#    $MANDIR               Directory where man pages are installed.
#    $DOCDIR               Directory where documentation files are installed.
#    $LICENSEDIR           Directory where license files are installed.
#    $INCLUDEDIR           Install path for headers (e.g., /usr/include).
#    $LIBDIR               Install path for libraries (e.g., /usr/lib).
#    $STRIP_DT_RPATH       Remove DT_RPATH and DT_RUNPATH from ELF files (Y/n).
#    $DOCS                 List of doc files in $(WRKSRC) to install to $(DOCDIR).
#    $LICENSES             List of license files in $(WRKSRC) to install to
#                          $(LICENSEDIR).
# This script is called by the build system scripts and should not be called
# directly.

CWD="$(pwd)"
TOPSRC="$(readlink -q -f ${TOPSRC})"
LOG="${WRKDIR}/fakeinst.log"
ERR="${WRKDIR}/fakeinst.err"
DT_NEEDED="${WRKDIR}/DT_NEEDED"
LISTED_DEPS="${WRKDIR}/listed.deps"
NEEDED_DEPS="${WRKDIR}/needed.deps"
HAVE_DEPS="${WRKDIR}/have.deps"
ALL_PLISTS="${WRKDIR}/expanded-plists"
STATUS=0

find_empty() {
    find $1 -type f | while read pathname ; do
        [ -s ${pathname} ] || echo "*** WARNING: $(echo ${pathname} | sed -e "s|^${WRKINST}||g") is empty"
    done
}

# Install any documentation files
if [ ! -z "${DOCS}" ]; then
    mkdir -p ${WRKINST}${DOCDIR}

    for doc in ${DOCS} ; do
        if [ -d ${WRKSRC} ]; then
            ls -1 ${WRKSRC}/${doc} >/dev/null 2>&1
            if [ $? -eq 0 ]; then
                ( cd ${WRKSRC} ; cp -prv --parents ${doc} ${WRKINST}${DOCDIR} )
                continue
            fi
        fi

        ls -1 ${CWD}/${doc} >/dev/null 2>&1
        if [ $? -eq 0 ]; then
            ( cd ${CWD} ; cp -prv --parents ${doc} ${WRKINST}${DOCDIR} )
            continue
        fi

        echo "WARNING: Unable to find ${doc} for ${DOCDIR}" >&2
    done

    find ${WRKINST}${DOCDIR} -type d | xargs chmod ugo+rx
    find ${WRKINST}${DOCDIR} -type f | xargs chmod ugo+r
    find_empty ${WRKINST}${DOCDIR}
fi

# Install any license files
if [ ! -z "${LICENSES}" ]; then
    mkdir -p ${WRKINST}${LICENSEDIR}

    for lic in ${LICENSES} ; do
        if [ -d ${WRKSRC} ]; then
            ls -1 ${WRKSRC}/${lic} >/dev/null 2>&1
            if [ $? -eq 0 ]; then
                ( cd ${WRKSRC} ; cp -prv --parents ${lic} ${WRKINST}${LICENSEDIR} )
                continue
            fi
        fi

        ls -1 ${CWD}/${lic} >/dev/null 2>&1
        if [ $? -eq 0 ]; then
            ( cd ${CWD} ; cp -prv --parents ${lic} ${WRKINST}${LICENSEDIR} )
            continue
        fi

        echo "WARNING: Unable to find ${lic} for ${LICENSEDIR}" >&2
    done

    find ${WRKINST}${LICENSEDIR} -type d | xargs chmod ugo+rx
    find ${WRKINST}${LICENSEDIR} -type f | xargs chmod ugo+r
    find_empty ${WRKINST}${LICENSEDIR}
fi

[ "${SKIP_FAKEINST}" = "y" ] && exit 0

if [ -z "${WRKDIR}" ] || [ ! -d "${WRKDIR}" ]; then
    exit 1
fi

if [ -z "${WRKSRC}" ] || [ ! -d "${WRKSRC}" ]; then
    exit 2
fi

if [ -z "${WRKINST}" ] || [ ! -d "${WRKINST}" ]; then
    exit 4
fi

[ -z "${WRKDEBUG}" ] && exit 5
[ -z "${PACKAGE}" ] && exit 6

if [ "${SEPARATE_BUILD}" = "y" ]; then
    if [ -z "${WRKBLD}" ] || [ ! -d ${WRKBLD} ]; then
        exit 7
    fi
fi

[ -z "${FAKEINST_TARGET}" ] && exit 9

unique() {
    [ -f ${1} ] || return
    sort ${1} | uniq > ${1}.new
    mv ${1}.new ${1}
}

# Prepare a plist string for use by grep.  That is:
# 1) Escape the '+' character that may be in a filename.
# 2) Convert any '*' wildcards to the '.*' regex syntax.
escape_string() {
    echo "${1}" | sed -e 's|+|\\+|g' | sed -e 's|\*|.*|g'
}

if [ -d ${WRKSRC} ]; then
    echo ">>> Installing ${PACKAGE} in fake root..."

    if [ "${SEPARATE_BUILD}" = "y" ]; then
        cd ${WRKBLD}
    else
        cd ${WRKSRC}
    fi

    case "${CONFIGURE_STYLE}" in
        gnu|noconfigure|manual|cmake)
            if [ "${CONFIGURE_STYLE}" = "manual" ]; then
                echo "*** WARNING: CONFIGURE_STYLE=manual is deprecated, use CONFIGURE_STYLE=noconfigure" >&2
            fi

            ( env ${FAKEINST_ENV} \
                  DESTDIR="${WRKINST}" \
              ${FAKEINST_COMMAND} ${FAKEINST_TARGET} ${FAKEINST_ARGS} \
              2>&1 1>&3 | tee ${ERR} \
            ) 3>&1 1>&2 | tee ${LOG} || exit $?

            ;;
        haskell)
            ( env ${FAKEINST_ENV} \
                  LIBDIR="${LIBDIR}" \
                  PACKAGE="${PACKAGE}" \
                  DESTDIR="${WRKINST}" \
              ${FAKEINST_COMMAND} ${FAKEINST_TARGET} ${FAKEINST_ARGS} \
              2>&1 1>&3 | tee ${ERR} \
            ) 3>&1 1>&2 | tee ${LOG} || exit $?

            # Special handling for Haskell package docs
            if [ -d ${WRKINST}${DATADIR}/doc/${UPSTREAM_PKG}-${VER} ]; then
                rm -rf ${WRKINST}${DATADIR}/doc/${UPSTREAM_PKG}-${VER} || exit $?
            fi

            # license files are common in directories by themselves
            for legalfile in LICENSE COPYING ; do
                find "${WRKINST}${DATADIR}" -type f -name ${legalfile} 2>/dev/null | \
                while read legalfile ; do
                    if [ -f "${legalfile}" ]; then
                        dn="$(dirname "${legalfile}")"
                        echo "${dn}" | grep -q "^${WRKINST}${LICENSEDIR}" >/dev/null 2>&1
                        [ $? -eq 0 ] && continue
                        rm ${legalfile}
                        rmdir -p --ignore-fail-on-non-empty ${dn}
                    fi
                done
            done

            # symlink all shared libraries in to /usr/lib
            # This helps us depsolving when installing or upgrading packages as well as
            # running programs like quitxmonad that link with these shared libraries.
            if [ -d ${WRKINST}${LIBDIR} ]; then
                cd ${WRKINST}${LIBDIR}
                ghclibdir="$(ls -1d ghc-*)"

                if [ -d ${ghclibdir} ]; then
                    cd ${ghclibdir}
                    find . -type f -name "*.so" | while read libname ; do
                        baselibpath="$(echo "${libname}" | cut -c3-)"
                        baselibname="$(basename "${libname}")"
                        ln -sf ${ghclibdir}/${baselibpath} ${WRKINST}${LIBDIR}/${baselibname}
                    done
                fi
            fi

            ;;
        python)
            ( env ${FAKEINST_ENV} \
              ${PYTHON} setup.py ${FAKEINST_TARGET} \
                  --verbose -O1 --skip-build --root ${WRKINST} \
                  ${FAKEINST_ARGS} \
              2>&1 1>&3 | tee ${ERR} \
            ) 3>&1 1>&2 | tee ${LOG} || exit $?

            # Special handling for Python module installation
            # (this may not exist in all things that build with setup.py, but
            # we have seen it in more than one package)
            if [ -d "${WRKINST}${WRKSYSROOT}" ]; then
                mkdir -p ${WRKDIR}/tmp
                mv ${WRKINST}${WRKSYSROOT}/* ${WRKDIR}/tmp
                rmdir -p --ignore-fail-on-non-empty ${WRKINST}${WRKSYSROOT}
                cp -pr ${WRKDIR}/tmp/* ${WRKINST}
                rmdir ${WRKDIR}/tmp
            fi

            ;;
        qmake)
            if [ -f ${WRKSYSROOT}${BINDIR}/qt.conf ]; then
                rm -f ${WRKSYSROOT}${BINDIR}/qt.conf
            fi

            ( env ${FAKEINST_ENV} \
                  INSTALL_ROOT="${WRKINST}" \
              ${FAKEINST_COMMAND} ${FAKEINST_TARGET} ${FAKEINST_ARGS} \
              2>&1 1>&3 | tee ${ERR} \
            ) 3>&1 1>&2 | tee ${LOG} || exit $?

            # everything will be installed to WRKSYSROOT under WRKINST, so
            # manually move things to they are root in WRKINST
            if [ -d ${WRKINST}${WRKSYSROOT} ]; then
                cp -pr ${WRKINST}${WRKSYSROOT}/* ${WRKINST}
                rm -rf ${WRKINST}${WRKSYSROOT}/*
                rmdir -p --ignore-fail-on-non-empty ${WRKINST}${WRKSYSROOT}
            fi

            ;;
        meson)
            ( env ${FAKEINST_ENV} \
                  DESTDIR="${WRKINST}" \
              ninja -C ${WRKBLD} ${FAKEINST_TARGET} ${FAKEINST_ARGS} \
              2>&1 1>&3 | tee ${ERR} \
            ) 3>&1 1>&2 | tee ${LOG} || exit $?

            ;;
        cargo)
            cd "${WRKBLD}"
            ( env ${FAKEINST_ENV} \
              cargo install -v --path . ${FAKEINST_ARGS} ${PKG} \
              2>&1 1>&3 | tee ${ERR} \
            ) 3>&1 1>&2 | tee ${LOG} || exit $?

            find "${WRKINST}" -type f -name ".crates.toml" | xargs rm -f

            ;;
        *)
            echo "*** Do not understand CONFIGURE_STYLE=${CONFIGURE_STYLE}"
            exit 10
            ;;
    esac
fi

# Remove $WRKSRC and $WRKSYSROOT references from installed files.
# NOTE:  Put the longest patterns before shorter patterns.
if [ ! -z "${WRKSYSROOT}" ] && [ ! -z "${WRKSRC}" ]; then
    find ${WRKINST} -type f | while read f ; do
        file -b --mime-type "${f}" 2>/dev/null | grep -q -E "^text/" >/dev/null 2>&1
        [ $? -eq 0 ] || continue

        grep -q -E "(${WRKSYSROOT}|${WRKSRC})" "${f}" >/dev/null 2>&1
        [ $? -eq 0 ] || continue

        sed -i -e "s|'LDFLAGS=-L${WRKSYSROOT}${LIBDIR}\ -Wl,--sysroot=${WRKSYSROOT}'|'LDFLAGS=-L${LIBDIR}'|g" "${f}"
        sed -i -e "s|-Wl,--sysroot=${WRKSYSROOT}||g" "${f}"
        sed -i -e "s|--sysroot=${WRKSYSROOT}||g" "${f}"
        sed -i -e "s|'--with-sysroot=${WRKSYSROOT}'||g" "${f}"
        sed -i -e "s|-I${WRKSYSROOT}${INCLUDEDIR}|-I${INCLUDEDIR}|g" "${f}"
        sed -i -e "s|-L${WRKSYSROOT}${LIBDIR}|-L${LIBDIR}|g" "${f}"
        sed -i -e "s|${WRKSYSROOT}${INCLUDEDIR}|${INCLUDEDIR}|g" "${f}"
        sed -i -e "s|${WRKSYSROOT}${LIBDIR}|${LIBDIR}|g" "${f}"
        sed -i -e "s|${WRKSRC}|${WRKDIST}|g" "${f}"
    done
fi

# Clean up the man pages
if [ -d ${WRKINST}${MANDIR} ]; then
    # Create non symlink style man page references.
    find ${WRKINST}${MANDIR} -type l | while read manlink ; do
        section="$(basename $(dirname ${manlink}))"
        link="$(basename $(readlink ${manlink}))"
        rm -f ${manlink}
        manlink="$(dirname ${manlink})/$(basename ${manlink})"
        echo ".so ${section}/${link}" > ${manlink}
        chmod 0644 ${manlink}
    done

    # Warn about compressed man pages
    find ${WRKINST}${MANDIR} -type f | while read manpage ; do
        mtype="$(file -b --mime-type ${manpage} | cut -c-5)"
        if [ ! "${mtype}" = "text/" ]; then
            echo "*** ERROR: Unknown man page type for ${manpage}: $(file -b --mime-type ${manpage})" >&2
        fi
    done
fi

# Strip ELF files, add debugging symbols, collect ELF dependencies
if [ -d ${WRKINST} ] && [ "${NO_DEBUG}" = "n" ] && [ ! "${TARGET_MACH}" = "noarch" ]; then
    find ${WRKINST} -type f | while read filename ; do
        # Skip symlinks
        [ -L "${filename}" ] && continue

        # Gather MIME type of the file
        ftype="$(file -b --mime-type "${filename}" 2>/dev/null)"

        # Skip non-ELF files and likely ELF object (.o) files
        if ( [ ! "${ftype}" = "application/x-executable" ] &&
             [ ! "${ftype}" = "application/x-sharedlib" ] &&
             [ ! "${ftype}" = "application/x-pie-executable" ] &&
             [ ! "${ftype}" = "application/x-pie-sharedlib" ] ) || \
           [ "$(echo "${filename}" | rev | cut -c2)" = "o." ]; then
            continue
        fi

        # Skip unknown architecture files, which are basically things that come
        # with emulator software (e.g., firmware images)
        elfarch="$(objdump -x "${filename}" 2>/dev/null | grep ^architecture: | cut -d ' ' -f 2)"
        [ "${elfarch}" = "UNKNOWN!," ] && continue

        # Strip any DT_RPATH values unless told otherwise.
        if [ "${STRIP_DT_RPATH}" = "y" ]; then
            patchelf --remove-rpath "${filename}"
        fi

        # Find all ELF shared objects, copy them to $WRKDEBUG,
        # and strip them in $WRKINST.
        dn="$(dirname $(echo "${filename}" | sed -e "s|^${WRKINST}||g"))"
        debugfilename=${WRKDEBUG}${DEBUGDIR}${dn}/$(basename "${filename}").debug

        [ -d ${WRKDEBUG}${DEBUGDIR}${dn} ] || mkdir -p ${WRKDEBUG}${DEBUGDIR}${dn}

        owner="$(find "${filename}" -printf %U)"
        group="$(find "${filename}" -printf %G)"
        perms="$(find "${filename}" -printf %m)"

        ( cp -av ${filename} ${debugfilename} \
          2>&1 1>&3 | tee -a ${ERR} \
        ) 3>&1 1>&2 | tee -a ${LOG} || exit $?

        # Only perform objcopy step on dynamic executables and libraries
        ldd ${filename} >/dev/null 2>&1
        if [ $? -eq 0 ]; then
            ( objcopy -v --add-gnu-debuglink=${debugfilename} ${filename} \
              2>&1 1>&3 | tee -a ${ERR} \
            ) 3>&1 1>&2 | tee -a ${LOG} || exit $?
        fi

        # If this is a Haskell package, we need to fix up the rpath in
        # it to no longer point at $WRKSYSROOT.
        if [ "${CONFIGURE_STYLE}" = "haskell" ]; then
            if [ ! -z "$(patchelf --print-rpath ${filename} | grep "${WRKSYSROOT}")" ]; then
                OLDPATH="$(patchelf --print-rpath ${filename} | sed -e "s|^${filename}: RPATH=||")"
                NEWPATH="$(echo "${OLDPATH}" | sed -e "s|${WRKSYSROOT}||g")"
                patchelf --set-rpath ${NEWPATH} ${filename}
            fi
        fi

        # Use long options for strip(1).  The man page is not necessarily
        # correct about what the -g option does.  Better to be safe.
        STRIP_OPTS="-v -p -K .note.gnu.build-id --strip-debug"
        if [ "${ftype}" = "application/x-executable" ]; then
            STRIP_OPTS="${STRIP_OPTS} --strip-unneeded"
        fi

        ( strip ${STRIP_OPTS} ${filename} \
          2>&1 1>&3 | tee -a ${ERR} \
        ) 3>&1 1>&2 | tee -a ${LOG} || exit $?

        chown ${owner}:${group} "${filename}"
        chmod ${perms} "${filename}"

        # Collect DT_NEEDED names to verify the build dependencies.
        readelf -d ${filename} | grep "(NEEDED)" | cut -d '[' -f 2 | cut -d ']' -f 1 >> ${DT_NEEDED}
    done

    # Remove all .la files.
    find ${WRKINST} -type f -name '*.la' -exec rm {} \;
    find ${WRKINST} -type l -name '*.la' -exec rm {} \;
fi

# Verify the build dependencies are correct on this package.
if [ -f ${DT_NEEDED} ]; then
    sort ${DT_NEEDED} | uniq > ${DT_NEEDED}.new
    mv ${DT_NEEDED}.new ${DT_NEEDED}

    rm -f ${LISTED_DEPS} ${NEEDED_DEPS} ${HAVE_DEPS}
    len="$(($(echo ${TOPSRC} | wc -c) + 1))"

    if [ -z "${BUILD_DEPENDS}" ] && [ -f "${CWD}/buildreqs" ]; then
        BUILD_DEPENDS="$(${MAKE} -C ${CWD} show-buildreqs)" || exit $?
    fi

    for dep in ${BUILD_DEPENDS} ; do
        echo ${dep} >> ${LISTED_DEPS}
    done

    # Expand all the plist files first so we can check them
    for plist in $(find ${TOPSRC}/ -type f -name "plist*" 2>/dev/null | grep -v -E "(^${TOPSRC}/obj/|plist\.[cho5])" 2>/dev/null | grep -E "(plist|plist.${TARGET_MACH})$" 2>/dev/null) ; do
        p="$(dirname ${plist})"
        pkgpath="$(echo "${p}" | sed -e "s|^${TOPSRC}/||g")"
        sed -e "s|\${VER}|$(${MAKE} -C ${p} show var=VER)|g" ${plist} | \
        sed -e "s|\${REL}|$(${MAKE} -C ${p} show var=REL)|g" | \
        sed -e "s|\${TARGET_NAME}|$(${MAKE} -C ${p} show var=TARGET_NAME)|g" | \
        sed -e "s|\${TARGET_MACH}|$(${MAKE} -C ${p} show var=TARGET_MACH)|g" | \
        sed -e "s|^|${pkgpath}:|g" >> ${ALL_PLISTS}
    done

    # First, check to see if the dependency exists in an existing plist file
    while read needed ; do
        grep -q "${needed} " ${ALL_PLISTS} >/dev/null 2>&1
        if [ $? -eq 0 ]; then
            need="$(grep "${needed} " ${ALL_PLISTS} | cut -d ':' -f 1)"
            echo "${needed}" >> ${DT_NEEDED}.owned

            # Figure out what packages provide these dependencies and list
            # them in the needed deps list, except for anything we want to
            # ignore.
            for p in ${need} ; do
                if [ ! "$(basename ${p})" = "$(basename ${CWD})" ]; then
                    if [ -z "${IGNORE_REQS}" ]; then
                        echo "${p}" >> ${NEEDED_DEPS}
                    else
                        ignored=0
                        for req in ${IGNORE_REQS} ; do
                            if [ "${req}" = "${p}" ]; then
                                ignored=1
                                break
                            fi
                        done
                        [ ${ignored} -eq 0 ] && echo "${p}" >> ${NEEDED_DEPS}
                    fi
                fi
            done

            continue
        fi
    done < ${DT_NEEDED}

    # Second, if the dependency didn't exist as-is, see if a wildcard
    # line covers it.  We do this by gathering any wildcard lines in
    # plist file and seeing if any of those match something installed
    # in the sysroot or the pkgroot.  If there is a match, that means
    # either a dependent package owns it or this package owns it, so
    # then we can say this dependency is owned.
    while read needed ; do
        grep -q -E "(\?|\*)" ${ALL_PLISTS} >/dev/null 2>&1
        if [ $? -eq 0 ]; then
            # This plist file actually has wildcard lines, so iterate
            # over those.
            grep -E "(\?|\*)" ${ALL_PLISTS} | while read wild ; do
                # Filename may contain spaces, so parse accordingly
                RL="$(echo "${wild}" | rev)"
                LEN="$(echo "${wild}" | wc -c)"

                field=1
                limit=1
                end=""
                while [ ${field} -le ${limit} ]; do
                    val="$(echo "${RL}" | cut -d ' ' -f ${field})"

                    if [ ${field} -eq 1 ]; then
                        rval="$(echo "${val}" | rev)"
                        [ "${rval}" = "nochecksum" ] && limit=4 || limit=3
                    fi

                    field=$((${field} + 1))
                    end="${end}${val} "
                done

                filepath="$(echo "${wild}" | cut -c-$(($LEN - $(echo "${end}" | wc -c))) | cut -d ':' -f 2)"

                # If this package owns it, note that and move on.
                ls ${WRKINST}/${filepath} >/dev/null 2>&1
                if [ $? -eq 0 ]; then
                    ls -1 ${WRKINST}/${filepath} | while read line ; do
                        filename="$(basename "${line}")"
                        if [ "${filename}" = "${needed}" ]; then
                            echo "${needed}" >> ${DT_NEEDED}.owned
                            continue
                        fi
                    done
                fi

                # If we find it in the sysroot, assume another package
                # owns it and move on.
                ls ${WRKSYSROOT}/${filepath} >/dev/null 2>&1
                if [ $? -eq 0 ]; then
                    ls -1 ${WRKSYSROOT}/${filepath} | while read line ; do
                        filename="$(basename "/${line}")"
                        if [ "${filename}" = "${needed}" ]; then
                            echo "${needed}" >> ${DT_NEEDED}.owned
                            continue
                        fi
                    done
                fi
            done
        fi
    done < ${DT_NEEDED}
fi

if [ -s ${DT_NEEDED}.owned ]; then
    cat ${DT_NEEDED} ${DT_NEEDED}.owned | sort | uniq -u > ${DT_NEEDED}.unowned
fi

rm -f ${DT_NEEDED}.owned
[ -s ${NEEDED_DEPS} ] && unique ${NEEDED_DEPS}

if [ -s ${NEEDED_DEPS} ]; then
    while read needed ; do
        needed="$(escape_string "${needed}")"
        grep -s -q -E "^${needed}$" ${LISTED_DEPS}
        if [ $? -eq 0 ]; then
            grep -v -E "^${needed}$" ${LISTED_DEPS} > ${LISTED_DEPS}.new
            mv ${LISTED_DEPS}.new ${LISTED_DEPS}
            echo "${needed}" >> ${HAVE_DEPS}
        fi
    done < ${NEEDED_DEPS}
fi

if [ -s ${HAVE_DEPS} ]; then
    while read have ; do
        have="$(escape_string "${have}")"
        grep -v -E "^${have}$" ${NEEDED_DEPS} > ${NEEDED_DEPS}.new
        mv ${NEEDED_DEPS}.new ${NEEDED_DEPS}
    done < ${HAVE_DEPS}
fi

if [ -s ${NEEDED_DEPS} ]; then
    echo "*** WARNING: The following build dependencies are missing:" >&2
    while read needed ; do
        echo "***     ${needed}" >&2
    done < ${NEEDED_DEPS}
fi

if [ -s ${DT_NEEDED}.unowned ]; then
    unique ${DT_NEEDED}.unowned

    echo "*** WARNING: These build dependencies are not owned by any package:" >&2
    while read unowned ; do
        echo "***     ${unowned}" >&2
    done < ${DT_NEEDED}.unowned
fi

rm -f ${HAVE_DEPS} ${NEEDED_DEPS} ${LISTED_DEPS} ${DT_NEEDED}.unowned ${ALL_PLISTS}

[ -s ${LOG} ] || rm -f ${LOG}
[ -s ${ERR} ] || rm -f ${ERR}

exit ${STATUS}
