#!/bin/sh
#
# fetch.sh: Implement 'fetch' stage.
#
# {{{ Apache License version 2.0
#
# Copyright 2011-2018 David Cantrell <david.l.cantrell@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# }}}

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $TOPSRC               The top level project directory.
#    $MAKE                 The make command.
#    $WRKDIR               The top level work directory.
#    $OUR_MIRRORS          The list of our mirror download sites.
#    $FORCE_REFRESH        Skip download attempts from mirrors (y/n).
#    $URLS                 The URLS list from the package Makefile.
#    $DISTFILES            The top level distfiles directory.
#    $PKGDISTFILES         The package specific distfiles subdirectory.
# This script is called by the build system scripts and should not be called
# directly.

CURL="curl -A Mozilla -k -f -L --retry 3 --ftp-pasv"
CWD="$(pwd)"

if [ -z "${WRKDIR}" ] || [ ! -d "${WRKDIR}" ]; then
    exit 1
fi

if [ -z "${DISTFILES}" ] || [ ! -d "${DISTFILES}" ]; then
    exit 2
fi

if [ -z "${URLS}" ]; then
    [ -d "${CWD}/src" ] && exit 0
    [ -f "${CWD}/archives" ] || exit 0

    # URLS can be absent if archives contains nothing local
    remainder="$(grep -v "^#include" "${CWD}/archives" | grep -v "^$")"
    if [ ! -z "${remainder}" ]; then
        echo "*** Missing URLS variable in package Makefile." >&2
        exit 3
    fi
fi

[ -z "${PKGDISTFILES}" ] && exit 4

[ -z "${FORCE_DOWNLOAD}" ] || FORCE_DOWNLOAD="$(echo "${FORCE_DOWNLOAD}" | tr [A-Z] [a-z])"

fetch_file() {
    archive="${1}"
    force_download="${2}"

    # If the download is forced, do not try to fetch from our mirrors.
    if [ ${force_download} -eq 1 ]; then
        urls="${URLS}"
    else
        urls="${OUR_MIRRORS} ${URLS}"
    fi

    for url in ${urls} ; do
        # Skip if we already have it.
        if [ -d "${PKGDISTFILES}" ] && [ -f "${PKGDISTFILES}/${archive}" ] && [ ${force_download} -ne 1 ]; then
            grep "${archive}" ${WRKDIR}/fetch.log >/dev/null 2>&1

            if [ $? -eq 0 ]; then
                echo ">>> ${archive} found, skipping." >> ${WRKDIR}/.msg
            else
                echo "${archive}" >> ${WRKDIR}/fetch.log
            fi

            continue
        fi

        # Some URLS may end with a slash, some may not.  Be careful not
        # to add an extra slash on the end.
        echo "${url}" | grep -q "\/$" 2>/dev/null
        [ $? -eq 0 ] && slash="" || slash="/"

        # If a URL ends with '@', we want to use it as-is without adding the archive
        # name to the URL.
        redirect=0
        echo "${url}" | grep -q -E '@$' 2>/dev/null
        if [ $? -eq 0 ]; then
            url="$(echo "${url}" | sed -e 's|@$||g')"
            redirect=1
        fi

        echo ">>> Fetching ${archive}..."
        [ -d "${PKGDISTFILES}" ] || mkdir -p "${PKGDISTFILES}"
        if [ ${redirect} -eq 1 ]; then
            ${CURL} -o "${PKGDISTFILES}/${archive}" -e "${url}" ${url}
        else
            ${CURL} -o "${PKGDISTFILES}/${archive}" -e "${url}" ${url}${slash}${archive}
        fi

        if [ $? -eq 0 ]; then
            echo "${archive}" >> ${WRKDIR}/fetch.log
            force_download=0
        else
            rm -f "${PKGDISTFILES}/${archive}"
        fi

        # SourceForge check...*sigh*
        t="$(file -b --mime-type "${PKGDISTFILES}/${archive}" | cut -c-5)"
        echo "${url}" | grep -q "sourceforge\.net" >/dev/null 2>&1
        if [ $? -eq 0 ] && [ "${t}" = "text/" ]; then
            echo >&2
            echo "*** SOURCEFORGE BREAKS DOWNLOADING ***" >&2
            echo "The build system downloaded an HTML page for your archive because" >&2
            echo "SourceForge uses a client-side JavaScript-based redirect function" >&2
            echo "to provide you the download link.  It is impossible to script." >&2
            echo >&2
            echo "You need to manually download the files and copy them to distfiles." >&2
            echo "Sorry, but SourceForge is terrible." >&2
            echo >&2
            echo "${url}${slash}${archive}" >&2
            echo >&2
            exit 5
        fi
    done
}

# Fetch archives not included from another package.
if [ -f "${CWD}/archives" ]; then
    rm -f ${WRKDIR}/.msg

    grep -v "^#" "${CWD}/archives" | grep -v "^$" | \
    while read first second third ; do
        force_download=0

        if [ -z "${second}" ] && [ -z "${third}" ]; then
            filename="${first}"
            force_download=1
        else
            if [ -z "${third}" ]; then
                # archives file is in the openssl(1) dgst format
                filename="$(echo ${first} | cut -d '(' -f 2 | cut -d ')' -f 1)"
            else
                # archives file is in legacy Peachtree Linux format
                filename="${third}"
            fi
        fi

        [ "${FORCE_REFRESH}" = "y" ] && force_download=1

        # We may have the archive in an old location, move it
        if [ -d "${PKGDISTFILES}" ]; then
            if [ -f "${DISTFILES}/${filename}" ] && [ ! -f "${PKGDISTFILES}/${filename}" ]; then
                mv "${DISTFILES}/${filename}" "${PKGDISTFILES}"
            fi
        fi

        # Fetch this archive file
        fetch_file ${filename} ${force_download}
    done

    if [ -f ${WRKDIR}/.msg ]; then
        sort ${WRKDIR}/.msg | uniq
        rm -f ${WRKDIR}/.msg
    fi
fi

# Fetch any archives included from another package.
if [ -f "${CWD}/archives" ]; then
    # if there are no included archives files, exit
    grep -q -E "^#include" "${CWD}/archives" >/dev/null 2>&1
    [ $? -eq 0 ] || exit 0

    # change to each included package directory and fetch
    grep -E "^#include" "${CWD}/archives" | \
    while read directive includepath ; do
        pkgdir="${TOPSRC}/$(dirname $(echo ${includepath} | cut -d '<' -f 2 | cut -d '>' -f 1))"
        ${MAKE} --no-print-directory -C ${pkgdir} fetch
    done

    # symlink the included archives to our distfiles directory
    grep -E "^#include" "${CWD}/archives" | \
    while read directive includepath ; do
        includepath="$(echo ${includepath} | cut -d '<' -f 2 | cut -d '>' -f 1)"
        pkgdir="${TOPSRC}/$(dirname ${includepath})"
        includedarchives="${TOPSRC}/${includepath}"
        includeddistfiles="$(${MAKE} --no-print-directory -C ${pkgdir} show var=PKGDISTFILES)"

        grep -v "^#" "${includedarchives}" | grep -v "^$" | \
        while read first second third ; do
            [ -d "${PKGDISTFILES}" ] || mkdir -p "${PKGDISTFILES}"

            if [ -z "${third}" ]; then
                # archives file is in openssl(1) dgst format
                filename="$(echo ${first} | cut -d '(' -f 2 | cut -d ')' -f 1)"
            else
                # archives file is in legacy Peachtree Linux format
                filename="${third}"
            fi

            # create an absolute symlink, then convert it to relative
            ln -sf "${includeddistfiles}/${filename}" "${PKGDISTFILES}/${filename}"
            symlinks --help >/dev/null 2>&1
            [ ${?} -eq 127 ] && continue || symlinks -c "${PKGDISTFILES}"
        done
    done
fi
