#!/bin/sh
#
# install-chroot.sh: Install Peachtree Linux to the $CHROOT path.
#
# {{{ Apache License version 2.0
#
# Copyright 2012-2018 David Cantrell <david.l.cantrell@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# }}}

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $TOPSRC               The top level distribution source directory.
#    $COLLECTIONS          The names of all the collections.
#    $EXCLUDES             List of subgroups or packages to exclude, or
#                          a file containing the list of packages to exclude.
#    $RELEASE_DIST         The directory containing built packages.
#    $TARGET_MACH          The target architecture we are working with.
#    $CHROOT               The chroot path (cannot be '/').
# This script is called by the build system scripts and should not be called
# directly.

PATH=/bin:/sbin:/usr/bin:/usr/sbin:/usr/local/bin:/usr/local/sbin

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 1
fi

[ -z "${COLLECTIONS}" ] && exit 2

if [ -z "${RELEASE_DIST}" ] || [ ! -d "${RELEASE_DIST}" ]; then
    exit 3
fi

if [ -z "${CHROOT}" ]; then
    echo "*** You must specify the CHROOT= variable." >&2
    exit 4
fi

if [ -d "${CHROOT}" ] && [ "$(readlink -f ${CHROOT})" = "/" ]; then
    exit 5
fi

[ -d "${CHROOT}" ] || mkdir -p -m 0755 "${CHROOT}"

# Require all collections to be built.
for collection in ${COLLECTIONS} ; do
    if [ ! -d "${RELEASE_DIST}/${COLLECTION}" ]; then
        echo "*** The ${collection} collection has not been built." >&2
        exit 6
    fi
done

# For a chroot installation, the user must be root.
if [ $(id -u) -ne 0 ]; then
    echo "*** You must be root to perform a chroot installation." >&2
    exit 7
fi

# We need distadd.  Just assume the one we can exec is current enough.
distadd --help >/dev/null 2>&1
if [ ${?} -eq 127 ]; then
    echo "*** distadd command not found" >&2
    exit 8
fi

# We need etcmerge.  Just assume the one we can exec is current enough.
etcmerge --help >/dev/null 2>&1
if [ ${?} -eq 127 ]; then
    echo "*** etcmerge command not found" >&2
    exit 9
fi

TMPDIR="$(mktemp -d)"
PKGLIST="${TMPDIR}/pkglist"
EXCLUDESLIST="${TMPDIR}/excludeslist"

# Build the list of excludes
[ -f "${EXCLUDES}" ] && EXCLUDES="$(grep -v ^$ ${EXCLUDES} 2>/dev/null | grep -v ^#)"

if [ ! -z "${EXCLUDES}" ]; then
    rm -f "${EXCLUDESLIST}"

    for p in ${EXCLUDES} ; do
        if [ -d "${TOPSRC}/${p}" ]; then
            find ${TOPSRC}/${p} -type f -name plist | while read f ; do
                d="$(dirname ${f})"
                package="$(make --no-print-directory -C ${d} show var=PACKAGE)"
                echo "Excluding ${package}"
                echo "${package}" >> ${EXCLUDESLIST}
            done
        else
            echo "*** ${p} is not a valid excludes.miniroot definition" >&2
            exit 12
        fi
    done

    if [ -s "${EXCLUDESLIST}" ]; then
        sort ${EXCLUDESLIST} | uniq > ${EXCLUDESLIST}.new
        mv ${EXCLUDESLIST}.new ${EXCLUDESLIST}
    fi
fi

# Generate list of base packages to install
rm -f "${PKGLIST}"
for a in ${TARGET_MACH} noarch ; do
    if [ -f "${RELEASE_DIST}/base/${a}/00-ORDER" ]; then
        cat "${RELEASE_DIST}/base/${a}/00-ORDER" >> "${PKGLIST}"
    fi
done

# Filter excluded packages from base install list
if [ -s "${EXCLUDESLIST}" ]; then
    while read p ; do
        grep -v "^${p}$" "${PKGLIST}" > "${PKGLIST}.new"
        mv "${PKGLIST}.new" "${PKGLIST}"
    done < "${EXCLUDESLIST}"
fi

# Install the base packages
while read dist ; do
    pkg="${RELEASE_DIST}/base/${TARGET_MACH}/${dist}"
    [ -f "${pkg}" ] || pkg="${RELEASE_DIST}/base/noarch/${dist}"

    if [ ! -f "${pkg}" ]; then
        echo "*** Missing ${pkg}, try 'make world' again." >&2
        exit 10
    fi

    distadd -m -r ${CHROOT} -d "${pkg}"
done < "${PKGLIST}"

# Update library cache after installing the base collection
if [ -x "${CHROOT}/sbin/ldconfig" ]; then
    chroot ${CHROOT} /sbin/ldconfig 2>/dev/null
elif [ -x "${CHROOT}/usr/sbin/ldconfig" ]; then
    chroot ${CHROOT} /usr/sbin/ldconfig 2>/dev/null
fi

# Install the remaining collections
for collection in ${COLLECTIONS} ; do
    [ "${collection}" = "base" ] && continue

    for a in ${TARGET_MACH} noarch ; do
        while read p ; do
            # Skip this package if the user excluded it
            if [ -s "${EXCLUDESLIST}" ]; then
                grep -q "^${p}$" "${EXCLUDESLIST}" >/dev/null 2>&1
                [ ${?} -eq 0 ] && continue
            fi

            pkg="${RELEASE_DIST}/${collection}/${a}/${dist}"

            if [ ! -f "${pkg}" ]; then
                echo "*** Missing ${pkg}, try 'make world' again." >&2
                exit 11
            fi

            distadd -m -r ${CHROOT} -d "${pkg}"
        done < ${RELEASE_DIST}/${collection}/${a}/00-ORDER
    done

    # Update the library cache after the collection is installed
    if [ -x "${CHROOT}/sbin/ldconfig" ]; then
        chroot ${CHROOT} /sbin/ldconfig 2>/dev/null
    elif [ -x "${CHROOT}/usr/sbin/ldconfig" ]; then
        chroot ${CHROOT} /usr/sbin/ldconfig 2>/dev/null
    fi
done

# Install initial configuration files
etcmerge -r ${CHROOT} -i

# Perform initial boot up things that /etc/rc normally would do.
# These are things that a freshly installed system would have happen at boot up.
# Initialize/update cache files, set up other control files, etc.
if [ -x ${CHROOT}/usr/bin/dbus-uuidgen ]; then
    chroot ${CHROOT} /usr/bin/dbus-uuidgen --ensure
fi

if [ -f /etc/resolv.conf ]; then
    install -D -m 0644 /etc/resolv.conf ${CHROOT}/etc/resolv.conf
fi

if [ -x ${CHROOT}/usr/bin/env ]; then
    if [ -x ${CHROOT}/usr/bin/gdk-pixbuf-query-loaders ]; then
        chroot ${CHROOT} /usr/bin/env -u LC_ALL -u LC_COLLATE -u LANG \
            /usr/bin/gdk-pixbuf-query-loaders --update-cache
    fi

    if [ -x ${CHROOT}/usr/bin/pango-query-modules ]; then
        chroot ${CHROOT} /usr/bin/env -u LC_ALL -u LC_COLLATE -u LANG \
            /usr/bin/pango-querymodules > ${TMPDIR}/pango.modules
        install -D -m 0644 ${TMPDIR}/pango.modules \
            ${CHROOT}/etc/pango/pango.modules
    fi

    if [ -x ${CHROOT}/usr/bin/gio-querymodules ] && [ -d ${CHROOT}/usr/lib/gio/modules ]; then
        chroot ${CHROOT} /usr/bin/env -u LC_ALL -u LC_COLLATE -u LANG \
            /usr/bin/gio-querymodules /usr/lib/gio/modules
    fi
fi

# Clean up
rm -rf ${TMPDIR}
exit 0
