#!/bin/sh
#
# make-release.sh: Create bootable release media from the staging tree.
#
# {{{ Apache License version 2.0
#
# Copyright 2012-2018 David Cantrell <david.l.cantrell@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# }}}

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $TOPSRC               The top level distribution source directory.
#    $DISTFILES            The location of upstream distfiles.
#    $OUTPUT               The directory to write the miniroot.img to.
#    $CHROOT               The chroot path (cannot be '/').
#    $TARGET_MACH          The target architecture.
#    $RELEASE              The built release subdirectory.
#    $RELEASE_DIST         The built release dist subdirectory.
#    $DISTRIBUTION_VERSION The version of Peachtree Linux.
#    $KERNEL_VERSION       The kernel version included in this tree.
#    $KERNEL_WRKSRC        The extracted kernel source directory.
#    $EXTRA_PACKAGES       The list of packages we need to make the boot iso
#                          but that we exclude from the miniroot.img file.
# This script is called by the build system scripts and should not be called
# directly.

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 1
fi

if [ -z "${DISTFILES}" ] || [ ! -d "${DISTFILES}" ]; then
    exit 2
fi

if [ -z "${OUTPUT}" ] || [ ! -d "${OUTPUT}" ]; then
    exit 3
fi

[ -z "${CHROOT}" ] && exit 4

if [ ! -d "${CHROOT}" ] || [ "$(readlink -f ${CHROOT})" = "/" ]; then
    exit 5
fi

[ -z "${TARGET_MACH}" ] && exit 6

if [ -z "${RELEASE}" ] || [ ! -d "${RELEASE}" ]; then
    exit 7
fi

if [ -z "${RELEASE_DIST}" ] || [ ! -d "${RELEASE_DIST}" ] || [ ! -d "${RELEASE_DIST}" ]; then
    exit 16
fi

[ -z "${DISTRIBUTION_VERSION}" ] && exit 8
[ -z "${KERNEL_VERSION}" ] && exit 9

# The name of the kernel image file depends on the target architecture
if [ "${TARGET_MACH}" = "x86_64" ]; then
    KERNEL="vmlinuz-${TARGET_MACH}"
elif [ "${TARGET_MACH}" = "ppc64" ]; then
    KERNEL="vmlinux-${TARGET_MACH}"
else
    KERNEL="vmlinux"
fi

MINIROOT="${OUTPUT}/miniroot-${TARGET_MACH}.img"
RELEASE_ISO="${RELEASE}/media/${TARGET_MACH}/release-${DISTRIBUTION_VERSION}-${TARGET_MACH}-$(date +%Y%m%d).iso"
RELEASE_LABEL=PLDIST
TMPDIR="$(mktemp -d)"

# User must be root to create the miniroot
if [ $(id -u) -ne 0 ]; then
    echo "*** You must be root to perform a chroot installation." >&2
    exit 10
fi

# Clear out initramfs-root
rm -f "${CHROOT}/*"

# Create the checksum files.
gen_SUMS() {
    for t in sha256 ; do
        c="openssl dgst -${t}"
        o="$(echo ${t} | tr [a-z] [A-Z])"

        find ${RELEASE}/ -type d | while read dname ; do
            if [ "${dname}/" = "$(realpath ${RELEASE})/" ]; then
                ${c} * > "${TMPDIR}/${o}" 2>/dev/null
                mv "${TMPDIR}/${o}" "${RELEASE}"
                [ -s "${RELEASE}/${o}" ] || rm -f "${RELEASE}/${o}"
                continue
            fi

            ( cd "${dname}"
              ${c} * > "${TMPDIR}/${o}" 2>/dev/null
              mv "${TMPDIR}/${o}" .
            )

            [ -s "${dname}/${o}" ] || rm -f "${dname}/${o}"
        done
    done
}

# Write out the version information file for this release tree.
gen_VERSION() {
    . ${CHROOT}/etc/os-release
    echo "${NAME} version ${VER}" > ${RELEASE}/VERSION
}

# Generate a list of changes since the last release.
gen_ChangeLog() {
    branch="$(cd ${TOPSRC} ; git symbolic-ref --short HEAD)"

    if [ "${branch}" = "master" ]; then
        # Only show changes since the last tag for master builds
        PREVTAG="$(cd ${TOPSRC} ; git tag -l | tail -n 1)"
        ( cd ${TOPSRC} ; git log --name-status ${PREVTAG}..HEAD ) > ${RELEASE}/ChangeLog
    else
        # Show all changes on the branch for stable builds
        ( cd ${TOPSRC} ; git log --name-status ) > ${RELEASE}/ChangeLog
    fi
}

# Add necessary entries to the default /etc/fstab in the miniroot.
modify_fstab() {
    CWD="$(pwd)"
    [ -f ${CHROOT}/etc/fstab ] || exit 11

    cd ${CHROOT}/etc
    len=$(wc -l fstab | cut -d ' ' -f 1)
    hlen=$(($(grep -n "Put your entries here:" fstab | cut -d ':' -f 1) + 2))
    tlen=$((${len} - $(grep -n "The following mount points are required for normal system operation" fstab | cut -d ':' -f 1) + 3))
    head -n ${hlen} fstab > fstab.header
    tail -n ${tlen} fstab > fstab.footer
    echo "rootfs     /  rootfs    defaults             0 0" > fstab.middle
    cat fstab.header fstab.middle fstab.footer > fstab
    rm fstab.header fstab.middle fstab.footer
    cd ${CWD}
}

# Count packages we have against what's in the 00-ORDER files
# Look for those files in architecture specific subdirectories, which will
# make sure we don't error out if we're missing arch-specific collections
# on a different architecture.
cd ${RELEASE_DIST}
for c in $(find . -type f -name "00-ORDER" | grep "${TARGET_MACH}" | sort | uniq) ; do
    collection="$(basename $(dirname $(dirname "${c}")))"
    cd ${RELEASE_DIST}/${collection}

    ( cd ${TARGET_MACH} ; ls -1 *.dist > ${TMPDIR}/ls.out )
    expected=$(wc -l "${TARGET_MACH}/00-ORDER" | cut -d ' ' -f 1)

    if [ -d noarch ]; then
        ( cd noarch ; ls -1 *.dist >> ${TMPDIR}/ls.out )
        expected=$((${expected} + $(wc -l "noarch/00-ORDER" | cut -d ' ' -f 1)))
    fi

    found=$(wc -l ${TMPDIR}/ls.out | cut -d ' ' -f 1)

    if [ ${expected} -ne ${found} ]; then
        echo "*** ERROR: package count mismatch in ${collection} collection" >&2
        echo "***            Expected: ${expected}" >&2
        echo "***            Found:    ${found}" >&2

        if [ -d noarch ]; then
            sort ${TARGET_MACH}/00-ORDER noarch/00-ORDER | uniq > ${TMPDIR}/00-ORDER.sorted
        else
            sort ${TARGET_MACH}/00-ORDER | uniq > ${TMPDIR}/00-ORDER.sorted
        fi

        sort ${TMPDIR}/ls.out | uniq > ${TMPDIR}/ls.out.sorted
        diff -u ${TMPDIR}/00-ORDER.sorted ${TMPDIR}/ls.out.sorted | while read line ; do
            echo "*** ${line}"
        done

        exit 12
    fi
done

# Remove any existing autogenerated release tree files
find ${RELEASE}/ -type f -name SHA256 | xargs rm 2>/dev/null
find ${RELEASE}/ -type f -name PACKAGES | xargs rm 2>/dev/null

# Generate release tree content
mkdir -p ${RELEASE}/media/${TARGET_MACH}
mkdir -p ${RELEASE}/source
cp -p ${TOPSRC}/LICENSE ${RELEASE}
cp -p ${TOPSRC}/README ${RELEASE}
[ -f ${TOPSRC}/NEWS ] && cp -p ${TOPSRC}/NEWS ${RELEASE}
[ -f ${TOPSRC}/UPDATING ] && cp -p ${TOPSRC}/UPDATING ${RELEASE}
install -m 0755 ${TOPSRC}/doc/verify.sh ${RELEASE}/verify.sh
cat ${TOPSRC}/doc/INSTALL.in | \
    sed -e "s|%VERSION%|${DISTRIBUTION_VERSION}|g" | \
    fold -s -w 75 > ${RELEASE}/INSTALL
gen_VERSION
gen_ChangeLog

find ${RELEASE}/dist -type f -name "*.dist" | xargs dirname | sort | uniq | while read c ; do
    echo > "${c}/header"
    echo "Packages in $(echo "${c}" | sed -e "s|^$(realpath ${RELEASE})/||g")" >> "${c}/header"
    echo "$(date)" >> "${c}/header"
    echo >> "${c}/header"

    find "${c}" -type f -name '*.dist' | while read p ; do
        echo
        echo "------------------------------------------------------------"
        echo
        distinfo -A ${p} | sed -e "s|\ $(realpath ${RELEASE})/| |g"
    done >> "${c}/PACKAGES.n"

    echo "Total size of all packages:  $(grep "Package size:" "${c}/PACKAGES.n" | awk '{ total = total + $3 } END { print total }') bytes" >> "${c}/header"
    echo "Total installed size of all packages:  $(grep "Installed size:" "${c}/PACKAGES.n" | awk '{ total = total + $3 } END { print total }') bytes" >> "${c}/header"

    cat "${c}/header" "${c}/PACKAGES.n" > "${c}/PACKAGES"
    rm -f "${c}/header" "${c}/PACKAGES.n"
done

# Copy in distfiles (assumes prune-distfiles has been run).
rm -rf ${RELEASE}/source/distfiles
mkdir -p ${RELEASE}/source/distfiles
cp -pr ${DISTFILES}/* ${RELEASE}/source/distfiles

# If a tag for this release exists, use it to create the source archive,
# otherwise just create a source archive of the master branch.
cd "${TOPSRC}"

TAG=
( cd ${TOPSRC} ; git tag -l ) | grep -q -E "^v${DISTRIBUTION_VERSION}-${TARGET_MACH}$" >/dev/null 2>&1
[ ${?} -eq 0 ] && TAG="v${DISTRIBUTION_VERSION}-${TARGET_MACH}"

if [ -z "${TAG}" ]; then
    ( cd ${TOPSRC} ; git tag -l ) | grep -q -E "^v${DISTRIBUTION_VERSION}$" >/dev/null 2>&1
    [ ${?} -eq 0 ] && TAG="v${DISTRIBUTION_VERSION}"
fi

[ -z "${TAG}" ] && TAG="master"

echo -n "Creating src.tar.gz from ${TAG}..."
git archive --format=tar --prefix=src/ -o ${RELEASE}/source/src.tar ${TAG}

cd "${RELEASE}/source"
gzip -9f src.tar
echo "DONE"

# Copy in installer files and modify other default config files
ln -sf sbin/init ${CHROOT}/init
rm -f ${CHROOT}/etc/issue
rm -f ${CHROOT}/etc/issue.net
install -D -m 0644 $RELEASE/INSTALL ${CHROOT}/root/INSTALL
install -m 0755 ${TOPSRC}/inst/inst ${CHROOT}/sbin/inst
ln -sf inst ${CHROOT}/sbin/ttyinst
install -D -m 0644 ${TOPSRC}/inst/inst.8 ${CHROOT}/usr/share/man/man8/inst.8
gzip -9 ${CHROOT}/usr/share/man/man8/inst.8
install -D -m 0644 ${TOPSRC}/inst/ttyinst.8 ${CHROOT}/usr/share/man/man8/ttyinst.8
gzip -9 ${CHROOT}/usr/share/man/man8/ttyinst.8
install -m 0644 ${TOPSRC}/inst/issue ${CHROOT}/etc/issue
ln -sf issue ${CHROOT}/etc/issue.net
install -m 0644 ${TOPSRC}/inst/motd ${CHROOT}/etc/motd
install -m 0644 ${TOPSRC}/inst/rc.conf ${CHROOT}/etc/rc.conf
install -m 0644 ${TOPSRC}/inst/inst.conf ${CHROOT}/etc/inst.conf
sed -i -e 's|/sbin/agetty |/sbin/agetty -J |g' ${CHROOT}/etc/inittab
modify_fstab

# Trim the initramfs for the installer media.  We don't need a lot
# of the files we package.  We should aim for a small miniroot.img.
if [ -f ${TOPSRC}/inst/trimlist.miniroot ]; then
    grep -v "^$" ${TOPSRC}/inst/trimlist.miniroot 2>/dev/null | grep -v "^#" | while read trimline ; do
        [ -z "${trimline}" ] || rm -rfv ${CHROOT}/${trimline}
    done
fi

# Check the initramfs to see if any required libraries are missing.
echo '#!/bin/sh' > ${CHROOT}/checkit.sh
echo 'for d in bin sbin lib ; do' >> ${CHROOT}/checkit.sh
echo '    cd /usr/$d' >> ${CHROOT}/checkit.sh
echo '    for f in * ; do' >> ${CHROOT}/checkit.sh
echo '        ldd $f 2>/dev/null | grep -q "not found" >/dev/null 2>&1' >> ${CHROOT}/checkit.sh
echo '        if [ $? -eq 0 ]; then' >> ${CHROOT}/checkit.sh
echo '            echo "$f:"' >> ${CHROOT}/checkit.sh
echo '            ldd $f 2>/dev/null | grep "not found"' >> ${CHROOT}/checkit.sh
echo '        fi' >> ${CHROOT}/checkit.sh
echo '    done' >> ${CHROOT}/checkit.sh
echo 'done' >> ${CHROOT}/checkit.sh

chmod 0700 ${CHROOT}/checkit.sh
missing="$(chroot ${CHROOT} /checkit.sh)"

if [ ! -z "${missing}" ]; then
    echo "*** ERROR: missing required libraries in initramfs." >&2
    echo "${missing}" >&2
    exit 15
fi

rm -f ${CHROOT}/checkit.sh

# Create miniroot file.
# Use the kernel scripts to generate an external compressed initramfs.
if [ -z "${KERNEL_WRKSRC}" ] || [ ! -d "${KERNEL_WRKSRC}" ]; then
    echo "*** Missing base/kernel \${WRKSRC} directory." >&2
    echo "*** Building the release requires the kernel initramfs tools." >&2
    echo "*** Do a 'make world' in base/kernel, then try the 'make release' again." >&2
    exit 13
fi

echo -n "Generating compressed initramfs image (this takes a long time)..."
cd ${KERNEL_WRKSRC}
#make --no-print-directory -C usr gen_init_cpio
sh ./scripts/gen_initramfs_list.sh \
    -o $OUTPUT/initramfs.cpio.xz \
    -u squash -g squash \
    ${CHROOT}
mv ${OUTPUT}/initramfs.cpio.xz ${MINIROOT}
echo "done."

# Add additional packages to the $CHROOT that we did not want in the initramfs
# These installations are done here for the remained of the script which will pick
# and choose files from $CHROOT, but we don't actually need these files in the
# miniroot.img.  Complicated
if [ ! -z "${EXTRA_PACKAGES}" ]; then
    for extrapkg in ${EXTRA_PACKAGES} ; do
        distadd -m -r ${CHROOT} -d "${extrapkg}"
    done
fi

# Create the boot ISO for each architecture
if [ "${TARGET_MACH}" = "x86_64" ]; then
    # On x86_64, create hybrid boot ISO that uses ISOLINUX for BIOS
    # systems, grub for UEFI systems, and has isohybrid run against
    # it.  This gives us something that works as a CD, DVD, or USB
    # media.
    BOOT_LOADER_PLATFORM="x86_64-efi"

    # Set up ISOLINUX for BIOS systems
    mkdir -p ${RELEASE}/isolinux
    cp -p ${CHROOT}/usr/lib/syslinux/isolinux.bin ${RELEASE}/isolinux
    cp -p ${CHROOT}/usr/lib/syslinux/reboot.c32 ${RELEASE}/isolinux
    cp -p ${CHROOT}/usr/lib/syslinux/ldlinux.c32 ${RELEASE}/isolinux
    cp -p ${CHROOT}/usr/lib/syslinux/chain.c32 ${RELEASE}/isolinux
    cp -p ${CHROOT}/usr/lib/syslinux/menu.c32 ${RELEASE}/isolinux
    cp -p ${CHROOT}/usr/lib/syslinux/lib*.c32 ${RELEASE}/isolinux

    # Copy kernel and miniroot to the /boot directory
    mkdir -p ${RELEASE}/boot
    cp -p ${CHROOT}/boot/vmlinuz-${KERNEL_VERSION} ${RELEASE}/boot/${KERNEL}
    cp ${MINIROOT} ${RELEASE}/boot

    # Build the isolinux.cfg file
    sed -e "s|%DISTRIBUTION_VERSION%|${DISTRIBUTION_VERSION}|g" \
        -e "s|%TARGET_MACH%|${TARGET_MACH}|g" \
        ${TOPSRC}/inst/${TARGET_MACH}/bios/isolinux.cfg.in > ${RELEASE}/isolinux/isolinux.cfg

    # Generate the efiboot.img filesystem tree for UEFI systems
    mkdir -p ${TMPDIR}/efiboot/EFI/BOOT/fonts
    cp -p ${CHROOT}/usr/share/grub/unicode.pf2 ${TMPDIR}/efiboot/EFI/BOOT/fonts
    cp -p ${CHROOT}/usr/lib/grub/${BOOT_LOADER_PLATFORM}/BOOTX64-FOR-ISO-IMAGES.EFI ${TMPDIR}/efiboot/EFI/BOOT/BOOTX64.EFI
    ${CHROOT}/usr/bin/grub-mkimage -O ${BOOT_LOADER_PLATFORM} \
                                 -d ${CHROOT}/usr/lib/grub/${BOOT_LOADER_PLATFORM} \
                                 -o ${TMPDIR}/efiboot/EFI/BOOT/BOOTX64.EFI \
                                 -c ${TOPSRC}/inst/${TARGET_MACH}/uefi/grub-efiboot.img.cfg \
                                 -C xz \
                                 -p '/EFI/BOOT' \
                                 part_gpt part_msdos fat ext2 hfs hfsplus iso9660 \
                                 chain linux boot backtrace efifwsetup efinet tftp\
                                 configfile normal regexp minicmd reboot halt search \
                                 search_fs_file search_fs_uuid search_label gfxterm \
                                 gfxmenu all_video gzio echo loopback part_apple \
                                 loadenv font cat help ls png serial sleep syslinuxcfg \
                                 usb usbserial_common usbserial_ftdi usbserial_pl2303 \
                                 usbserial_usbdebug video xfs jpeg test
    sed -e "s|%DISTRIBUTION_VERSION%|${DISTRIBUTION_VERSION}|g" \
        -e "s|%TARGET_MACH%|${TARGET_MACH}|g" \
        ${TOPSRC}/inst/${TARGET_MACH}/uefi/grub.cfg.in > ${TMPDIR}/efiboot/EFI/BOOT/grub.cfg

    # We need an EFI boot tree on the actual iso filesystem because
    # the BOOTX64.EFI loader will read the embedded grub configuration
    # file which we have set to look for the PLDIST volume and use
    # that as its root, so /EFI/BOOT/grub.cfg on the iso filesystem
    # is the main one to use.
    cp -pr ${TMPDIR}/efiboot/EFI ${RELEASE}

    # Compute the size of efiboot.img
    # 17408 is the size of the GPT tables that parted creates, but there are
    # 2 of them. (Note from pjones via the mkefiboot code in the lorax project.)
    # Use rounded up MB for the size to make the disk image.  This pads the size
    # a bit to account for space needed for the disk label and filesystem.
    disksize=$(du -s -m ${TMPDIR}/efiboot | cut -f 1)

    # Create the efiboot.img file.  Start with a dd'ed file from /dev/zero,
    # then attach that to a loopback device.
    dd if=/dev/zero of=${RELEASE}/boot/efiboot.img bs=1M count=${disksize}
    lsmod | grep -q loop >/dev/null 2>&1
    [ $? -eq 1 ] && modprobe loop
    loopdev="$(losetup --find --show ${RELEASE}/boot/efiboot.img)"
    parted -a opt -s ${loopdev} mklabel gpt
    parted -a opt -s ${loopdev} mkpart ESP fat32 34816b 100%
    parted -a opt -s ${loopdev} set 1 boot on
    partprobe ${loopdev}
    mkfs.fat -f 1 -v ${loopdev}p1

    # Mount efiboot.img and copy files to it.
    tmpmount="$(mktemp -d -p ${TMPDIR})"
    mount ${loopdev}p1 ${tmpmount}
    cp -rv ${TMPDIR}/efiboot/* ${tmpmount}
    ( cd ${tmpmount} ; ls -lR )
    umount ${tmpmount}
    rmdir ${tmpmount}

    # Disconnect efiboot.img
    losetup --detach ${loopdev}

    # Generate checksum files
    gen_SUMS

    # sort ordering information for the ISO image
    SORTFILE="$(mktemp)"
    echo "/     0" > ${SORTFILE}
    echo "/boot 1" >> ${SORTFILE}

    # Create the ISO image
    rm -f ${RELEASE}/isolinux/boot.cat
    mkisofs -o ${RELEASE_ISO} -r -hide-rr-moved -v -d -N \
            -b isolinux/isolinux.bin -c isolinux/boot.cat \
            -no-emul-boot -boot-load-size 4 -boot-info-table \
            -eltorito-alt-boot -eltorito-platform efi -eltorito-boot boot/efiboot.img -no-emul-boot \
            -V "${RELEASE_LABEL}" \
            -A "Peachtree Linux" \
            -m "media" \
            -m "ppc64" \
            -sort ${SORTFILE} \
            -graft-points /=${RELEASE}/
    isohybrid --uefi ${RELEASE_ISO}

    # clean up
    rm -f ${SORTFILE}
elif [ "${TARGET_MACH}" = "ppc64" ]; then
    BOOT_LOADER_PLATFORM="powerpc-ieee1275"

    # Copy initial boot files in place
    mkdir -p ${RELEASE}/boot/${TARGET_MACH}/${BOOT_LOADER_PLATFORM}
    cp -p ${CHROOT}/boot/vmlinux-${KERNEL_VERSION} ${RELEASE}/boot/${KERNEL}
    cp ${MINIROOT} ${RELEASE}/boot
    cp -p ${TOPSRC}/inst/${TARGET_MACH}/ofboot.b ${RELEASE}/boot/${TARGET_MACH}/${BOOT_LOADER_PLATFORM}/ofboot.b

    # Create grub boot image and copy in grub modules
    ${CHROOT}/usr/bin/grub-mkimage -O ${BOOT_LOADER_PLATFORM} \
                                 -d ${CHROOT}/usr/lib/grub/${BOOT_LOADER_PLATFORM} \
                                 -p "()/boot/${TARGET_MACH}" \
                                 -o ${RELEASE}/boot/${TARGET_MACH}/${BOOT_LOADER_PLATFORM}/core.elf \
                                 part_apple part_msdos loopback \
                                 hfs hfsplus hfspluscomp ext2 fat iso9660 \
                                 search search_fs_file search_fs_uuid search_label \
                                 adler32 boot cat gzio xzio ieee1275_fb macbless \
                                 linux configfile
    cp -p ${CHROOT}/usr/lib/grub/${BOOT_LOADER_PLATFORM}/*.mod ${RELEASE}/boot/${TARGET_MACH}/${BOOT_LOADER_PLATFORM}
    cp -p ${CHROOT}/usr/lib/grub/${BOOT_LOADER_PLATFORM}/*.lst ${RELEASE}/boot/${TARGET_MACH}/${BOOT_LOADER_PLATFORM}

    # Install our grub install media config file
    sed -e "s|%DISTRIBUTION_VERSION%|${DISTRIBUTION_VERSION}|g" \
        -e "s|%TARGET_MACH%|${TARGET_MACH}|g" \
        < ${TOPSRC}/inst/${TARGET_MACH}/grub.cfg.in > ${RELEASE}/boot/${TARGET_MACH}/grub.cfg

    # Generate checksum files
    gen_SUMS

    # Create the ISO image
    mkisofs -o ${RELEASE_ISO} \
            -r -J -v -d -N -T -l \
            -chrp-boot -U -part -hfs \
            -V "${RELEASE_LABEL}" \
            -A "Peachtree Linux" \
            -m "media" \
            -m "x86_64" \
            -sysid PPC \
            -hfs-volid "${RELEASE_LABEL}" \
            -hfs-bless "${RELEASE}/boot/${TARGET_MACH}/${BOOT_LOADER_PLATFORM}" \
            -map ${TOPSRC}/inst/${TARGET_MACH}/mapping \
            -no-desktop \
            -allow-multidot \
            -graft-points /=${RELEASE}/
else
    echo "*** Unsupported target architecture: ${TARGET_MACH}" >&2
    exit 14
fi

# Generate checksums for the media files
cd "${RELEASE}/media/${TARGET_MACH}"
rm -f SHA256
openssl dgst -sha256 * > "${TMPDIR}/SHA256"
mv "${TMPDIR}/SHA256" "${RELEASE}/media/${TARGET_MACH}"

rm -rf ${TMPDIR}
exit 0
