#!/bin/sh
#
# makesum.sh: Generate additional data for the archives file.
#
# {{{ Apache License version 2.0
#
# Copyright 2011-2018 David Cantrell <david.l.cantrell@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# }}}

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $TOPSRC               The top level project directory.
#    $WRKDIR               The top level work directory.
#    $DIGEST_ALGORITHM     Message digest algorithm to use.
#    $PKGDISTFILES         The package specific distfiles subdirectory.
# This script is called by the build system scripts and should not be called
# directly.

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 1
fi

if [ -z "${WRKDIR}" ] || [ ! -d "${WRKDIR}" ]; then
    exit 2
fi

[ -z "${DIGEST_ALGORITHM}" ] && exit 3

if [ -z "${PKGDISTFILES}" ] || [ ! -d "${PKGDISTFILES}" ]; then
    exit 4
fi

CWD="$(pwd)"
INCLUDES="${WRKDIR}/archives.includes"
HANDLED="${WRKDIR}/archives.handled"

if [ -f "${CWD}/archives" ]; then
    cmd="openssl"
    cmdopts="dgst -$(echo "${DIGEST_ALGORITHM}" | tr '[A-Z]' '[a-z]')"

    ${cmd} --version >/dev/null 2>&1
    [ $? -eq 127 ] && exit 5

    # collect list of included archives files
    included="$(grep "^#include" "${CWD}/archives" | cut -d '<' -f 2 | cut -d '>' -f 1)"
    if [ ! -z "${included}" ]; then
        rm -f "${HANDLED}"

        for ifile in ${included} ; do
            cat "${TOPSRC}/${ifile}" >> "${HANDLED}"
        done
    fi

    rm -f "${WRKDIR}/archives"

    grep -v "^#" "${CWD}/archives" | grep -v "^$" | \
    while read a b ; do
        if [ ! -z "${a}" ] && [ ! -z "${b}" ]; then
            echo "${a} ${b}" >> ${WRKDIR}/archives
            continue
        fi

        archive="$(echo ${a} | cut -d '(' -f 2 | cut -d ')' -f 1)"
        if [ ! -f "${PKGDISTFILES}/${archive}" ]; then
            echo "*** Missing ${PKGDISTFILES}/${archive}, try fetch again." >&2
            exit 6
        fi

        if [ -f "${HANDLED}" ] && [ ! -z "${b}" ]; then
            grep -q -E "${b}$" "${HANDLED}" >/dev/null 2>&1
            [ $? -eq 0 ] && continue
        fi

        ( cd ${PKGDISTFILES} ; ${cmd} ${cmdopts} ${archive} ) >> ${WRKDIR}/archives
    done

    # keep any include lines
    grep "^#include" "${CWD}/archives" > "${INCLUDES}"

    if [ -f "${WRKDIR}/archives" ]; then
        if [ -s "${INCLUDES}" ]; then
            # remove expanded lines that we are including
            while read x ; do
                oa="$(echo "${x}" | cut -d '<' -f 2 | cut -d '>' -f 1)"
                [ -f ${TOPSRC}/${oa} ] || continue

                while read y ; do
                    grep -v "^${y}" "${WRKDIR}/archives" > "${WRKDIR}/a.new"
                    [ -f "${WRKDIR}/a.new" ] && mv "${WRKDIR}/a.new" "${WRKDIR}/archives"
                done < ${TOPSRC}/${oa}
            done < ${INCLUDES}

            cat "${INCLUDES}" > ${CWD}/archives
            rm -f "${INCLUDES}"
        else
            rm -f ${CWD}/archives
        fi

        # only take the fully expanded lines beginning with an algorithm
        grep -E "\(.+\)=" "${WRKDIR}/archives" | sort | uniq >> ${CWD}/archives
        rm -rf ${WRKDIR}/archives
    fi

    rm -rf "${HANDLED}"
fi
