#!/bin/sh
#
# merge-releases.sh: Combine multiple release trees in to $RELEASE
#
# {{{ Apache License version 2.0
#
# Copyright 2017 David Cantrell <david.l.cantrell@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# }}}

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $RELEASE              The destination /usr/release tree.
#    $OTHER_RELEASES       Comma-delimited list of other /usr/release trees.
# This script is called by the build system scripts and should not be called
# directly.

if [ -z "${RELEASE}" ] || [ ! -d "${RELEASE}" ]; then
    exit 1
fi

if [ -z "${OTHER_RELEASES}" ]; then
    echo "*** Missing OTHER_RELEASES environment variable." >&2
    exit 2
fi

CWD="$(pwd)"
TMPDIR="$(mktemp -d)"

for other in $(echo "${OTHER_RELEASES}" | tr ',' '\n') ; do
    if [ ! -d "${other}" ]; then
        echo "*** ${other} is not a directory, skipping." >&2
        continue
    fi

    # Copy in all new and different files, backup existing ones to have the
    # same name but ending with ~
    rsync -abviuP ${other}/ ${RELEASE}/

    # Iterate over all files ending with ~ and eliminate dupes and combine
    # things we can combine (text files we know the format for).
    find ${RELEASE}/ -type f -name "*~" | while read oldfile ; do
        newfile="$(echo "${oldfile}" | rev | cut -c2- | rev)"
        oldsha="$(openssl dgst -sha256 ${oldfile} | rev | cut -d ' ' -f 1 | rev)"
        newsha="$(openssl dgst -sha256 ${newfile} | rev | cut -d ' ' -f 1 | rev)"

        if [ "${oldsha}" = "${newsha}" ]; then
            rm "${oldfile}"
        else
            distinfo -Ss "${oldfile}" > ${TMPDIR}/vars
            unset SIZE
            . ${TMPDIR}/vars
            oldsize="${SIZE}"

            distinfo -Ss "${newfile}" > ${TMPDIR}/vars
            unset SIZE
            . ${TMPDIR}/vars
            newsize="${SIZE}"

            if [ "${oldsize}" = "${newsize}" ]; then
                rm "${oldfile}"
            else
                echo "${newfile} contains differences, unable to resolve"
            fi
        fi
    done
done

rm -rf "${TMPDIR}"
