#!/bin/sh
#
# order.sh: Sort all of the packages in the specified collection in
#           build dependency order and save that order to files.
#
# {{{ Apache License version 2.0
#
# Copyright 2018 David Cantrell <david.l.cantrell@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# }}}

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $TOPSRC               The top level distribution source directory.
#    $COLLECTION           The name of the collection.
#    $MAKE                 The make command.
# This script is called by the build system scripts and should not be called
# directly.

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 1
fi

if [ -z "${COLLECTION}" ] || [ ! -d "${TOPSRC}/${COLLECTION}" ]; then
    exit 2
fi

TMPDIR="$(mktemp -d)"
ALLPKGS="${TMPDIR}/allpkgs"
TSORTED="${TMPDIR}/tsorted"
BUILT="${TMPDIR}/built"

makecmd="env WORLD=y ${MAKE} --no-print-directory"

get_all_packages() {
    find ${TOPSRC}/${COLLECTION} -type f -name pkginfo | \
    while read package ; do
        pkgpath="$(dirname ${package} | sed -e "s|^${TOPSRC}/||g")"
        skip="$(${makecmd} -C ${pkgpath} show var=SKIP)"
        [ "$(echo ${skip} | tr [A-Z] [a-z])" = "y" ] || echo "${pkgpath}"
    done
}

get_edges() {
    for package in $(get_all_packages) ; do
        deps="$(${MAKE} -C ${package} show var=BUILD_DEPENDS)"

        if [ -z "${deps}" ] && [ -f "${package}/buildreqs" ]; then
            deps="$(${MAKE} -C ${package} show-buildreqs)" || exit ${?}
        fi

        if [ ! -z "${deps}" ]; then
            for dep in ${deps} ; do
                echo "${package} ${dep}"
                echo "${package} ${package}"
            done
        fi
    done
}

# Generate build dependency order for this collection.
echo -n "Generating build order for the ${COLLECTION} collection..."

rm -f ${BUILT}
rm -f ${ALLPKGS}
get_all_packages > ${ALLPKGS}
get_edges | tsort > "${TSORTED}"

# See if we have any circular dependencies
if [ ${?} -eq 1 ]; then
    echo
    echo >&2
    echo "*** BUILD_DEPENDS are circular." >&2
    echo "*** See 'make show-order COLLECTION=[name]' for more info." >&2
    echo "*** Also, consider backing out recent BUILD_DEPENDS changes to see" >&2
    echo "*** where the loop was introduced." >&2
    echo >&2
    exit 1
fi

[ -d ${TOPSRC}/mk/order ] || mkdir -p ${TOPSRC}/mk/order
tac "${TSORTED}" | grep "^${COLLECTION}" > ${TOPSRC}/mk/order/${COLLECTION}.list
echo "done."
rm -rf ${TMPDIR}

exit 0
