#!/bin/sh
#
# package.sh: Implement package stage.
#
# {{{ Apache License version 2.0
#
# Copyright 2011-2018 David Cantrell <david.l.cantrell@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# }}}

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $WRKDIR               The top level work directory.
#    $WRKINST              The staging root directory.
#    $WRKDEBUG             The debug directory.
#    $PKG                  The package name.
#    $VER                  The package version.
#    $REL                  The package release.
#    $TARGET_NAME          The target arch name (e.g., vax-peachtree-linux).
#    $TARGET_MACH          arch(1) or 'uname -m' value (or 'noarch').
#    $TOPSRC               The top level directory of the project.
#    $RELEASE_DIST         The release tree directory of the project.
#    $PKGCOLLECTION        The name of the package collection.
#    $NAME_PREFIX_CONFLICT Set to a list of package names that name $PKG as
#                          their package name prefix.
#    $MULTIPLE_VERSIONS    (y/n), y=There a multiple versions of this $PKG
#                          in this $PKGCOLLECTION (e.g., gtk+)
#    $DIGEST_ALGORITHM     The digest algorithm to use (e.g., sha1) when
#                          generating message digests for built packages.
#    $COMPRESSION          Compression algorithm to use in dist payloads.  It
#                          is lzma by default, but that's expensive on slower
#                          processors.  Packages can override it by setting
#                          this variable to an accepted distmk(1) -z setting.
# This script is called by the build system scripts and should not be called
# directly.

if [ -z "${WRKDIR}" ] || [ ! -d "${WRKDIR}" ]; then
    exit 1
fi

if [ -z "${WRKINST}" ] || [ ! -d "${WRKINST}" ]; then
    exit 2
fi

[ -z "${PKG}" ] && exit 3
[ -z "${VER}" ] && exit 4
[ -z "${REL}" ] && exit 5
[ -z "${TARGET_MACH}" ] && exit 6

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 7
fi

if [ -z "${PKGCOLLECTION}" ] || [ ! -d "${TOPSRC}/${PKGCOLLECTION}" ]; then
    exit 8
fi

[ -z "${DIGEST_ALGORITHM}" ] && exit 9
[ -z "${RELEASE_DIST}" ] && exit 12
[ -z "${COMPRESSION}" ] && exit 13

CWD="$(pwd)"
TMPDIST="$(mktemp -d)"
LOG="${WRKDIR}/package.log"
ERR="${WRKDIR}/package.err"
DEBUGLOG="${WRKDIR}/package-debug.log"
DEBUGERR="${WRKDIR}/package-debug.err"

DISTDIR="${RELEASE_DIST}/${PKGCOLLECTION}/${TARGET_MACH}"
DEBUGDIR="${RELEASE_DIST}/debug/${PKGCOLLECTION}/${TARGET_MACH}"

# Message digest command
digestcmd="openssl"
digestopts="dgst -$(echo ${DIGEST_ALGORITHM} | tr [A-Z] [a-z])"
${digestcmd} --version >/dev/null 2>&1
[ $? -eq 127 ] && exit 10

escape_pkgname() {
    echo "${1}" | sed -e 's|\+|\\\+|g' -e 's|\-|\\\-|g'
}

remove_old_pkg() {
    _pkgdir="${1}"
    _curpkg="${2}"
    _msg="${3}"

    ls -1 ${_pkgdir}/${PKG}-*.dist 2>/dev/null | while read line ; do
        pkgname="$(basename ${line})"

        # Special handling for package names that confuse the wildcards
        # or that otherwise need special handling.
        if [ ! -z "${NAME_PREFIX_CONFLICT}" ]; then
            conflict=0
            for c in ${NAME_PREFIX_CONFLICT} ; do
                c="$(escape_pkgname ${c})"
                echo "${pkgname}" | grep -E -s -q "^${c}"
                if [ $? -eq 0 ]; then
                    conflict=1
                    break
                fi
            done
            [ ${conflict} -eq 1 ] && continue
        elif [ "$(echo ${MULTIPLE_VERSIONS} | tr [A-Z] [a-z])" = "y" ]; then
            c="$(escape_pkgname "${PKG}-")"
            echo "${pkgname}" | grep -E -s -q "^${c}"
            if [ $? -eq 0 ]; then
                majorver="$(echo ${VER} | cut -d '.' -f 1)"
                echo "${pkgname}" | grep -E -s -q "^${c}${majorver}"
                [ $? -eq 1 ] && continue
            fi
        fi

        if [ ! "${pkgname}" = "${_curpkg}" ]; then
            echo -n "Removing old ${_msg} ${pkgname}..."
            rm -f ${line}

            if [ -f ${_pkgdir}/repodata.db ]; then
                sqlite3 ${_pkgdir}/repodata.db "DELETE FROM dists WHERE filename='${pkgname}'"
            fi

            if [ -f ${_pkgdir}/00-ORDER ]; then
                sed -i -e "s|^${pkgname}$|${_curpkg}|g" ${_pkgdir}/00-ORDER
            fi

            if [ -f ${_pkgdir}/00-MULTI ]; then
                sed -i -e "s|^${pkgname}$|${_curpkg}|g" ${_pkgdir}/00-MULTI
            fi

            echo "done."
        fi
    done
}

init_repodata_db() {
    DISTDIR="${1}"
    [ -d ${DISTDIR} ] || return 1
    sqlite3 ${DISTDIR}/repodata.db 'CREATE TABLE IF NOT EXISTS
                                    dists (filename TEXT NOT NULL UNIQUE,
                                           pkg TEXT NOT NULL,
                                           ver TEXT NOT NULL,
                                           rel INTEGER NOT NULL,
                                           repo TEXT NOT NULL,
                                           arch TEXT NOT NULL,
                                           desc TEXT,
                                           maintainer TEXT,
                                           license TEXT,
                                           pkgsize UNSIGNED BIG INT NOT NULL,
                                           size UNSIGNED BIG INT NOT NULL,
                                           debug BOOLEAN,
                                           digest TEXT NOT NULL);'
}

add_repodata() {
    DISTDIR="${1}"
    curpkg="${2}"

    [ -d ${DISTDIR} ] || exit 1
    [ -f ${DISTDIR}/${curpkg} ] || exit 2

    ( cd ${DISTDIR}
      eval $(distinfo -A -S "${curpkg}")
      [ "${DEBUG}" = "y" ] && DEBUG=true || DEBUG=false
      DIGEST="$(${digestcmd} ${digestopts} "${curpkg}" | cut -d ' ' -f 2)"
      sqlite3 "${DISTDIR}/repodata.db" "INSERT OR REPLACE INTO
                                        dists (filename, pkg, ver, rel, license,
                                               repo, arch, maintainer, desc,
                                               pkgsize, size, debug, digest)
                                        VALUES ('${FILENAME}', '${PKG}', '${VER}',
                                                '${REL}', '${LICENSE}', '${REPO}',
                                                '${ARCH}', '${MAINTAINER}', '${DESC}',
                                                '${PKGSIZE}', '${SIZE}', '${DEBUG}',
                                                '${DIGEST}')"
    )
}

# Force packages without ELF content to have TARGET_MACH=noarch
if [ "${TARGET_MACH}" = "noarch" ]; then
    find ${WRKINST} -type f -exec file -b {} \; | grep -s ELF >/dev/null 2>&1
    if [ $? -eq 0 ]; then
        echo "*** ELF files found in WRKINST but TARGET_MACH = noarch." >&2
        exit 11
    fi
fi

[ -d ${DISTDIR} ] || mkdir -p ${DISTDIR}
[ -f ${DISTDIR}/repodata.db ] || init_repodata_db ${DISTDIR}

# Combine multiple plist files for the package and replace variables
PLIST=
if [ -f "${CWD}/plist" ] && [ ! -f "${CWD}/plist.${TARGET_MACH}" ]; then
    PLIST="${CWD}/plist"
elif [ ! -f "${CWD}/plist" ] && [ -f "${CWD}/plist.${TARGET_MACH}" ]; then
    PLIST="${CWD}/plist.${TARGET_MACH}"
elif [ -f "${CWD}/plist" ] && [ -f "${CWD}/plist.${TARGET_MACH}" ]; then
    PLIST="${CWD}/plist ${CWD}/plist.${TARGET_MACH}"
else
    # For new packages, we may not have a plist file yet, so just use an
    # empty file here to move things along for the packager.
    touch ${TMPDIST}/emptyplist
    PLIST="${TMPDIST}/emptyplist"
fi

cat ${PLIST} | sort | uniq | \
sed -e "s|\${VER}|${VER}|g" | \
sed -e "s|\${REL}|${REL}|g" | \
sed -e "s|\${TARGET_NAME}|${TARGET_NAME}|g" > ${TMPDIST}/plist

if [ ! "${TARGET_MACH}" = "noarch" ]; then
    sed -i -e "s|\${TARGET_MACH}|${TARGET_MACH}|g" ${TMPDIST}/plist
fi

# Create the dist package
( distmk -t ${WRKINST} -p ${PKG} -v ${VER} -r ${REL} -l ${TMPDIST}/plist \
      -m ${CWD} -a ${TARGET_MACH} -o ${DISTDIR} -R ${PKGCOLLECTION} -c -z ${COMPRESSION} \
  2>&1 1>&3 | tee ${ERR} \
) 3>&1 1>&2 | tee ${LOG} || exit $?

curpkg="${PKG}-${VER}-${REL}.${TARGET_MACH}.dist"
add_repodata ${DISTDIR} ${curpkg}

# Remove old packages
remove_old_pkg ${DISTDIR} ${curpkg} "package"

# Is this a multiple version package?  If so, note it.
if [ "$(echo ${MULTIPLE_VERSIONS} | tr [A-Z] [a-z])" = "y" ]; then
    echo "${curpkg}" >> "${DISTDIR}/00-MULTI"
    sort "${DISTDIR}/00-MULTI" | uniq > "${DISTDIR}/00-MULTI.new"
    mv "${DISTDIR}/00-MULTI.new" "${DISTDIR}/00-MULTI"
fi

# Do we need a debug package?
if [ -d ${WRKDEBUG} ]; then
    [ -d ${DEBUGDIR} ] || mkdir -p ${DEBUGDIR}
    [ -f ${DEBUGDIR}/repodata.db ] || init_repodata_db ${DEBUGDIR}

    distplist -t ${WRKDEBUG} -u root -g root > ${TMPDIST}/plist.debug
    ( distmk -t ${WRKDEBUG} -p ${PKG} -v ${VER} -r ${REL} -d \
          -l ${TMPDIST}/plist.debug -m ${CWD} -a ${TARGET_MACH} \
          -o ${DEBUGDIR} -R ${PKGCOLLECTION} -c \
      2>&1 1>&3 | tee ${DEBUGERR} \
    ) 3>&1 1>&2 | tee ${DEBUGLOG} || exit $?

    curpkg="${PKG}-${VER}-${REL}.debug.${TARGET_MACH}.dist"
    add_repodata ${DEBUGDIR} ${curpkg}

    # Remove old debug packages
    remove_old_pkg ${DEBUGDIR} ${curpkg} "debug package"
fi

rm -rf ${TMPDIST}

[ -s "${LOG}" ] || rm -f "${LOG}"
[ -s "${DEBUGLOG}" ] || rm -f "${DEBUGLOG}"
[ -s "${ERR}" ] || rm -f "${ERR}"
[ -s "${DEBUGERR}" ] || rm -f "${DEBUGERR}"
