#!/bin/sh
#
# prune-distfiles.sh: Implement prune-distfiles target.
#
# {{{ Apache License version 2.0
#
# Copyright 2012-2017 David Cantrell <david.l.cantrell@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# }}}

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $WRKDIR               The top level work directory.
#    $DISTFILES            The top level distfiles directory.
#    $PKGDISTFILES         The package specifc distfiles subdirectory.
#    $SHARED_DISTFILES     List of packages sharing distfiles locations.
#    $TOPSRC               The top level of the source tree.
# This script is called by the build system scripts and should not be called
# directly.

if [ -z "${WRKDIR}" ] || [ ! -d "${WRKDIR}" ]; then
    exit 1
fi

if [ -z "${DISTFILES}" ] || [ ! -d "${DISTFILES}" ]; then
    exit 2
fi

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 3
fi

CWD="$(pwd)"

# Detect format of the archives file and filter it down to just
# a list of the archive filenames.
filter_distfiles() {
    grep -q -E "\(.+\)=" "${1}" >/dev/null 2>&1
    if [ ${?} -eq 0 ]; then
        # archives file is in openssl(1) dgst format
        grep -vE "(^#|^$)" "${1}" | cut -d '(' -f 2 | cut -d ')' -f 1 | sort | uniq
    else
        # archives file is in legacy Peachtree Linux format
        grep -vE "(^#|^$)" "${1}" | rev | cut -d ' ' -f 1 | rev | sort | uniq
    fi
}

if [ -f "${WRKDIR}/archives" ]; then
    if [ -z "${PKGDISTFILES}" ] || [ ! -d "${PKGDISTFILES}" ]; then
        exit 4
    fi

    keep="$(mktemp -d -p "${PKGDISTFILES}")"

    filter_distfiles "${WRKDIR}/archives" | \
    while read archive ; do
        echo "Keeping ${archive}"
        mv "${PKGDISTFILES}/$archive" "${keep}"
    done

    if [ ! -z "${SHARED_DISTFILES}" ]; then
        for shared in ${SHARED_DISTFILES} ; do
            [ -f ${TOPSRC}/${shared}/archives ] || exit 5
            filter_distfiles "${TOPSRC}/${shared}/archives" | \
            while read archive ; do
                echo "Keeping ${archive}"
                mv "${PKGDISTFILES}/${archive}" "${keep}"
            done
        done
    fi

    basekeep="$(basename ${keep})"
    cd "${PKGDISTFILES}"
    for f in * ; do
        bf="$(basename ${f})"
        if [ "${bf}" = "${basekeep}" ] || [ "${bf}" = "SHA256" ]; then
            continue
        fi
        echo "Removing ${f}"
        rm "${f}"
    done

    mv ${keep}/* "${PKGDISTFILES}"
    rmdir "${keep}"
fi
