#!/bin/sh
#
# test.sh: Implement test stage.
#
# {{{ Apache License version 2.0
#
# Copyright 2011-2018 David Cantrell <david.l.cantrell@gmail.com>
#                     Chris Lumens <chris@bangmoney.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# }}}

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $WRKDIR               The top level work directory.
#    $WRKSRC               The extracted source directory.
#    $WRKBLD               The source build directory.
#    $PACKAGE              The full package name.
#    $CONFIGURE_STYLE      The method used to configure $WRKSRC for builds.
#    $PYTHON               Python interpreter to use.
#    $SEPARATE_BUILD       (y/n) Use source build directories?
#    $TEST_ENV             Environment variables passed to $(TEST_COMMAND).
#    $TEST_COMMAND         The make command.
#    $TEST_TARGET          The name of the test target in the Makefile.
#    $TEST_ARGS            Extra arguments passed to $(TEST_COMMAND).
#    $SKIP_TEST            (y/n) Run tests?
#    $TEST_COMMAND         The make command.
# This script is called by the build system scripts and should not be called
# directly.

[ "${SKIP_TEST}" = "y" ] && exit 0

if [ -z "${WRKDIR}" ] || [ ! -d "${WRKDIR}" ]; then
    exit 1
fi

if [ "${SEPARATE_BUILD}" = "y" ]; then
    if [ -z "${WRKBLD}" ] || [ ! -d "${WRKBLD}" ]; then
        exit 2
    fi
else
    if [ -z "${WRKSRC}" ] || [ ! -d "${WRKSRC}" ]; then
        exit 3
    fi
fi

CWD="$(pwd)"
LOG="${WRKDIR}/test.log"
ERR="${WRKDIR}/test.err"

echo ">>> Testing ${PACKAGE}..."

if [ "${CONFIGURE_STYLE}" = "python" ]; then
    cd ${WRKSRC}

    ( env ${TEST_ENV} \
      ${PYTHON} setup.py ${TEST_TARGET} ${TEST_ARGS} \
      2>&1 1>&3 | tee ${ERR} \
    ) 3>&1 1>&2 | tee ${LOG} || exit $?
elif [ "${CONFIGURE_STYLE}" = "meson" ]; then
    cd ${WRKSRC}

    ( env ${TEST_ENV} ninja -C ${WRKBLD} ${TEST_TARGET} ${TEST_ARGS} \
      2>&1 1>&3 | tee ${ERR} \
    ) 3>&1 1>&2 | tee ${LOG} || exit $?
elif [ "${CONFIGURE_STYLE}" = "cargo" ]; then
    cd "${WRKBLD}"
    ( env ${TEST_ENV} cargo test -v --release ${TEST_ARGS}
      2>&1 1>&3 | tee ${ERR} \
    ) 3>&1 1>&2 | tee ${LOG} || exit $?
else
    if [ -d "${WRKBLD}" ] || [ -d "${WRKSRC}" ]; then
        [ "${SEPARATE_BUILD}" = "y" ] && D=${WRKBLD} || D=${WRKSRC}

        ( env ${TEST_ENV} \
          ${TEST_COMMAND} -C ${D} ${TEST_TARGET} ${TEST_ARGS} \
          2>&1 1>&3 | tee ${ERR} \
        ) 3>&1 1>&2 | tee ${LOG} || exit $?
    fi
fi

[ -s "${LOG}" ] || rm -f "${LOG}"
[ -s "${ERR}" ] || rm -f "${ERR}"
