#!/bin/sh
#
# update-keyring.sh: If GPGKEYS is set, fetch and store to PKG.gpg
#
# {{{ Apache License version 2.0
#
# Copyright 2017-2018 David Cantrell <david.l.cantrell@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# }}}

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $GPGKEYS              The GPG key ID(s) or URI(s) to a key file.
#    $KEYRING              The full path to the keyring file to use.
#    $TRUSTDB              The full path to the GPG trust database file.
#    $GPG                  The gpg command and common options to use.
# This script is called by the build system scripts and should not be called
# directly.

[ -z "${GPGKEYS}" ] && exit 0
[ -z "${KEYRING}" ] && exit 1
[ -z "${TRUSTDB}" ] && exit 2
[ -z "${GPG}" ] && exit 3

CWD="$(pwd)"
CURL="curl -A Mozilla -k -f -L --retry 3 --ftp-pasv"
GPGCMD="${GPG} --no-default-keyring --keyring ${KEYRING} --trustdb-name ${TRUSTDB}"

# Helper function that trusts newly imported keys.
trust_keys() {
    # Tips from here:
    # https://blog.tersmitten.nl/how-to-ultimately-trust-a-public-key-non-interactively.html
    ${GPGCMD} --list-keys --fingerprint --with-colons | \
        sed -E -n -e 's/^fpr:::::::::([0-9A-F]+):$/\1:6:/p' | \
        ${GPGCMD} --import-ownertrust
    return $?
}

# Start fresh
rm -f ${KEYRING} ${TRUSTDB}

# If the value of GPGKEYS is a URI, fetch that file and import it to the
# keyring. Some projects may not publish their keys to a PGP keyserver, so
# we need to get the key manually and create the keyring.
#
# Otherwise, we pull the key from the PGP keyserver and import it.
for key in ${GPGKEYS} ; do
    if [ "$(echo "${key}" | cut -c-6)" = "ftp://" ] || [ "$(echo "${key}" | cut -c-7)" = "http://" ] || [ "$(echo "${key}" | cut -c-8)" = "https://" ]; then
        TMPDIR="$(mktemp -d)"
        ${CURL} -o ${TMPDIR}/key ${key} || exit 4
        ${GPGCMD} --import ${TMPDIR}/key || exit 5
        trust_keys || exit 6
        rm -rf ${TMPDIR}
    elif [ "$(echo "${key}" | cut -c-6)" = "hkp://" ]; then
        trimmed="$(echo "${key}" | cut -c7-)"
        keyserver="$(echo "${trimmed}" | cut -d '/' -f 1)"
        key="$(echo "${trimmed}" | cut -d '/' -f 2)"

        # Fetch the signing key for this package from a specific keyserver
        ${GPGCMD} --keyserver hkp://${keyserver} --recv-keys ${key}
        [ ${?} -eq 1 ] && exit 7

        # Note that we trust this key when verifying signatures
        trust_keys || exit 8
    else
        # Fetch the public signing key for this package
        ${GPGCMD} --keyserver hkp://keys.gnupg.net --recv-keys ${key}
        [ ${?} -eq 1 ] && exit 9

        # Note that we trust this key when verifying signatures
        trust_keys || exit 10
    fi
done

chmod 0644 ${KEYRING} ${TRUSTDB}
rm -f ${KEYRING}~

exit 0
