#!/bin/sh
#
# usereqs.sh: Install packages listed in buildreqs to the sysroot tree
#             specified.
#
# {{{ Apache License version 2.0
#
# Copyright 2011-2017 David Cantrell <david.l.cantrell@gmail.com>
#                     Chris Lumens <chris@bangmoney.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# }}}

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $TOPSRC               The top level distribution source directory.
#    $WRKDIR               The package work directory.
#    $WRKSYSROOT           The directory passed to --sysroot on gcc(1).
#    $MAKE                 The make(1) command.
#    $DIST_DOWNLOAD        Try to download build dependencies (y/n).
#    $DIST_URL             Site to download build dependencies from.
#    $RELEASE              The release directory containing built packages.
#    $DISTADD_TMPDIR       Temp directory for distadd(8) to use.
# This script is called by the build system scripts and should not be called
# directly.

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 1
fi

[ -z "${WRKDIR}" ] && exit 2
[ -z "${WRKSYSROOT}" ] && exit 3
[ -z "${MAKE}" ] && exit 4
[ -z "${DISTADD_TMPDIR}" ] && exit 5

[ -d "${DISTADD_TMPDIR}" ] || mkdir -p "${DISTADD_TMPDIR}"
CWD="$(pwd)"

# Some URLs may end with a slash, some may not.  Be careful not
# to add an extra slash on the end.
echo "${DIST_URL}" | grep -q "\/$" 2>/dev/null
[ ${?} -eq 0 ] && slash="" || slash="/"

# Gather list of buildreqs
if [ -f "${CWD}/buildreqs" ]; then
    BUILDREQS="$(${MAKE} show-buildreqs)" || exit ${?}
fi

[ -z "${BUILDREQS}" ] && exit 0

# We use a single sysroot, so wipe it and start a new one.
rm -rf "${WRKSYSROOT}"
mkdir -p "${WRKSYSROOT}"

# distadd (standard Peachtree Linux location is /usr/sbin)
distadd="/usr/sbin/distadd"

# Find curl, or disable downloading
CURL="curl -A Mozilla -k -f -L --retry 3 --ftp-pasv"
${CURL} >/dev/null 2>&1
[ ${?} -eq 127 ] && DIST_DOWNLOAD=n

[ -d "${WRKDIR}" ] || mkdir -p "${WRKDIR}"
LOG="${WRKDIR}/usereqs.log"
ERR="${WRKDIR}/usereqs.err"

# Generic variable gathering function for packages.
get_var() {
    env -u TARGET_MACH ${MAKE} -C ${TOPSRC}/${1} show var=${2}
}

# Called by _download_helper() to perform digest verification.
_verify_digest() {
    path="${1}"
    package="${2}"
    algorithm="${3}"
    digest="${4}"

    [ -z "${digest}" ] && return 1

    # what digest command are we using
    digestcmd="openssl"
    digestopts="dgst -$(echo ${algorithm} | tr [A-Z] [a-z])"
    ${digestcmd} --version >/dev/null 2>&1
    [ ${?} -eq 127 ] && return

    computed="$(${digestcmd} ${digestopts} "${path}/${package}" 2>/dev/null | cut -d ' ' -f 2)"
    [ "${computed}" = "${digest}" ] && return 0 || return 1
}

# Called by download_dist() to fetch regular and debug packages.
_download_helper() {
    package="${1}"
    path="${2}"
    collection="${3}"
    algorithm="${4}"
    webdir="${5}"
    digest=

    [ -d "${path}" ] || mkdir -p "${path}"

    if [ ! -f "${path}/repodata.db" ]; then
        echo ">>> Downloading repodata for ${collection}..."
        ${CURL} -o "${path}/repodata.db" "${DIST_URL}${slash}/${webdir}/${collection}/repodata.db"
    fi

    if [ -f "${path}/repodata.db" ]; then
        digest="$(sqlite3 "${path}/repodata.db" "SELECT digest FROM dists WHERE filename='${package}'")"
    fi

    if [ -f "${path}/${package}" ]; then
        _verify_digest ${path} ${package} ${algorithm} ${digest} && return
    else
        echo ">>> Downloading ${package}..."
        ${CURL} -o "${path}/${package}" "${DIST_URL}${slash}/${webdir}/${collection}/${package}"

        if [ ${?} -ne 0 ]; then
            echo ">>> FAILED"

            if [ -f "${path}/repodata.db" ]; then
                sqlite3 "${path}/repodata.db" "DELETE FROM dists WHERE filename='${package}'"
            fi

            rm -f "${path}/${package}"
            return
        fi
    fi

    _verify_digest ${path} ${package} ${algorithm} ${digest} && return

    if [ ${?} -ne 0 ]; then
        if [ -f "${path}/repodata.db" ]; then
            sqlite3 "${path}/repodata.db" "DELETE FROM dists WHERE filename='${package}'"
        fi

        rm -f "${path}/${package}"
        return
    fi
}

# Attempt a download and cache of this dep's built packages.
download_dist() {
    [ "$(echo ${DIST_DOWNLOAD} | tr [A-Z] [a-z])" = "n" ] && return

    # many variables
    dep="${1}"
    package="$(get_var ${dep} PACKAGE)"
    debugpackage="$(get_var ${dep} DEBUGPACKAGE)"
    collection="$(get_var ${dep} PKGCOLLECTION)"
    algorithm="$(get_var ${dep} DIGEST_ALGORITHM)"
    pkgpath="$(dirname $(get_var ${dep} FULLPKGPATH))"
    debugpkgpath="$(dirname $(get_var ${dep} FULLDEBUGPKGPATH))"

    # try to download the main dist
    _download_helper ${package} ${pkgpath} ${collection} ${algorithm} "dist"

    # try to download the debug dist (this is not fatal if it's missing)
    _download_helper ${debugpackage} ${debugpkgpath} ${collection} ${algorithm} "dist/debug"
}

# Step one:  Verify all build dependencies exist.
( for dep in ${BUILDREQS} ; do
      fullpkgpath="$(get_var ${dep} FULLPKGPATH)"

      [ -f ${fullpkgpath} ] || download_dist ${dep}

      if [ ! -f ${fullpkgpath} ]; then
          echo "*** Missing ${fullpkgpath}, cannot continue." >&2
          exit 4
      fi
  done \
  2>&1 1>&3 | tee ${ERR} \
) 3>&1 1>&2 | tee ${LOG} || exit ${?}

# Step two:  Install build dependencies.
( for dep in ${BUILDREQS} ; do
      fullpkgpath="$(get_var ${dep} FULLPKGPATH)"
      multiple_versions="$(get_var ${dep} MULTIPLE_VERSIONS)"

      [ "$(echo ${multiple_versions} | tr [A-Z] [a-z])" = "y" ] && MULTIOPT="-m" || MULTIOPT=
      ${distadd} --tmpdir=${DISTADD_TMPDIR} ${MULTIOPT} -S -d -r ${WRKSYSROOT} ${fullpkgpath}
  done \
  2>&1 1>&3 | tee -a ${ERR} \
) 3>&1 1>&2 | tee -a ${LOG} || exit ${?}

[ -s "${LOG}" ] || rm -f "${LOG}"
[ -s "${ERR}" ] || rm -f "${ERR}"
