#
# Master definitions
#
# {{{ Apache License version 2.0
#
# Copyright 2002-2017 David Cantrell <david.l.cantrell@gmail.com>
#                     Elliott Potter <epotter.e.potter@nasa.gov>
#                     David Shea <david@reallylongword.org>
#                     Chris Lumens <chris@bangmoney.org>
#                     Michael Imamura <zoogie@lugatgt.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# }}}

# Site-specific overrides to build system settings.
-include /etc/default/make.conf
-include /etc/make.conf
-include $(HOME)/.peachtree/make.conf

# Commands
# We only need to list commands here if we want to explicitly call a
# particular one.  Keep this list small.
ECHO := /bin/echo

# Current working directory
CWD := $(shell pwd)

# Distribution name
DISTRIBUTION_NAME = Peachtree Linux

# Distribution version (only for non-kernel builds)
DISTRIBUTION_VERSION = 3.4.0

# Target architecture (can be set in a package Makefile to 'noarch')
ifndef TARGET_MACH
TARGET_MACH = $(shell uname -m | sed -e "s/i[[:digit:]]86/i386/g")
endif

# For noarch packages, this is "noarch-peachtree-linux", but we don't care.
TARGET_NAME = $(TARGET_MACH)-peachtree-linux

# Kernel architecture name (used in the kernel source directory)
ifeq ($(TARGET_MACH),x86_64)
KERN_ARCH = x86
endif

ifeq ($(TARGET_MACH),ppc64)
KERN_ARCH = powerpc
endif

# Number of parallel make processes to run (equals # of CPUs)
# If the nproc command is not found, set a default value.
PARALLEL_JOBS ?= $(shell nproc 2>/dev/null)
ifeq ($(PARALLEL_JOBS),)
PARALLEL_JOBS := 3
endif

# Common make options
MAKEFLAGS += --no-print-directory

# Find the top level directory if we don't already know it.
ifndef TOPSRC
TOPSRCERR = "*** Unable to find top level directory."
TOPSRC := $(shell while [ ! -f mk/pkg.mk ]; do if [ "$$(pwd)" = "/" ]; then $(ECHO) $(TOPSRCERR) >&2; exit 1; fi; cd ..; done; $(ECHO) "$$(pwd)")
endif

MKBINDIR = $(TOPSRC)/mk/bin

# Distribution-wide helper scripts
worldbuilder  := $(MKBINDIR)/world-builder.sh
cleanchroot   := $(MKBINDIR)/clean-chroot.sh
installchroot := $(MKBINDIR)/install-chroot.sh
makerelease   := $(MKBINDIR)/make-release.sh
showorphans   := $(MKBINDIR)/show-orphans.sh
parsebool     := $(MKBINDIR)/parsebool.sh
mergereleases := $(MKBINDIR)/merge-releases.sh
incver        := $(MKBINDIR)/incver.sh
order         := $(MKBINDIR)/order.sh

# Paths
RELEASE_BASE   ?= release
DISTFILES_BASE ?= distfiles
OBJ_BASE       ?= obj

ifndef RELEASE
RELEASE := $(shell readlink -f $(TOPSRC)/$(RELEASE_BASE))
endif

ifndef DISTFILES
DISTFILES := $(shell readlink -f $(TOPSRC)/$(DISTFILES_BASE))
endif

ifndef OBJ
OBJ := $(shell readlink -f $(TOPSRC)/$(OBJ_BASE))
endif

ifndef MINIROOT
MINIROOT := $(OBJ)/miniroot
endif

# Package subdirectories are versioned and split by architecture.
# This may seem pointless, but it allows for a local $(RELEASE)
# tree to contain multiple releases and it allows noarch packages
# to be shared among all supported architectures.  The architecture
# split is handled in the helper scripts and inst, primarily
# because the collection name is added before the architecture
# split.
RELEASE_DIST = $(RELEASE)/dist/$(DISTRIBUTION_VERSION)

# Digest algorithms for archive files
# This can be md5, sha1, sha256, sha384, or sha512
DIGEST_ALGORITHM ?= sha512

# Global CFLAGS and make customizations
ifeq ($(findstring -Wall,$(CFLAGS)),)
CFLAGS += -Wall
endif

ifeq ($(findstring -g,$(CFLAGS)),)
CFLAGS += -g
endif

ifeq ($(findstring -O2,$(CFLAGS)),)
CFLAGS += -O2
endif

ifeq ($(findstring -pipe,$(CFLAGS)),)
endif

# Build with -fPIC by default.  Disable with POSITION_INDEPENDENT_CODE=n in the
# package Makefile.
ifdef POSITION_INDEPENDENT_CODE
POSITION_INDEPENDENT_CODE := $(shell $(parsebool) $(POSITION_INDEPENDENT_CODE))
else
POSITION_INDEPENDENT_CODE ?= y
endif

ifeq ($(POSITION_INDEPENDENT_CODE),y)
CFLAGS += -fPIC
endif

# Build with -fstack-protector-strong by default.  Disable with
# STACK_PROTECTOR_STRONG=n in the package Makefile.
ifdef STACK_PROTECTOR_STRONG
STACK_PROTECTOR_STRONG := $(shell $(parsebool) $(STACK_PROTECTOR_STRONG))
else
STACK_PROTECTOR_STRONG ?= y
endif

ifeq ($(STACK_PROTECTOR_STRONG),y)
CFLAGS += -fstack-protector-strong

# On ppc64, we need to explicitly link with the SSP library.
ifeq ($(TARGET_MACH),ppc64)
LDFLAGS += -lssp
endif
endif

# Build with _FORTIFY_SOURCE=2 by default.  Disable with FORTIFY_SOURCE=n in
# the package Makefile.
ifdef FORTIFY_SOURCE
FORTIFY_SOURCE := $(shell $(parsebool) $(FORTIFY_SOURCE))
else
FORTIFY_SOURCE ?= y
endif

ifeq ($(FORTIFY_SOURCE),y)
CFLAGS += -D_FORTIFY_SOURCE=2
endif

# Build without BIND_NOW in the linker.  Enable BIND_NOW by setting
# LAZY_EVALUATIONS=n in the package Makefile.
ifdef LAZY_EVALUATIONS
LAZY_EVALUATIONS := $(shell $(parsebool) $(LAZY_EVALUATIONS))
else
LAZY_EVALUATIONS ?= n
endif

ifeq ($(LAZY_EVALUATIONS),n)
LDFLAGS += -z now
CFLAGS += -Wl,-z,now
endif

# The x86_64 systems we have span two generations of Intel instruction set
# architecture.  Intel uses a name for each ISA, so it's hard to tell what came
# first and so on.  Here is a rough list of recent ISAs from Intel:
#
#     Core                    core2
#     Nehalem                 nehalem
#     Sandy Bridge            sandybridge
#     Ivy Bridge              ivybridge
#     Haswell                 haswell
#     Broadwell               broadwell
#     Skylake                 skylake
#     Bonnell                 bonnell
#     Silvermont              silvermont
#     Knight's Landing        knl
#     Skylake Server          skylake-avx512
#
# With 'Core' being first and the bottom of the list being most recent.  Note
# that gcc has dedicated definitions for AMD processors and other manufacturers
# processors, but they are not listed here as we are 100% Intel for our x86_64
# systems.  All of our current systems consist of the following processors:
#
#     Intel Xeon E5-2630      Haswell
#     Intel Core i5-3320M     Ivy Bridge
#     Intel Core i7-4600U     Haswell
#     Intel Core i5-4570      Haswell
#     Intel Core i7-6600U     Skylake
#
# When setting default compiler flags, we want to restrict the ISA to the
# lowest common denominator so we don't get builds that will throw illegal
# instruction signals on the older systems.  This is most easily seen in
# programs that make heavy use of floating point operations.  The gmp package
# is a common culprit.
#
# WHERE DID THESE VALUES COME FROM?
# Reading /proc/cpuinfo and using Wikipedia to figure out the ISA.
#
# WON'T BUILDING FOR AN OLDER ISA RESULT IN SLOWER CODE ON A NEWER ISA?
# No.
#
# BUT WHAT ABOUT ALL THOSE FEATURES THAT WE WON'T BE USING ON THE NEW ISA?
# It doesn't matter.
#
# We set the -march flag to the ISA and then always set -mtune=generic.  It is
# important to note that gcc does not have an -march=generic option.  These
# values can be found here:
#
#  http://gcc.gnu.org/onlinedocs/gcc/x86-Options.html
#
ifeq ($(TARGET_MACH),x86_64)
	CFLAGS += -mtune=generic
endif

# The only ppc64 systems we have are Apple Power Mac G5 towers
ifeq ($(TARGET_MACH),ppc64)
	CFLAGS += -mtune=G5
endif

CXXFLAGS ?= $(CFLAGS)

# Some environment variables that we may have set for our own personal use
# may cause problems with the build system.  Unset them here.
GREP_OPTIONS =

# Our mirrors, which we should try before the main sites.
_SUBDIR := $(shell $(ECHO) "$(PKG)" | cut -c1 | tr [A-Z] [a-z])
OUR_MIRRORS += https://www.burdell.org/peachtree/distfiles/$(_SUBDIR)/$(PKG)/

# Base URL for downloading packages.  Under this directory should be a
# subdirectory named 'dist'.
DIST_URL ?= https://www.burdell.org/peachtree/releases/devel/

# Should world builds attempt to download a built version first?
DIST_DOWNLOAD ?= y

# World build defaults

# If set, the world build will be wrapped in "nice -n $(WORLD_PRIORITY)".
# See the man page for nice(1) for more information.
WORLD_PRIORITY ?= 19

# Command used to time the build (comment out to disable).
WORLD_TIME_CMD ?= /usr/bin/time -v

# Temporary directory for distadd(8) to use during builds.
DISTADD_TMPDIR ?= /var/tmp

# Keep each package $(WRKDIR) around after building?  (y/n)
# If you have limited space for $(OBJ), you should set this to 'n'.  The
# only things kept after a world build for each package are the package
# and the debug package (if applicable).
KEEP_WRKDIR ?= y
