#
# Package Makefile defintions and targets
# (include this file at the end of package Makefiles)
#
# {{{ Apache License version 2.0
#
# Copyright 2002-2018 David Cantrell <david.l.cantrell@gmail.com>
#                     Chris Lumens <chris@bangmoney.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# }}}

# Find the top level directory
TOPSRCERR = "*** Unable to find top level directory."
TOPSRC   := $(shell while [ ! -f mk/pkg.mk ]; do if [ "$$(pwd)" = "/" ]; then /bin/echo $(TOPSRCERR) >&2; exit 1; fi; cd ..; done; /bin/echo "$$(pwd)")

include $(TOPSRC)/mk/defs.mk
include $(TOPSRC)/mk/common.mk
include $(TOPSRC)/mk/urls.mk

# Each STAGE can have a pre or post leaf stage provided by the packager.
# If they exist, we execute them in pre-STAGE, STAGE, post-STAGE order.
LEAF_STAGES = pre post
STAGES      = update-patches update-plist fetch checksum extract \
              patch build test fakeinst package clean

# Create a list of all targets and declare as .PHONY so the make does
# not error if the packager has not provided them.
ALL_TARGETS  = $(foreach s, $(LEAF_STAGES), $(addprefix $(s)-, $(STAGES)))
ALL_TARGETS += $(STAGES)
.PHONY: $(ALL_TARGETS)

# Helper commands for each target
prunedistfiles   = $(MKBINDIR)/prune-distfiles.sh
checkupstream    = $(MKBINDIR)/check-upstream.sh
ishttpurl        = $(MKBINDIR)/is-http-url.sh
issourceforgeurl = $(MKBINDIR)/is-sourceforge-url.sh
fetch            = $(MKBINDIR)/fetch.sh
checksum         = $(MKBINDIR)/checksum.sh
extract          = $(MKBINDIR)/extract.sh
patch            = $(MKBINDIR)/patch.sh
build            = $(MKBINDIR)/build.sh
test             = $(MKBINDIR)/test.sh
fakeinst         = $(MKBINDIR)/fakeinst.sh
updatepatches    = $(MKBINDIR)/update-patches.sh
updateplist      = $(MKBINDIR)/update-plist.sh
updatekeyring    = $(MKBINDIR)/update-keyring.sh
makesum          = $(MKBINDIR)/makesum.sh
cleansum         = $(MKBINDIR)/cleansum.sh
buildreqs        = $(MKBINDIR)/buildreqs.sh
usereqs          = $(MKBINDIR)/usereqs.sh
package          = $(MKBINDIR)/package.sh
hardening-check  = $(MKBINDIR)/hardening-check.sh

# plist file names
PLIST     := $(CWD)/plist
PLISTMACH := $(CWD)/plist.$(TARGET_MACH)

# patches subdirectory
PATCHES   := $(CWD)/patches

# Full package name
PACKAGE      := $(PKG)-$(VER)-$(REL).$(TARGET_MACH).dist
DEBUGPACKAGE := $(PKG)-$(VER)-$(REL).debug.$(TARGET_MACH).dist

# Debug path for gdb
DEBUGDIR = $(PREFIX)/debug

# Package-specific build directories
WRKDIR      := $(OBJ)/$(PKG)-$(VER)-$(REL).$(TARGET_MACH)
WRKDIST     ?= $(PKG)-$(VER)
WRKSRC      := $(WRKDIR)/$(WRKDIST)
WRKBLD      := $(WRKDIR)/build-$(TARGET_MACH)
WRKINST     := $(WRKDIR)/pkgroot-$(TARGET_MACH)
WRKDEBUG    := $(WRKDIR)/debug-$(TARGET_MACH)
WRKSYSROOT  := $(OBJ)/sysroot

# Package location specifics
PKGDISTFILES     := $(DISTFILES)/$(shell $(ECHO) $(PKG) | cut -c1 | tr [A-Z] [a-z])/$(PKG)
PKGCOLLECTION    := $(shell pwd | sed -e "s|^$(TOPSRC)/||g" | cut -d '/' -f 1)
FULLPKGPATH      := $(RELEASE_DIST)/$(PKGCOLLECTION)/$(TARGET_MACH)/$(PACKAGE)
FULLDEBUGPKGPATH := $(RELEASE_DIST)/debug/$(PKGCOLLECTION)/$(TARGET_MACH)/$(DEBUGPACKAGE)

# Are we skipping this package for now?
SKIP_REASON ?=
ifdef SKIP
SKIP := $(shell $(parsebool) $(SKIP))
else
ifneq ($(TARGET_MACH),noarch)
ifneq ($(TARGET_MACH),$(shell uname -m))
SKIP ?= y
SKIP_REASON = ($(TARGET_MACH) only)
else
SKIP ?= n
endif
else
SKIP ?= n
endif
endif

# We are not in a world build unless the caller says we are.
ifdef WORLD
WORLD := $(shell $(parsebool) $(WORLD))
else
WORLD ?= n
endif

#########
# WORLD #
#########

ifeq ($(WORLD),y)

# Add --sysroot= to CFLAGS and CXXFLAGS?
ifdef WORLD_CFLAGS_SYSROOT
WORLD_CFLAGS_SYSROOT := $(shell $(parsebool) $(WORLD_CFLAGS_SYSROOT))
else
WORLD_CFLAGS_SYSROOT ?= y
endif

ifdef WORLD_CXXFLAGS_SYSROOT
WORLD_CXXFLAGS_SYSROOT := $(shell $(parsebool) $(WORLD_CXXFLAGS_SYSROOT))
else
WORLD_CXXFLAGS_SYSROOT ?= y
endif

# Add --with-sysroot= to CONFIGURE_OPTS?
ifdef WORLD_WITH_SYSROOT
WORLD_WITH_SYSROOT := $(shell $(parsebool) $(WORLD_WITH_SYSROOT))
else
WORLD_WITH_SYSROOT ?= y
endif

# Add -I to CFLAGS?
ifdef WORLD_CFLAGS_INCLUDEDIR
WORLD_CFLAGS_INCLUDEDIR := $(shell $(parsebool) $(WORLD_CFLAGS_INCLUDEDIR))
else
WORLD_CFLAGS_INCLUDEDIR ?= y
endif

# Add -L to LDFLAGS?
ifdef WORLD_LDFLAGS_LIBDIR
WORLD_LDFLAGS_LIBDIR := $(shell $(parsebool) $(WORLD_LDFLAGS_LIBDIR))
else
WORLD_LDFLAGS_LIBDIR ?= y
endif

# Add -Wl,--sysroot= to LDFLAGS?
ifdef WORLD_LDFLAGS_WL_SYSROOT
WORLD_LDFLAGS_WL_SYSROOT := $(shell $(parsebool) $(WORLD_LDFLAGS_WL_SYSROOT))
else
WORLD_LDFLAGS_WL_SYSROOT ?= y
endif

# Make sure world builds know about sysroot bin and lib directories.
ifeq ($(CONFIGURE_STYLE),gnu)
WORLD_CONFIGURE_ENV_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),haskell)
WORLD_CONFIGURE_ENV_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),qmake)
WORLD_CONFIGURE_ENV_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),cmake)
WORLD_CONFIGURE_ENV_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),meson)
WORLD_CONFIGURE_ENV_PATH ?= y
endif

ifdef WORLD_CONFIGURE_ENV_PATH
WORLD_CONFIGURE_ENV_PATH := $(shell $(parsebool) $(WORLD_CONFIGURE_ENV_PATH))
endif

ifeq ($(CONFIGURE_STYLE),gnu)
WORLD_CONFIGURE_ENV_LD_LIBRARY_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),haskell)
WORLD_CONFIGURE_ENV_LD_LIBRARY_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),qmake)
WORLD_CONFIGURE_ENV_LD_LIBRARY_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),meson)
WORLD_CONFIGURE_ENV_LD_LIBRARY_PATH ?= y
endif

ifdef WORLD_CONFIGURE_ENV_LD_LIBRARY_PATH
WORLD_CONFIGURE_ENV_LD_LIBRARY_PATH := $(shell $(parsebool) $(WORLD_CONFIGURE_ENV_LD_LIBRARY_PATH))
endif

# Add path overrides to build, test, and fakeinst environments?
ifdef WORLD_BUILD_ENV_PATH
WORLD_BUILD_ENV_PATH := $(shell $(parsebool) $(WORLD_BUILD_ENV_PATH))
else
WORLD_BUILD_ENV_PATH ?= y
endif

ifdef WORLD_BUILD_ENV_LD_LIBRARY_PATH
WORLD_BUILD_ENV_LD_LIBRARY_PATH := $(shell $(parsebool) $(WORLD_BUILD_ENV_LD_LIBRARY_PATH))
else
WORLD_BUILD_ENV_LD_LIBRARY_PATH ?= y
endif

ifdef WORLD_BUILD_ENV_PKG_CONFIG
WORLD_BUILD_ENV_PKG_CONFIG := $(shell $(parsebool) $(WORLD_BUILD_ENV_PKG_CONFIG))
else
WORLD_BUILD_ENV_PKG_CONFIG ?= y
endif

ifdef WORLD_TEST_ENV_LD_LIBRARY_PATH
WORLD_TEST_ENV_LD_LIBRARY_PATH := $(shell $(parsebool) $(WORLD_TEST_ENV_LD_LIBRARY_PATH))
else
WORLD_TEST_ENV_LD_LIBRARY_PATH ?= y
endif

ifdef WORLD_TEST_ENV_PATH
WORLD_TEST_ENV_PATH := $(shell $(parsebool) $(WORLD_TEST_ENV_PATH))
else
WORLD_TEST_ENV_PATH ?= y
endif

ifdef WORLD_FAKEINST_ENV_PKG_CONFIG
WORLD_FAKEINST_ENV_PKG_CONFIG := $(shell $(parsebool) $(WORLD_FAKEINST_ENV_PKG_CONFIG))
else
WORLD_FAKEINST_ENV_PKG_CONFIG ?= y
endif

ifdef WORLD_FAKEINST_ENV_PATH
WORLD_FAKEINST_ENV_PATH := $(shell $(parsebool) $(WORLD_FAKEINST_ENV_PATH))
else
ifeq ($(CONFIGURE_STYLE),python)
WORLD_FAKEINST_ENV_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),haskell)
WORLD_FAKEINST_ENV_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),meson)
WORLD_FAKEINST_ENV_PATH ?= y
endif
WORLD_FAKEINST_ENV_PATH ?= n
endif

ifdef WORLD_FAKEINST_ENV_LD_LIBRARY_PATH
WORLD_FAKEINST_ENV_LD_LIBRARY_PATH := $(shell $(parsebool) $(WORLD_FAKEINST_ENV_LD_LIBRARY_PATH))
else
WORLD_FAKEINST_ENV_LD_LIBRARY_PATH ?= y
endif

# Always add the pkg-config override variables to CONFIGURE_ENV
ifdef WORLD_CONFIGURE_ENV_PKG_CONFIG
WORLD_CONFIGURE_ENV_PKG_CONFIG := $(shell $(parsebool) $(WORLD_CONFIGURE_ENV_PKG_CONFIG))
else
WORLD_CONFIGURE_ENV_PKG_CONFIG ?= y
endif

# Pass SYSROOT= to the environments of certain stages, but allow the
# package to disable it.
#
# gobject-introspection honors the $SYSROOT environment variable
ifdef WORLD_CONFIGURE_ENV_SYSROOT
WORLD_CONFIGURE_ENV_SYSROOT := $(shell $(parsebool) $(WORLD_CONFIGURE_ENV_SYSROOT))
else
WORLD_CONFIGURE_ENV_SYSROOT ?= y
endif

ifdef WORLD_BUILD_ENV_SYSROOT
WORLD_BUILD_ENV_SYSROOT := $(shell $(parsebool) $(WORLD_BUILD_ENV_SYSROOT))
else
WORLD_BUILD_ENV_SYSROOT ?= y
endif

ifdef WORLD_TEST_ENV_SYSROOT
WORLD_TEST_ENV_SYSROOT := $(shell $(parsebool) $(WORLD_TEST_ENV_SYSROOT))
else
WORLD_TEST_ENV_SYSROOT ?= y
endif

ifdef WORLD_FAKEINST_ENV_SYSROOT
WORLD_FAKEINST_ENV_SYSROOT := $(shell $(parsebool) $(WORLD_FAKEINST_ENV_SYSROOT))
else
WORLD_FAKEINST_ENV_SYSROOT ?= y
endif

# If configure style is gnu, run autoreconf on the source?
# (If 'y', we need to make sure the build uses sysroot Perl libs.)
ifeq ($(CONFIGURE_STYLE),gnu)
ifdef AUTORECONF
AUTORECONF := $(shell $(parsebool) $(AUTORECONF))
else
AUTORECONF ?= n
endif

ifeq ($(AUTORECONF),y)
WORLD_USE_PERL5LIB ?= y
CONFIGURE_ENV += AUTOM4TE=$(WRKSYSROOT)$(BINDIR)/autom4te \
                 AUTOCONF=$(WRKSYSROOT)$(BINDIR)/autoconf \
                 AUTOHEADER=$(WRKSYSROOT)$(BINDIR)/autoheader \
                 AUTOMAKE=$(WRKSYSROOT)$(BINDIR)/automake \
                 ACLOCAL=$(WRKSYSROOT)$(BINDIR)/aclocal \
                 ACLOCAL_PATH=$(WRKSYSROOT)$(DATADIR)/alocal:$(WRKSYSROOT)$(X11_DATADIR)/aclocal \
                 AUTOPOINT=$(WRKSYSROOT)$(BINDIR)/autopoint \
                 LIBTOOLIZE=$(WRKSYSROOT)$(BINDIR)/libtoolize \
                 M4=$(WRKSYSROOT)$(BINDIR)/m4
BUILD_ENV += AUTOM4TE=$(WRKSYSROOT)$(BINDIR)/autom4te \
             AUTOCONF=$(WRKSYSROOT)$(BINDIR)/autoconf \
             AUTOHEADER=$(WRKSYSROOT)$(BINDIR)/autoheader \
             AUTOMAKE=$(WRKSYSROOT)$(BINDIR)/automake \
             ACLOCAL=$(WRKSYSROOT)$(BINDIR)/aclocal \
             ACLOCAL_PATH=$(WRKSYSROOT)$(DATADIR)/alocal:$(WRKSYSROOT)$(X11_DATADIR)/aclocal \
             AUTOPOINT=$(WRKSYSROOT)$(BINDIR)/autopoint \
             LIBTOOLIZE=$(WRKSYSROOT)$(BINDIR)/libtoolize \
             M4=$(WRKSYSROOT)$(BINDIR)/m4
endif
endif

# Python needs to know where to find its modules during world builds.  The
# easy way to do that is tell Python to replace its ${exec_prefix} and
# ${prefix} variables throughout its configuration with a temporary path.
# That path can be specified in the PYTHONHOME environment variable making
# this process way way easy.
ifeq ($(CONFIGURE_STYLE),python)
WORLD_USE_PYTHONHOME ?= y
else
WORLD_USE_PYTHONHOME ?= n
endif

ifdef WORLD_USE_PYTHONHOME
WORLD_USE_PYTHONHOME := $(shell $(parsebool) $(WORLD_USE_PYTHONHOME))
endif

# Default to Python 3.x
PYTHON ?= python3

# If meson is used, it needs to find the sysroot Python
ifeq ($(CONFIGURE_STYLE),meson)
WORLD_USE_PYTHONHOME := y
endif

# Perl needs to know where its library files are during world builds.
# By default this is disabled.  If you set it to 'y', the PERL5LIB
# variable is available to you to append or override.  This variable
# is passed in the env variables for the configure, build, test, and
# fakeinst stages.  See the perlrun(1) man page for information on the
# PERL5LIB variable.
ifdef WORLD_USE_PERL5LIB
WORLD_USE_PERL5LIB := $(shell $(parsebool) $(WORLD_USE_PERL5LIB))
else
WORLD_USE_PERL5LIB ?= n
endif

# Add --sysroot= to CFLAGS and CXXFLAGS?
ifeq ($(WORLD_CFLAGS_SYSROOT),y)
CFLAGS += --sysroot=$(WRKSYSROOT)
endif

ifeq ($(WORLD_CXXFLAGS_SYSROOT),y)
CXXFLAGS += --sysroot=$(WRKSYSROOT)
endif

# Add --with-sysroot= to CONFIGURE_OPTS?
ifeq ($(CONFIGURE_STYLE),gnu)
ifeq ($(WORLD_WITH_SYSROOT),y)
CONFIGURE_OPTS += --with-sysroot=$(WRKSYSROOT)
endif
endif

# Add -I to CFLAGS?
ifeq ($(WORLD_CFLAGS_INCLUDEDIR),y)
CFLAGS += -I$(WRKSYSROOT)$(INCLUDEDIR)
CXXFLAGS += -I$(WRKSYSROOT)$(INCLUDEDIR)

ifeq ($(X11),y)
CFLAGS += -I$(WRKSYSROOT)$(X11_INCLUDEDIR)
CXXFLAGS += -I$(WRKSYSROOT)$(X11_INCLUDEDIR)
endif
endif

# Add -L to LDFLAGS?
ifeq ($(WORLD_LDFLAGS_LIBDIR),y)
LDFLAGS += -L$(WRKSYSROOT)$(LIBDIR)

ifeq ($(X11),y)
LDFLAGS += -L$(WRKSYSROOT)$(X11_LIBDIR)
endif
endif

# Add -Wl,--sysroot= to LDFLAGS?
ifeq ($(WORLD_LDFLAGS_WL_SYSROOT),y)
LDFLAGS += -Wl,--sysroot=$(WRKSYSROOT)
endif

# Make sure world builds know about sysroot bin and lib directories.
ifeq ($(WORLD_CONFIGURE_ENV_PATH),y)
CONFIGURE_ENV += PATH=$(WRKSYSROOT)$(BINDIR):$(WRKSYSROOT)$(X11_BINDIR):$(PATH)
endif

ifeq ($(WORLD_CONFIGURE_ENV_LD_LIBRARY_PATH),y)
CONFIGURE_ENV += LD_LIBRARY_PATH=$(WRKSYSROOT)$(LIBDIR):$(WRKSYSROOT)$(X11_LIBDIR)
endif

# Add path overrides to build, test, and fakeinst environments?
ifeq ($(WORLD_BUILD_ENV_PATH),y)
BUILD_ENV += PATH=$(WRKSYSROOT)$(BINDIR):$(WRKSYSROOT)$(X11_BINDIR):$(PATH)
endif

ifeq ($(WORLD_BUILD_ENV_LD_LIBRARY_PATH),y)
BUILD_ENV += LD_LIBRARY_PATH=$(WRKSYSROOT)$(LIBDIR):$(WRKSYSROOT)$(X11_LIBDIR)
endif

ifeq ($(WORLD_TEST_ENV_PATH),y)
TEST_ENV += PATH=$(WRKSYSROOT)$(BINDIR):$(WRKSYSROOT)$(X11_BINDIR):$(PATH)
endif

ifeq ($(WORLD_TEST_ENV_LD_LIBRARY_PATH),y)
TEST_ENV += LD_LIBRARY_PATH=$(WRKSYSROOT)$(LIBDIR):$(WRKSYSROOT)$(X11_LIBDIR)
endif

ifeq ($(WORLD_FAKEINST_ENV_PATH),y)
FAKEINST_ENV += PATH=$(WRKSYSROOT)$(BINDIR):$(WRKSYSROOT)$(X11_BINDIR):$(PATH)
endif

ifeq ($(WORLD_FAKEINST_ENV_LD_LIBRARY_PATH),y)
FAKEINST_ENV += LD_LIBRARY_PATH=$(WRKSYSROOT)$(LIBDIR):$(WRKSYSROOT)$(X11_LIBDIR)
endif

# Add the pkg-config override variables to CONFIGURE_ENV
ifeq ($(WORLD_CONFIGURE_ENV_PKG_CONFIG),y)
CONFIGURE_ENV += PKG_CONFIG_SYSROOT_DIR=$(WRKSYSROOT) \
                 PKG_CONFIG_PATH=$(WRKSYSROOT)$(LIBDIR)/pkgconfig:$(WRKSYSROOT)$(X11_LIBDIR)/pkgconfig:$(LIBDIR)/pkgconfig:$(DATADIR)/pkgconfig$(EXTRA_PKG_CONFIG_PATH)
endif

# Add the pkg-config override variables to BUILD_ENV
ifeq ($(WORLD_BUILD_ENV_PKG_CONFIG),y)
BUILD_ENV += PKG_CONFIG_SYSROOT_DIR=$(WRKSYSROOT) \
             PKG_CONFIG_PATH=$(WRKSYSROOT)$(LIBDIR)/pkgconfig:$(WRKSYSROOT)$(X11_LIBDIR)/pkgconfig:$(LIBDIR)/pkgconfig:$(DATADIR)/pkgconfig$(EXTRA_PKG_CONFIG_PATH)
endif

ifeq ($(WORLD_FAKEINST_ENV_PKG_CONFIG),y)
FAKEINST_ENV += PKG_CONFIG_SYSROOT_DIR=$(WRKSYSROOT) \
                PKG_CONFIG_PATH=$(WRKSYSROOT)$(LIBDIR)/pkgconfig:$(WRKSYSROOT)$(X11_LIBDIR)/pkgconfig:$(LIBDIR)/pkgconfig:$(DATADIR)/pkgconfig$(EXTRA_PKG_CONFIG_PATH)
endif

# Pass SYSROOT= to the environments of certain stages, but allow the
# package to disable it.
#
# gobject-introspection honors the $SYSROOT environment variable
ifeq ($(WORLD_CONFIGURE_ENV_SYSROOT),y)
CONFIGURE_ENV += SYSROOT=$(WRKSYSROOT)
endif

ifeq ($(WORLD_BUILD_ENV_SYSROOT),y)
BUILD_ENV += SYSROOT=$(WRKSYSROOT)
endif

ifeq ($(WORLD_TEST_ENV_SYSROOT),y)
TEST_ENV += SYSROOT=$(WRKSYSROOT)
endif

ifeq ($(WORLD_FAKEINST_ENV_SYSROOT),y)
FAKEINST_ENV += SYSROOT=$(WRKSYSROOT)
endif

# Python needs to know where to find its modules during world builds.  The
# easy way to do that is tell Python to replace its ${exec_prefix} and
# ${prefix} variables throughout its configuration with a temporary path.
# That path can be specified in the PYTHONHOME environment variable making
# this process way way easy.
ifeq ($(WORLD_USE_PYTHONHOME),y)
CONFIGURE_ENV += PYTHONHOME=$(WRKSYSROOT)$(PREFIX)
BUILD_ENV += PYTHONHOME=$(WRKSYSROOT)$(PREFIX)
TEST_ENV += PYTHONHOME=$(WRKSYSROOT)$(PREFIX)
FAKEINST_ENV += PYTHONHOME=$(WRKSYSROOT)$(PREFIX)
endif

# Perl needs to know where its library files are during world builds.
# By default this is disabled.  If you set it to 'y', the PERL5LIB
# variable is available to you to append or override.  This variable
# is passed in the env variables for the configure, build, test, and
# fakeinst stages.  See the perlrun(1) man page for information on the
# PERL5LIB variable.
ifeq ($(WORLD_USE_PERL5LIB),y)
PERL5LIB ?= $(WRKSYSROOT)$(LIBDIR)/perl:$(WRKSYSROOT)$(DATADIR)/perl:$(WRKSYSROOT)$(LIBDIR)/perl/vendor_perl:$(WRKSYSROOT)$(DATADIR)/perl/vendor_perl
CONFIGURE_ENV += PERL5LIB=$(PERL5LIB) \
                 PERL_LIB=$(WRKSYSROOT)$(DATADIR)/perl \
                 PERL_ARCHLIB=$(WRKSYSROOT)$(LIBDIR)/perl \
                 PERL_INC=$(WRKSYSROOT)$(LIBDIR)/perl/CORE
BUILD_ENV += PERL5LIB=$(PERL5LIB) \
             PERL_LIB=$(WRKSYSROOT)$(DATADIR)/perl \
             PERL_ARCHLIB=$(WRKSYSROOT)$(LIBDIR)/perl \
             PERL_INC=$(WRKSYSROOT)$(LIBDIR)/perl/CORE
TEST_ENV += PERL5LIB=$(PERL5LIB) \
            PERL_LIB=$(WRKSYSROOT)$(DATADIR)/perl \
            PERL_ARCHLIB=$(WRKSYSROOT)$(LIBDIR)/perl \
            PERL_INC=$(WRKSYSROOT)$(LIBDIR)/perl/CORE
FAKEINST_ENV += PERL5LIB=$(PERL5LIB) \
                PERL_LIB=$(WRKSYSROOT)$(DATADIR)/perl \
                PERL_ARCHLIB=$(WRKSYSROOT)$(LIBDIR)/perl \
                PERL_INC=$(WRKSYSROOT)$(LIBDIR)/perl/CORE
endif

endif

# Unset LD_RUN_PATH
BUILD_ENV += --unset=LD_RUN_PATH
TEST_ENV += --unset=LD_RUN_PATH
FAKEINST_ENV += --unset=LD_RUN_PATH

# Make sure we use the sysroot gconv files
BUILD_ENV += GCONV_PATH=$(WRKSYSROOT)$(LIBDIR)/gconv
TEST_ENV += GCONV_PATH=$(WRKSYSROOT)$(LIBDIR)/gconv
FAKEINST_ENV += GCONV_PATH=$(WRKSYSROOT)$(LIBDIR)/gconv

# Set QT_SYSROOT for qmake builds
ifeq ($(CONFIGURE_STYLE),qmake)
BUILD_ENV += QT_SYSROOT=$(WRKSYSROOT)
endif

# Use different compression algorithm in distmk depending on architecture.
ifeq ($(TARGET_MACH),ppc64)
COMPRESSION ?= gzip
else
COMPRESSION ?= lzma
endif

# Main driver for packages, we want a 'make' with no arguments to just
# run through build and stop.  Users have to explicitly tell it to fakeinst,
# package, or clean.
all: do-post-build

# Utility targets (helpful to other targets or to packagers)
do-mkdir-wrkdir:
	@mkdir -p $(WRKDIR)

do-mkdir-wrkinst:
	@mkdir -p $(WRKINST)

do-mkdir-distfiles:
	@mkdir -p $(DISTFILES)

do-mkdir-release:
	@mkdir -p $(RELEASE)

do-mkdir-patches:
	@mkdir -p $(PATCHES)

do-mkdir-wrksysroot:
	@mkdir -p $(WRKSYSROOT)

do-pp-archives: do-mkdir-wrkdir
	@if [ -f "$(CWD)/archives" ]; then \
		cpp -E -dU -pipe -fdirectives-only -nostdinc -I$(TOPSRC) \
			$(CWD)/archives -o $(WRKDIR)/archives ; \
	fi

# This target is called during 'make world' runs.  It installs the packages
# listed in the buildreqs file to a sysroot tree.  The 'make world' stage will
# pass --sysroot to gcc so that the package build uses headers and libraries
# from that tree rather than the active system.  This is similar to doing a
# chroot build except that the runtime commands are called from the running
# system, it's only headers and libraries that are read from the sysroot tree.
usereqs: do-mkdir-wrksysroot do-mkdir-wrkdir
	@if [ "$(CACHE_SYSROOT)" = "n" ]; then \
		env TOPSRC='$(TOPSRC)' \
		    WRKDIR='$(WRKDIR)' \
		    WRKSYSROOT='$(WRKSYSROOT)' \
		    PKGCOLLECTION='$(PKGCOLLECTION)' \
		    MAKE='$(MAKE)' \
		    DIST_URL='$(DIST_URL)' \
		    DIST_DOWNLOAD='$(DIST_DOWNLOAD)' \
		    DIGEST_ALGORITHM='$(DIGEST_ALGORITHM)' \
		    RELEASE='$(RELEASE)' \
		    DISTADD_TMPDIR='$(DISTADD_TMPDIR)' \
		$(usereqs) || exit 1 ; \
	fi

# Helper target for world-builder to show the contents of buildreqs.
show-buildreqs: do-mkdir-wrkdir
	@env TOPSRC='$(TOPSRC)' \
	     WRKDIR='$(WRKDIR)' \
	     REDUCE='n' \
	     IGNORE_REQS='$(IGNORE_REQS)' \
	     TARGET_MACH='$(TARGET_MACH)' \
	$(buildreqs) || exit 1

# Utility function to reduce an existing buildreqs file.
reduce-buildreqs: do-mkdir-wrkdir
	@env TOPSRC='$(TOPSRC)' \
	     WRKDIR='$(WRKDIR)' \
	     REDUCE='y' \
	     IGNORE_REQS='$(IGNORE_REQS)' \
	     TARGET_MACH='$(TARGET_MACH)' \
	$(buildreqs) || exit 1

# Run the hardening-check tool on all ELF files in the packaging root
hardening-check: do-mkdir-wrkinst
	@env WRKINST='$(WRKINST)' LIBDIR='$(LIBDIR)' $(hardening-check) || exit 1

# Perform a world-style build of the package.  This is intended for use
# when a world build fails and you want to work on the immediate package
# that failed.  Do not perform a top level 'make realclean' because those
# built packages will be needed by this test build.

ifeq ($(CACHE_SYSROOT),)
CACHE_SYSROOT ?= n
endif

world:
	@if [ "$(SKIP)" = "y" ]; then \
		echo "Skipping $(PACKAGE) $(SKIP_REASON), set SKIP=n to override." ; \
	else \
		env WORLD=y CACHE_SYSROOT='$(CACHE_SYSROOT)' $(MAKE) clean || exit $$? ; \
		env WORLD=y CACHE_SYSROOT='$(CACHE_SYSROOT)' $(MAKE) usereqs || exit $$? ; \
		env WORLD=y CACHE_SYSROOT='$(CACHE_SYSROOT)' $(MAKE) package || exit $$? ; \
		if [ "$(KEEP_WRKDIR)" = "n" ]; then \
			env WORLD=y CACHE_SYSROOT='$(CACHE_SYSROOT)' $(MAKE) clean || exit $$? ; \
		fi ; \
	fi

quickworld:
	@env CACHE_SYSROOT='y' $(MAKE) world || exit $$?

comma:=,

show:
	@if [ -z "$(var)" ]; then \
		$(ECHO) "Usage: make show var=VARIABLE[,VARIABLE2,...]" ; \
	else \
		$(ECHO) $(foreach v,$(subst $(comma), ,$(var)),"$($(v))") ; \
	fi

# Increment the release number and apply the change to the Makefile
bumprel:
	@sed -i -e "s/^REL\ *\=\ *$(REL)/REL\ =\ $$(($(REL) + 1))/g" Makefile

# Show the PKG-VER-REL of a package
verrel:
	@echo "$(PKG)-$(VER)-$(REL)"

# Show the license information for the package
license:
	@. $(CWD)/pkginfo ; \
	[ -s $(CWD)/legal ] && legal="" || legal=" (missing legal file)" ; \
	printf "%-35s %-35s\n" $(PKG)-$(VER)-$(REL) "$${license}$${legal}"

# Remove old distfiles, keep current ones
prune-distfiles: do-pp-archives do-mkdir-distfiles
	@env WRKDIR='$(WRKDIR)' \
	     DISTFILES='$(DISTFILES)' \
	     PKGDISTFILES='$(PKGDISTFILES)' \
	     SHARED_DISTFILES='$(SHARED_DISTFILES)' \
	     TOPSRC='$(TOPSRC)' \
	$(prunedistfiles)

# Variables specific to the upstream version checking target.
# Packages can override these settings in 'upstream' file.

# Default URL for Haskell packages is different than other packages.
ifeq ($(CONFIGURE_STYLE),haskell)
UPSTREAM_URLS ?= $(HACKAGE_UPSTREAM)
else
UPSTREAM_URLS ?= $(URLS)
endif

# hackage.haskell.org no longer offers directory browsing, so we're going to
# scrape the HTML output.  It's ugly, but that's really the easiest option
# here.  This will likely change again, so we need to keep an eye on it.  I
# noticed it this time around because 'make check-upstream' in every haskell
# package failed.
ifeq ($(UPSTREAM_URLS),$(HACKAGE_UPSTREAM))
UPSTREAM_REGEXP ?= Downloads.*$(PKG)-[0-9\.]
UPSTREAM_PREFIX ?= .*Downloads\ *$(PKG)-
endif

ifeq ($(shell $(ishttpurl) $(UPSTREAM_URLS)),y)
UPSTREAM_REGEXP ?= $(PKG)[-_\.]+[0-9\.]+.+\.(tar|t[bglx]z|tbz2|zip)+(\.[gx]z|\.bz2)*
endif

# Default upstream regexps
UPSTREAM_REGEXP ?= ^$(PKG)[-_\.]+[0-9\.]+.+\.(tar|t[bglx]z|tbz2|zip)+(\.[gx]z|\.bz2)*$$
UPSTREAM_PREFIX ?= $(PKG)[-_\.]+
UPSTREAM_SUFFIX ?= \.(tar|t[bglx]z|tbz2|zip)+(\.[gx]z|\.bz2)*.*

# Check for new upstream version of the package's source.
# See upstream.template for more information.
check-upstream:
	@env PKG='$(PKG)' \
	     VER='$(VER)' \
	     UPSTREAM_URLS='$(UPSTREAM_URLS)' \
	     UPSTREAM_REGEXP='$(UPSTREAM_REGEXP)' \
	     UPSTREAM_PREFIX='$(UPSTREAM_PREFIX)' \
	     UPSTREAM_SUFFIX='$(UPSTREAM_SUFFIX)' \
	     TOPSRC='$(TOPSRC)' \
	$(checkupstream)

# Given an archives file with new entries of just filenames, expand
# to lines for each supported hash type.  e.g., if we have:
#
#     bash-4.2.tar.gz
#
# In the archives file, replace it with one of:
#
#     MD5 <digest>  bash-4.2.tar.gz
#     SHA1 <digest>  bash-4.2.tar.gz
#     SHA256 <digest>  bash-4.2.tar.gz
#     SHA384 <digest>  bash-4.2.tar.gz
#     SHA512 <digest>  bash-4.2.tar.gz
#
# The line that is output is based on the preferred digest algorithm
# currently in use in the tree (set in defs.mk).
#
# Really, all this does is run the coreutils commands to generate the
# digests for the new archive files.  It's just meant as a time saver
# for instances where we are creating a new package or updating an
# existing one.
#
# You can put your new archive names in the archives file, then run
#     make fetch
#     make makesum
# To download them and compute new digest hashes for them.
makesum: do-mkdir-wrkdir
	@env TOPSRC='$(TOPSRC)' \
	     WRKDIR='$(WRKDIR)' \
	     DIGEST_ALGORITHM='$(DIGEST_ALGORITHM)' \
	     PKGDISTFILES='$(PKGDISTFILES)' \
	$(makesum) || exit 1

# Reset the archives file to just be include lines and archive filenames.
cleansum: do-mkdir-wrkdir
	@env WRKDIR='$(WRKDIR)' $(cleansum) || exit 1

# Walk $(WRKSRC) and generate patches for any files that have a matching
# file of the same name that ends with ".orig".
update-patches: do-mkdir-patches pre-update-patches do-update-patches post-update-patches
do-update-patches:
	@env WRKDIR='$(WRKDIR)' \
	     WRKSRC='$(WRKSRC)' \
	     PATCHES='$(PATCHES)' \
	$(updatepatches) || exit 1

# Generate a packing list file for the package.  The packing list file
# is the same format as the dist-tools attrs(5) file, which shows the
# file or directory name, owner, group, and octal permissions.
update-plist: pre-update-plist do-update-plist post-update-plist
do-update-plist:
	@env VER='$(VER)' \
	     REL='$(REL)' \
	     WRKDIR='$(WRKDIR)' \
	     WRKINST='$(WRKINST)' \
	     PACKAGE='$(PACKAGE)' \
	     TARGET_NAME='$(TARGET_NAME)' \
	     TARGET_MACH='$(TARGET_MACH)' \
	$(updateplist) || exit 1

# If $(GPGKEYS) is set, update the $(PKG).gpg file with the key pointed
# to by this variable.
update-keyring:
	@env GPG='$(GPG)' \
	     GPGKEYS='$(GPGKEYS)' \
	     KEYRING='$(KEYRING)' \
	     TRUSTDB='$(TRUSTDB)' \
	$(updatekeyring) || exit 1

# List the files and directories in $WRKINST, if it exists
list-wrkinst:
	@if [ -d "$(WRKINST)" ]; then \
		find $(WRKINST) | sed -e "s|$(WRKINST)||g" | grep -v -E "^$$" ; \
	else \
		$(ECHO) "$(WRKINST) does not exist." ; \
	fi

# STAGE: fetch

# Set this to 'y' to tell the fetch target to not download from mirrors.
ifdef FORCE_REFRESH
FORCE_REFRESH := $(shell $(parsebool) $(FORCE_REFRESH))
else
FORCE_REFRESH ?= n
endif

fetch: do-pp-archives do-mkdir-distfiles pre-fetch do-fetch post-fetch
do-fetch:
	@env TOPSRC='$(TOPSRC)' \
	     MAKE='$(MAKE)' \
	     WRKDIR='$(WRKDIR)' \
	     OUR_MIRRORS='$(OUR_MIRRORS)' \
	     FORCE_REFRESH='$(FORCE_REFRESH)' \
	     URLS='$(URLS)' \
	     PKGDISTFILES='$(PKGDISTFILES)' \
	     DISTFILES='$(DISTFILES)' \
	$(fetch) || exit 1

checksum: fetch
	@env WRKDIR='$(WRKDIR)' \
	     PKGDISTFILES='$(PKGDISTFILES)' \
	$(checksum) || exit 1

# STAGE: extract

ifdef GPGKEYS
KEYRING ?= $(CWD)/keyring.gpg
TRUSTDB ?= $(CWD)/trustdb.gpg

# gpg command line to run in the different build targets.  By putting
# it here, we only have to update the common options in one place.
GPG ?= gpg --no-use-agent --no-sig-cache --no-default-keyring --allow-non-selfsigned-uid --trust-model pgp

ifdef GPG_KEYSERVER
GPG += --keyserver $(GPG_KEYSERVER)
endif

ifeq ($(GPG_SIGNATURE_ENDINGS),)
GPG_SIGNATURE_ENDINGS ?= .sig .asc .sign
endif
endif

extract: checksum pre-extract do-extract post-extract
do-extract:
	@env PKGDISTFILES='$(PKGDISTFILES)' \
	     WRKDIR='$(WRKDIR)' \
	     WRKDIST='$(WRKDIST)' \
	     GPG='$(GPG)' \
	     KEYRING='$(KEYRING)' \
	     TRUSTDB='$(TRUSTDB)' \
	     GPG_SIGNATURE_ENDINGS='$(GPG_SIGNATURE_ENDINGS)' \
	$(extract) || exit 1

# STAGE: patch
# All patches are assumed to have been generated by 'update-patches'.  We
# apply with -p1 to trim the first directory level, -F0 for no fuzz factor,
# and we back up the orignal to a file of the same name ending with ".orig"
# (so that 'update-patches' can be run later too).

# Default directory trim level (-p option on patch)
PATCHLEVEL ?= 1

patch: extract do-mkdir-wrkdir pre-patch do-patch post-patch
do-patch:
	@env WRKDIR='$(WRKDIR)' \
	     WRKSRC='$(WRKSRC)' \
	     PATCHES='$(PATCHES)' \
	     PATCHLEVEL='$(PATCHLEVEL)' \
	$(patch) || exit 1

# STAGE: build

ifdef X11
X11 := $(shell $(parsebool) $(X11))
else
X11 ?= n
endif

ifdef SKIP_BUILD
SKIP_BUILD := $(shell $(parsebool) $(SKIP_BUILD))
else
SKIP_BUILD ?= n
endif

# Default paths
PREFIX         ?= /usr
BINDIR         ?= $(PREFIX)/bin
SBINDIR        ?= $(PREFIX)/sbin
LIBEXECDIR     ?= $(PREFIX)/libexec
LIBDIR         ?= $(PREFIX)/lib
INCLUDEDIR     ?= $(PREFIX)/include
DATADIR        ?= $(PREFIX)/share
MANDIR         ?= $(DATADIR)/man

X11_PREFIX     ?= /usr/X11
X11_BINDIR     ?= $(X11_PREFIX)/bin
X11_SBINDIR    ?= $(X11_PREFIX)/sbin
X11_LIBEXECDIR ?= $(X11_PREFIX)/libexec
X11_LIBDIR     ?= $(X11_PREFIX)/lib
X11_INCLUDEDIR ?= $(X11_PREFIX)/include
X11_DATADIR    ?= $(X11_PREFIX)/share
X11_MANDIR     ?= $(X11_PREFIX)/man

DOCDIR         ?= /usr/share/doc/$(PKG)
LICENSEDIR     ?= /usr/share/licenses/$(PKG)
SYSCONFDIR     ?= /etc
INITDIR        ?= $(SYSCONFDIR)/init.d
LOCALSTATEDIR  ?= /var

# Default configure style
CONFIGURE_STYLE ?= gnu

ifeq ($(CONFIGURE_STYLE),gnu)
# Default ./configure options (for CONFIGURE_STYLE=gnu)
CONFIGURE_OPTS += --sysconfdir=$(SYSCONFDIR) \
                  --localstatedir=$(LOCALSTATEDIR) \

ifeq ($(X11),y)
CONFIGURE_OPTS += --prefix=$(X11_PREFIX) \
                  --bindir=$(X11_BINDIR) \
                  --sbindir=$(X11_SBINDIR) \
                  --libexecdir=$(X11_LIBEXECDIR) \
                  --libdir=$(X11_LIBDIR) \
                  --includedir=$(X11_INCLUDEDIR) \
                  --datadir=$(X11_DATADIR) \
                  --mandir=$(X11_MANDIR)

else
CONFIGURE_OPTS += --prefix=$(PREFIX) \
                  --bindir=$(BINDIR) \
                  --sbindir=$(SBINDIR) \
                  --libexecdir=$(LIBEXECDIR) \
                  --libdir=$(LIBDIR) \
                  --includedir=$(INCLUDEDIR) \
                  --datadir=$(DATADIR) \
                  --mandir=$(MANDIR)
endif
endif

ifeq ($(CONFIGURE_STYLE),haskell)
# Conditionally set the location of helper programs based on what is
# available in sysroot.
ALEXBIN = $(shell [ -x $(WRKSYSROOT)$(BINDIR)/alex ] && echo $(WRKSYSROOT)$(BINDIR)/alex || echo $(BINDIR)/alex)
ALEXDATA = $(shell [ -x $(WRKSYSROOT)$(DATADIR)/haskell/alex ] && echo $(WRKSYSROOT)$(DATADIR)/haskell/alex || echo $(WRKSYSROOT)$(DATADIR)/haskell/alex)
HAPPYBIN = $(shell [ -x $(WRKSYSROOT)$(BINDIR)/happy ] && echo $(WRKSYSROOT)$(BINDIR)/happy || echo $(BINDIR)/happy)
HAPPYDATA = $(shell [ -x $(WRKSYSROOT)$(DATADIR)/haskell/happy ] && echo $(WRKSYSROOT)$(DATADIR)/haskell/happy || echo $(WRKSYSROOT)$(DATADIR)/haskell/happy)
HSCOLOURBIN = $(shell [ -x $(WRKSYSROOT)$(BINDIR)/HsColour ] && echo $(WRKSYSROOT)$(BINDIR)/HsColour || echo $(BINDIR)/HsColour)

CONFIGURE_OPTS += --enable-shared \
                  --enable-executable-dynamic \
                  --datadir=$(DATADIR)/haskell \
                  --extra-include-dirs=$(WRKSYSROOT)$(INCLUDEDIR):$(WRKSYSROOT)$(X11_INCLUDEDIR) \
                  --extra-lib-dirs=$(WRKSYSROOT)$(LIBDIR):$(WRKSYSROOT)$(X11_LIBDIR) \
                  --with-alex=$(ALEXBIN) \
                  --with-happy=$(HAPPYBIN) \
                  --with-hscolour=$(HSCOLOURBIN) \
                  --alex-options=--template=$(ALEXDATA) \
                  --happy-options=--template=$(HAPPYDATA) \
                  --gcc-options=-Wl,--sysroot=$(WRKSYSROOT)

ifeq ($(X11),y)
CONFIGURE_OPTS += --prefix=$(X11_PREFIX) \
                  --libdir=$(X11_LIBDIR) \
                  --libexecdir=$(X11_LIBEXECDIR) \
                  --libexecsubdir=$(PKG)
else
CONFIGURE_OPTS += --prefix=$(PREFIX) \
                  --libdir=$(LIBDIR) \
                  --libexecdir=$(LIBEXECDIR) \
                  --libexecsubdir=$(PKG)
endif
endif

ifeq ($(CONFIGURE_STYLE),cmake)
CONFIGURE_OPTS += -DCMAKE_SYSROOT:PATH=$(WRKSYSROOT) \
                  -DCMAKE_MODULE_PATH:PATH=$(WRKSYSROOT)$(DATADIR)/cmake/Modules:$(WRKSYSROOT)$(X11_DATADIR)/cmake/Modules \
                  -DCMAKE_VERBOSE_MAKEFILE=ON \
                  -DBUILD_SHARED_LIBS:BOOL=ON \
                  -DCMAKE_POLICY_DEFAULT_CMP0017=OLD \
                  -DSYSCONF_INSTALL_DIR:PATH=$(SYSCONFDIR)

ifeq ($(X11),y)
CONFIGURE_OPTS += -DCMAKE_INSTALL_PREFIX:PATH=$(X11_PREFIX) \
                  -DINCLUDE_INSTALL_DIR:PATH=$(X11_INCLUDEDIR) \
                  -DLIB_INSTALL_DIR:PATH=$(X11_LIBDIR) \
                  -DSHARE_INSTALL_PREFIX:PATH=$(X11_DATADIR)
else
CONFIGURE_OPTS += -DCMAKE_INSTALL_PREFIX:PATH=$(PREFIX) \
                  -DINCLUDE_INSTALL_DIR:PATH=$(INCLUDEDIR) \
                  -DLIB_INSTALL_DIR:PATH=$(LIBDIR) \
                  -DSHARE_INSTALL_PREFIX:PATH=$(DATADIR)
endif
endif

ifeq ($(CONFIGURE_STYLE),meson)
CONFIGURE_OPTS += --sysconfdir $(SYSCONFDIR) \
                  --localstatedir $(LOCALSTATEDIR) \
                  --backend ninja \
                  --default-library shared

ifeq ($(X11),y)
CONFIGURE_OPTS += --prefix $(X11_PREFIX) \
                  --libdir $(X11_LIBDIR) \
                  --libexecdir $(X11_LIBEXECDIR) \
                  --bindir $(X11_BINDIR) \
                  --sbindir $(X11_SBINDIR) \
                  --includedir $(X11_INCLUDEDIR) \
                  --datadir $(X11_DATADIR) \
                  --mandir $(X11_MANDIR)
else
CONFIGURE_OPTS += --prefix $(PREFIX) \
                  --libdir $(LIBDIR) \
                  --libexecdir $(LIBEXECDIR) \
                  --bindir $(BINDIR) \
                  --sbindir $(SBINDIR) \
                  --includedir $(INCLUDEDIR) \
                  --datadir $(DATADIR) \
                  --mandir $(MANDIR)
endif
endif

ifeq ($(CONFIGURE_STYLE),python)
CONFIGURE_OPTS += --verbose
endif

# Disable certain things by default for GNU style builds.
ifeq ($(CONFIGURE_STYLE),gnu)
CONFIGURE_ENV += am_cv_langinfo_codeset=no

# Pass --disable-dependency-tracking to ./configure script?
ifdef DISABLE_DEPENDENCY_TRACKING
DISABLE_DEPENDENCY_TRACKING := $(shell $(parsebool) $(DISABLE_DEPENDENCY_TRACKING))
else
DISABLE_DEPENDENCY_TRACKING ?= y
endif

ifeq ($(DISABLE_DEPENDENCY_TRACKING),y)
CONFIGURE_OPTS += --disable-dependency-tracking
endif

# Pass --disable-silent-rules to ./configure script?
ifdef DISABLE_SILENT_RULES
DISABLE_SILENT_RULES := $(shell $(parsebool) $(DISABLE_SILENT_RULES))
else
DISABLE_SILENT_RULES ?= y
endif

ifeq ($(DISABLE_SILENT_RULES),y)
CONFIGURE_OPTS += --disable-silent-rules
endif
endif

# Separate build directory from the source directory (recommended)
ifeq ($(CONFIGURE_STYLE),haskell)
SEPARATE_BUILD ?= n
endif

ifeq ($(CONFIGURE_STYLE),python)
SEPARATE_BUILD ?= n
endif

ifeq ($(CONFIGURE_STYLE),qmake)
SEPARATE_BUILD ?= n
endif

ifeq ($(CONFIGURE_STYLE),cargo)
SEPARATE_BUILD ?= n
endif

ifdef SEPARATE_BUILD
SEPARATE_BUILD := $(shell $(parsebool) $(SEPARATE_BUILD))
else
SEPARATE_BUILD ?= y
endif

ifeq ($(SEPARATE_BUILD),n)
WRKBLD := $(WRKSRC)
endif

# Default target
ifeq ($(CONFIGURE_STYLE),python)
BUILD_TARGET ?= build
endif

# Default command
BUILD_COMMAND ?= $(MAKE)

BUILD_TARGET ?= all

build: patch do-mkdir-wrkdir pre-build do-build post-build
do-build:
	@if [ "$(SKIP)" = "y" ]; then \
		echo "Skipping $(PACKAGE) $(SKIP_REASON), set SKIP=n to override." ; \
	else \
		env WRKDIR='$(WRKDIR)' \
		    WRKSRC='$(WRKSRC)' \
		    WRKBLD='$(WRKBLD)' \
		    WRKSYSROOT='$(WRKSYSROOT)' \
		    PREFIX='$(PREFIX)' \
		    SYSCONFDIR='$(SYSCONFDIR)' \
		    BINDIR='$(BINDIR)' \
		    LIBDIR='$(LIBDIR)' \
		    LIBEXECDIR='$(LIBEXECDIR)' \
		    DATADIR='$(DATADIR)' \
		    INCLUDEDIR='$(INCLUDEDIR)' \
		    SEPARATE_BUILD='$(SEPARATE_BUILD)' \
		    CONFIGURE_STYLE='$(CONFIGURE_STYLE)' \
		    CONFIGURE_ENV='$(CONFIGURE_ENV)' \
		    CONFIGURE_OPTS='$(CONFIGURE_OPTS)' \
		    AUTORECONF='$(AUTORECONF)' \
		    PKG='$(PKG)' \
		    VER='$(VER)' \
		    PACKAGE='$(PACKAGE)' \
		    CFLAGS='$(CFLAGS)' \
		    CXXFLAGS='$(CXXFLAGS)' \
		    LDFLAGS='$(LDFLAGS)' \
		    PYTHON='$(PYTHON)' \
		    BUILD_ENV='$(BUILD_ENV)' \
		    BUILD_TARGET='$(BUILD_TARGET)' \
		    BUILD_ARGS='$(BUILD_ARGS)' \
		    BUILD_COMMAND='$(BUILD_COMMAND)' \
		    PARALLEL_JOBS='$(PARALLEL_JOBS)' \
		    TOPSRC='$(TOPSRC)' \
		    SKIP_BUILD='$(SKIP_BUILD)' \
		$(build) || exit 1 ; \
	fi

# STAGE: test

ifeq ($(CONFIGURE_STYLE),haskell)
SKIP_TEST ?= y
endif

ifdef SKIP_TEST
SKIP_TEST := $(shell $(parsebool) $(SKIP_TEST))
else
SKIP_TEST ?= n
endif

ifeq ($(SKIP_TEST),n)
ifeq ($(CONFIGURE_STYLE),meson)
TEST_TARGET ?= test
endif

TEST_TARGET ?= check
endif

TEST_COMMAND ?= $(MAKE)

test: build do-mkdir-wrkdir pre-test do-test post-test
do-test:
	@if [ "$(SKIP)" = "y" ]; then \
		echo "Skipping $(PACKAGE) $(SKIP_REASON), set SKIP=n to override." ; \
	else \
		env WRKDIR='$(WRKDIR)' \
		    WRKSRC='$(WRKSRC)' \
		    WRKBLD='$(WRKBLD)' \
		    PACKAGE='$(PACKAGE)' \
		    CONFIGURE_STYLE='$(CONFIGURE_STYLE)' \
		    PYTHON='$(PYTHON)' \
		    SEPARATE_BUILD='$(SEPARATE_BUILD)' \
		    TEST_ENV='$(TEST_ENV)' \
		    TEST_COMMAND='$(TEST_COMMAND)' \
		    TEST_TARGET='$(TEST_TARGET)' \
		    TEST_ARGS='$(TEST_ARGS)' \
		    SKIP_TEST='$(SKIP_TEST)' \
		$(test) || exit 1 ; \
	fi

# STAGE: fakeinst

# Default target
FAKEINST_TARGET ?= install
FAKEINST_COMMAND ?= $(MAKE)

# Disable the debug package creation?
NO_DEBUG ?= n

# Strip DT_RPATH values unless instructed otherwise
ifdef STRIP_DT_RPATH
STRIP_DT_RPATH := $(shell $(parsebool) $(STRIP_DT_RPATH))
else
STRIP_DT_RPATH ?= y
endif

ifdef SKIP_FAKEINST
SKIP_FAKEINST := $(shell $(parsebool) $(SKIP_FAKEINST))
else
SKIP_FAKEINST ?= n
endif

fakeinst: test do-mkdir-wrkinst pre-fakeinst do-fakeinst post-fakeinst
do-fakeinst:
	@if [ "$(SKIP)" = "y" ]; then \
		echo "Skipping $(PACKAGE) $(SKIP_REASON), set SKIP=n to override." ; \
	else \
		env PKG='$(PKG)' \
		    VER='$(VER)' \
		    WRKDIR='$(WRKDIR)' \
		    WRKDIST='$(WRKDIST)' \
		    WRKSRC='$(WRKSRC)' \
		    WRKBLD='$(WRKBLD)' \
		    PREFIX='$(PREFIX)' \
		    BINDIR='$(BINDIR)' \
		    LIBDIR='$(LIBDIR)' \
		    SEPARATE_BUILD='$(SEPARATE_BUILD)' \
		    WRKSYSROOT='$(WRKSYSROOT)' \
		    WRKINST='$(WRKINST)' \
		    WRKDEBUG='$(WRKDEBUG)' \
		    DEBUGDIR='$(DEBUGDIR)' \
		    NO_DEBUG='$(NO_DEBUG)' \
		    CONFIGURE_STYLE='$(CONFIGURE_STYLE)' \
		    PYTHON='$(PYTHON)' \
		    PKG='$(PKG)' \
		    VER='$(VER)' \
		    PACKAGE='$(PACKAGE)' \
		    FAKEINST_ENV='$(FAKEINST_ENV)' \
		    FAKEINST_TARGET='$(FAKEINST_TARGET)' \
		    FAKEINST_ARGS='$(FAKEINST_ARGS)' \
		    DATADIR='$(DATADIR)' \
		    MANDIR='$(MANDIR)' \
		    DOCDIR='$(DOCDIR)' \
		    LICENSEDIR='$(LICENSEDIR)' \
		    FAKEINST_COMMAND='$(FAKEINST_COMMAND)' \
		    MAKE='$(MAKE)' \
		    TOPSRC='$(TOPSRC)' \
		    TARGET_NAME='$(TARGET_NAME)' \
		    TARGET_MACH='$(TARGET_MACH)' \
		    INCLUDEDIR='$(INCLUDEDIR)' \
		    TARGET_MACH='$(TARGET_MACH)' \
		    STRIP_DT_RPATH='$(STRIP_DT_RPATH)' \
		    DOCS='$(DOCS)' \
		    LICENSES='$(LICENSES)' \
		    IGNORE_REQS='$(IGNORE_REQS)' \
		    SKIP_FAKEINST='$(SKIP_FAKEINST)' \
		$(fakeinst) || exit 1 ; \
	fi

# STAGE: package
package: fakeinst do-mkdir-release pre-package do-package post-package
do-package:
	@if [ "$(SKIP)" = "y" ]; then \
		echo "Skipping $(PACKAGE) $(SKIP_REASON), set SKIP=n to override." ; \
	else \
		env WRKDIR='$(WRKDIR)' \
		    WRKINST='$(WRKINST)' \
		    PKG='$(PKG)' \
		    VER='$(VER)' \
		    REL='$(REL)' \
		    TARGET_NAME='$(TARGET_NAME)' \
		    TARGET_MACH='$(TARGET_MACH)' \
		    TOPSRC='$(TOPSRC)' \
		    RELEASE_DIST='$(RELEASE_DIST)' \
		    WRKDEBUG='$(WRKDEBUG)' \
		    PKGCOLLECTION='$(PKGCOLLECTION)' \
		    NAME_PREFIX_CONFLICT='$(NAME_PREFIX_CONFLICT)' \
		    MULTIPLE_VERSIONS='$(MULTIPLE_VERSIONS)' \
		    DIGEST_ALGORITHM='$(DIGEST_ALGORITHM)' \
		    DIST_URL='$(DIST_URL)' \
		    DIST_DOWNLOAD='$(DIST_DOWNLOAD)' \
		    PACKAGE='$(PACKAGE)' \
		    COMPRESSION='$(COMPRESSION)' \
		$(package) || exit 1 ; \
	fi

# STAGE: clean
clean: pre-clean do-clean post-clean
do-clean:
	-rm -rf $(WRKDIR)

help:
	@echo
	@$(ECHO) "Peachtree Linux package building system"
	@echo
	@$(ECHO) "Package building consists of multiple stages, which can be supplemented with"
	@$(ECHO) "leaf stages.  The following table shows what stages there are and which you"
	@$(ECHO) "can override:"
	@echo
	@$(ECHO) "    +----------+------------+-----------+"
	@$(ECHO) "    |          |   LEAFS    |           |"
	@$(ECHO) "    | STAGE    | PRE | POST | OVERRIDE? |"
	@$(ECHO) "    +----------+-----+------+-----------+"
	@$(ECHO) "    | fetch    | yes | yes  | yes       |"
	@$(ECHO) "    | checksum | no  | no   | no        |"
	@$(ECHO) "    | extract  | yes | yes  | yes       |"
	@$(ECHO) "    | patch    | yes | yes  | no        |"
	@$(ECHO) "    | build    | yes | yes  | yes       |"
	@$(ECHO) "    | test     | yes | yes  | yes       |"
	@$(ECHO) "    | fakeinst | yes | yes  | yes       |"
	@$(ECHO) "    | package  | yes | yes  | no        |"
	@$(ECHO) "    | clean    | yes | yes  | yes       |"
	@$(ECHO) "    +----------+-----+------+-----------+"
	@echo
	@$(ECHO) "The different stages have a dependent order, specifying one stage will"
	@$(ECHO) "invoke dependent stages in order.  The general order is as they are shown"
	@$(ECHO) "in the table.  Each stage with a pre or post leaf will have that target run"
	@$(ECHO) "either before or after the main stage target, respectively."
	@echo
	@$(ECHO) "If you override one of the stages, the leaf stages are disabled unless you"
	@$(ECHO) "explicitly set them up in your package Makefile."
	@echo
	@$(ECHO) "Typing 'make' with no target specified will run through the build target."
	@echo
	@$(ECHO) "Stage targets:"
	@$(ECHO) "    fetch            Fetch the source files listed in 'archives', store in"
	@$(ECHO) "                     $(DISTFILES)."
	@$(ECHO) "                     Run as 'env FORCE_REFRESH=y make fetch' to skip any"
	@$(ECHO) "                     mirrors and download from the upstream location."
	@$(ECHO) "    checksum         Verify checksums of downloaded source files."
	@$(ECHO) "    extract          Extract source files listed in 'archives', to"
	@$(ECHO) "                     $(WRKDIR)."
	@$(ECHO) "    patch            Apply patches."
	@$(ECHO) "    build            Configure and build source and /sbin source if."
	@$(ECHO) "                     applicable."
	@$(ECHO) "    test             Run test suite included with $(PACKAGE)."
	@$(ECHO) "    fakeinst         Install software to $(WRKINST)."
	@$(ECHO) "    package          Generate an installable package for $(PACKAGE)."
	@$(ECHO) "    clean            Remove $(WRKDIR)."
	@echo
	@$(ECHO) "User-provided leaf targets:"
	@$(ECHO) "    pre-fetch        Runs before 'fetch'."
	@$(ECHO) "    post-fetch       Runs after 'fetch'."
	@$(ECHO) "    pre-extract      Runs before 'extract'."
	@$(ECHO) "    post-extract     Runs after 'extract'."
	@$(ECHO) "    pre-patch        Runs before 'patch'."
	@$(ECHO) "    post-patch       Runs after 'patch'."
	@$(ECHO) "    pre-build        Runs before 'build'."
	@$(ECHO) "    post-build       Runs after 'build'."
	@$(ECHO) "    pre-test         Runs before 'test'."
	@$(ECHO) "    post-test        Runs after 'test'."
	@$(ECHO) "    pre-fakeinst     Runs before 'fakeinst'."
	@$(ECHO) "    post-fakeinst    Runs after 'fakeinst'."
	@$(ECHO) "    pre-package      Runs before 'package'."
	@$(ECHO) "    post-package     Runs after 'package'."
	@$(ECHO) "    pre-clean        Runs before 'clean'."
	@$(ECHO) "    post-clean       Runs after 'clean'."
	@echo
	@$(ECHO) "Helper targets for packagers:"
	@$(ECHO) "    makesum          Generate new checksums in 'archives' for new source"
	@$(ECHO) "                     files."
	@$(ECHO) "    cleansum         Clean out the 'archives' file and leave just filenames"
	@$(ECHO) "                     and #include lines.  This is a helper target when updating"
	@$(ECHO) "                     a package to a new upstream version."
	@$(ECHO) "    update-patches   Walk source tree and generate any new patches for files"
	@$(ECHO) "                     ending with .orig."
	@$(ECHO) "    update-keyring   Update the GPG keyring file if GPGKEYS is set."
	@$(ECHO) "    update-plist     Update the packing list file."
	@$(ECHO) "    list-wrkinst     Show contents of $(WRKINST)"
	@$(ECHO) "    show var=VAR     Show the value of variable VAR."
	@$(ECHO) "                     e.g., make show var=WRKDIR"
	@$(ECHO) "    bumprel          Increment value of REL by 1 (requires REL be an integer)"
	@$(ECHO) "    verrel           Show the package triple PKG-VER-REL."
	@$(ECHO) "    license          Show licensing information for this package"
	@$(ECHO) "    prune-distfiles  Remove old distfiles, keep current ones"
	@$(ECHO) "    check-upstream   Check for a new upstream release"
	@$(ECHO) "    show-buildreqs   Show the packages required to build this package"
	@$(ECHO) "    reduce-buildreqs Shrink existing buildreqs file to minimum required"
	@$(ECHO) "    hardening-check  Run hardening-check script on ELF files in $(WRKINST)"
	@echo
	@$(ECHO) "The business end of package maintenance:"
	@$(ECHO) "    world            Peform all steps through 'make package' but in the style"
	@$(ECHO) "                     of a top level world build.  Requires a populated release"
	@$(ECHO) "                     tree with built packages to satisfy the buildreqs."
	@$(ECHO) "    quickworld       Exactly like world, but does not wipe and reinstall the"
	@$(ECHO) "                     sysroot tree.  Just uses the one in place already."
	@echo
