#!/bin/sh -x
#
# Begin make-ca-bundle.sh
# Script to populate OpenSSL's CApath from a bundle of PEM formatted CAs
#
# The file certdata.txt must exist in the local directory
# Version number is obtained from the version of the data.
#
# Usage: make-ca.sh [certdata] [tmpdir] [convert script]
#
# Authors: DJ Lucas
#          Bruce Dubbs
#
# Version 20120211
#
# Modified 30-Jul-2013 for Peachtree Linux by David Cantrell

mydate() {
    local y=$(echo $1 | cut -d" " -f4)
    local M=$(echo $1 | cut -d" " -f1)
    local d=$(echo $1 | cut -d" " -f2)
    local m

    [ ${d} -lt 10 ] && d="0${d}"

    case $M in
        Jan) m="01";;
        Feb) m="02";;
        Mar) m="03";;
        Apr) m="04";;
        May) m="05";;
        Jun) m="06";;
        Jul) m="07";;
        Aug) m="08";;
        Sep) m="09";;
        Oct) m="10";;
        Nov) m="11";;
        Dec) m="12";;
    esac

    certdate="${y}${m}${d}"
}

certdata="${1}"

if [ ! -r ${certdata} ]; then
    echo "${certdata} must be readable" >&2
    exit 1
fi

TEMPDIR="${2}"
TRUSTATTRIBUTES="CKA_TRUST_SERVER_AUTH"
BUNDLE="ca-bundle.crt"
CONVERTSCRIPT="${3}"

if [ ! -d "${TEMPDIR}" ]; then
    echo "ERROR: ${TEMPDIR} is not a directory." >&2
    exit 3
fi

if [ ! -x "${CONVERTSCRIPT}" ]; then
    echo "ERROR: ${CONVERTSCRIPT} is not an executable." >&2
    exit 4
fi

cd ${TEMPDIR}
mkdir -p "${TEMPDIR}/certs"
rm -f ${TEMPDIR}/certs/*      # Make sure the directory is clean
unset CERTBEGINLIST CERTDATA CERTENDLIST certebegin certend

# Get a list of staring lines for each cert
CERTBEGINLIST=$(grep -n "^# Certificate" "${certdata}" | cut -d ":" -f1)

# Get a list of ending lines for each cert
CERTENDLIST=$(grep -n "^CKA_TRUST_STEP_UP_APPROVED" "${certdata}" | cut -d ":" -f 1)

# Start a loop
for certbegin in ${CERTBEGINLIST}; do
    for certend in ${CERTENDLIST}; do
        if test "${certend}" -gt "${certbegin}"; then
            break
        fi
    done

    # Dump to a temp file with the name of the file as the beginning line number
    sed -n "${certbegin},${certend}p" "${certdata}" > "${TEMPDIR}/certs/${certbegin}.tmp"
done

for tempfile in ${TEMPDIR}/certs/*.tmp; do
    # Make sure that the cert is trusted...
    grep "CKA_TRUST_SERVER_AUTH" "${tempfile}" | \
      egrep "TRUST_UNKNOWN|NOT_TRUSTED" > /dev/null

    if test "${?}" = "0"; then
        # Throw a meaningful error and remove the file
        cp "${tempfile}" ${TEMPDIR}/tempfile.cer
        ${CONVERTSCRIPT} > ${TEMPDIR}/tempfile.crt
        keyhash=$(openssl x509 -noout -in ${TEMPDIR}/tempfile.crt -hash)
        echo "Certificate ${keyhash} is not trusted!  Removing..."
        rm -f ${TEMPDIR}/tempfile.cer ${TEMPDIR}/tempfile.crt "${tempfile}"
        continue
    fi

    # If execution made it to here in the loop, the temp cert is trusted
    # Find the cert data and generate a cert file for it

    cp "${tempfile}" ${TEMPDIR}/tempfile.cer
    ${CONVERTSCRIPT} > ${TEMPDIR}/tempfile.crt
    keyhash=$(openssl x509 -noout -in ${TEMPDIR}/tempfile.crt -hash)
    mv ${TEMPDIR}/tempfile.crt "${TEMPDIR}/certs/${keyhash}.pem"
    rm -f ${TEMPDIR}/tempfile.cer "${tempfile}"
    echo "Created ${keyhash}.pem"
done

# Remove blacklisted files
# MD5 Collision Proof of Concept CA
if test -f ${TEMPDIR}/certs/8f111d69.pem; then
    echo "Certificate 8f111d69 is not trusted!  Removing..."
    rm -f ${TEMPDIR}/certs/8f111d69.pem
fi

# Remove expired certificates
certs=$(find ${TEMPDIR}/certs -type f -name "*.pem" -o -name "*.crt")
today=$(date +%Y%m%d)

for cert in ${certs}; do
    notafter=$(openssl x509 -enddate -in "${cert}" -noout)
    date=$(echo ${notafter} | sed 's/^notAfter=//')
    mydate "${date}"

    if [ ${certdate} -lt ${today} ]; then
        echo "${cert} expired on ${certdate}! Removing..."
        rm -f "${cert}"
    fi
done

# Finally, generate the bundle and clean up.
cat ${TEMPDIR}/certs/*.pem > ${TEMPDIR}/${BUNDLE}
